/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	CLEMENT					*/
/*	Its purpose is to handle SMTP protocole	*/
/*	on SMTP server premises and crunch data	*/
/*	and get clearance from the main Mentor	*/
/*	server to pass the E-mail to local	*/
/*	smtp server (sendmail, postfixe, etc)	*/
/*						*/
/************************************************/
#include	<unistd.h>
#include	<errno.h>
#include	<syslog.h>
#include	<string.h>
#include	<stdio.h>
#include	<signal.h>
#include	"lowtyp.h"
#include	"utlprc.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"subprc.h"
#include	"unijrl.h"
#include	"unilck.h"
#include	"unisql.h"
#include	"unidns.h"
#include	"gesclm.h"
#include	"lvlsta.h"
#include	"lvlcnf.h"
#include	"lvlpro.h"
#include	"lvlipc.h"
#include	"modpro.h"
#include	"modckr.h"
#include	"modmen.h"
#include	"modtrt.h"
#include	"modtst.h"

typedef	struct	{
	CNFTYP *cnf;	/*config information	*/
	LSTTYP **lst;	/*connexion list	*/
	MXCTYP *mxc;	/*message exchange area	*/
	u_int nbrchkr;	/*# of chkrc process	*/
	pid_t mntpid;	/*mentor interface pid	*/
	pid_t *chkrlst;	/*checker process list	*/
	STATYP *stats;	/*statistics collection	*/
	}WRKTYP;
/*

*/
/************************************************/
/*						*/
/*	Procedure to display the program usage	*/
/*						*/
/************************************************/
static void usage()

{
(void) fprintf(stderr,"%s Version %s\n",appname,clm_getvers());
(void) fprintf(stderr,"usage:  %s "
		      "[-b base] "
		      "[-c nbrchkr] "
		      "[-d debug] "
		      "[-f ] "
		      "[-h ] "
		      "[-o ] "
		      "[-t ] "
		      "[-v ] "
		      "\n",appname);
(void) fprintf(stderr,"\t\t-b base      : application working directory\n");
(void) fprintf(stderr,"\t\t-c nbrchkr   : override default checker number\n");
(void) fprintf(stderr,"\t\t-d debug     : debug level\n");
(void) fprintf(stderr,"\t\t-f           : foreground\n");
(void) fprintf(stderr,"\t\t-h           : display usage\n");
(void) fprintf(stderr,"\t\t-o           : allowing ORGN within SMTP protocol\n");
(void) fprintf(stderr,"\t\t-t test|?    : conduct specific modul test\n");
(void) fprintf(stderr,"\t\t-v           : verbose debug\n");
(void) exit(-1);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report about the current	*/
/*	version number against official version	*/
/*						*/
/************************************************/
static void reportvers()

{
if ((debug==0)||(verbose==true)) {
  char *msg;
  char *dnsvers;

  msg=(char *)0;
  dnsvers=(char *)0;
  switch (dns_chksafeversion(&dnsvers)) {
    case -1	:
       (void) asprintf(&msg,"\nYour current %s version is <%s> "
			    "while the last released one is <%s>.\n"
			    "You should update your %s program",
			    appname,clm_getvers(),dnsvers,appname);
      break;
    case  0	:
      break;
    case +1	:
       (void) asprintf(&msg,"\nYour current %s-%s is a BETA "
			    "version, the last stable release is <%s>.\n"
			    "Please report any discrepancy or bugs "
			    "you may find to <clement@safe.ca>.\n",
			    appname,clm_getvers(),dnsvers);
      break;
    }
  if ((msg!=(char *)0)&&(strlen(msg)>0)) {
    (void) fprintf(stderr,"%s\n",msg);
    (void) jrl_journalling(true,APJRL,"Note:%s",msg);
    (void) free(msg);
    }
  dnsvers=clm_freestr(dnsvers);
  }
}
/*

*/
/************************************************/
/*						*/
/*	procedure to check if daemon is up	*/
/*						*/
/************************************************/
static int isdbdaemonup()

{
return sql_isdaemonup((const char *)0,(const char *)0,(const char *)0);
}
/*

*/
/************************************************/
/*						*/
/*	procedure to prepare a message exchange	*/
/*	area, return false if not successful.	*/
/*						*/
/************************************************/
static int prepmsgbox(WRKTYP *wrk,u_int kilobytes,u_int nbrproc)

{
#define	OPPR	"clement.c:prepmsgbox,"

int status;

status=false;
nbrproc+=2;
if (wrk->lst!=(LSTTYP **)0) {
  register int i;

  for (i=0;wrk->lst[i]!=(LSTTYP *)0;i++) {
    nbrproc+=wrk->lst[i]->binding->iteration;
    }
  }
if ((wrk->mxc=ipc_getmsgxchg(kilobytes,nbrproc))!=(MXCTYP *)0) {
  if (ipc_subscribe(wrk->mxc->bptr,(u_long)0,true)!=true)
    (void) clm_alert(0,"%s Unable to subscribe to message exchgr",OPPR);
  else
    status=true;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to start the mentor process 	*/
/*						*/
/************************************************/
static void getmentor(WRKTYP *wrk)

{
if (wrk->mntpid==(pid_t)0) {
  GACTYP *gat;

  gat=(GACTYP *)0;
  (void) prc_nozombie();
  switch (wrk->mntpid=fork()) {
    case -1	:
      (void) clm_alert(0,"%s Unable to fork getmentor process (error=<%s>)",
			  appname,strerror(errno));
      break;
    case  0	:
      (void) jrl_closejournals();
      /*module own garbage collector	*/
      gat=gac_open(2,appname,"mentor");
      (void) gac_start(gat);
      (void) clm_setjournal(APJRL,d_log,appname);
      (void) men_collector(wrk->mxc->bptr,wrk->mxc->mntpool);
      (void) jrl_closejournals();
      (void) clm_trashmem(2,gat);
      (void) prc_cleantitle();
      (void) closelog();
      (void) sleep(1);    /*avoiding avalanche    */
      (void) exit(0);
      break;
    default	:
      break;
    }
  }
else {
  if (prc_checkprocess(wrk->mntpid)==false)
    wrk->mntpid=(pid_t)0;
  }
}
/*

*/
/************************************************/
/*						*/
/*	procedure to start checkers process	*/
/*						*/
/************************************************/
static void checkers(WRKTYP *wrk)

{
register u_int i;

(void) prc_nozombie();
for (i=0;i<wrk->nbrchkr;i++) {
  if (wrk->chkrlst[i]==(pid_t)0) {
    GACTYP *gat;

    gat=(GACTYP *)0;
    switch (wrk->chkrlst[i]=fork()) {
      case -1	:
	(void) clm_alert(0,"%s Unable to fork process (error=<%s>)",
                            appname,strerror(errno));
	break;
      case  0	:
	/*freeing as much as memory as possible		*/
        wrk->cnf=cnf_freeconfig(wrk->cnf);
        wrk->lst=pro_freelst(wrk->lst);
	(void) jrl_closejournals();
        /*module own garbage collector			*/
        gat=gac_open(2,appname,"checker");
        (void) gac_start(gat);
	(void) clm_setjournal(APJRL,d_log,appname);
	(void) clm_setjournal(EVCKR,d_log,"checker");
	(void) ckr_checker(wrk->mxc,wrk->nbrchkr,i+1);
	(void) jrl_closejournals();
    	(void) clm_trashmem(2,gat);
	(void) prc_cleantitle();
	(void) closelog();
        (void) sleep(1);    /*avoiding avalanche    */
        (void) exit(0);
        break;
      default	:
	break;
      }
    }
  else {
    if (prc_checkprocess(wrk->chkrlst[i])==false)
      wrk->chkrlst[i]=(pid_t)0;
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	procedure to stop started sub process	*/
/*						*/
/************************************************/
static void stopsubproc(WRKTYP *wrk)

{
register u_int i;
PRCTYP *prc;

prc=prc_addtoprc((PRCTYP *)0,wrk->mntpid);
for (i=0;i<wrk->nbrchkr;i++)
  prc=prc_addtoprc(prc,wrk->chkrlst[i]);
prc=prc_childkill(prc,5);
}
/*

*/
/************************************************/
/*						*/
/*	procedure to collect ALL data coming	*/
/*	within a statistic msg			*/
/*	process.				*/
/*						*/
/************************************************/
STATYP *collecting(STATYP *stats,char *msg)

{
char **split;

if ((split=clm_splitstring(msg,'\n'))!=(char **)0) {
  stats=sta_msgtostats(stats,split);
  split=(char **)vec_freelstlst((void **)split,VECFREE);
  }
return stats;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to start/check/stop a transitr*/
/*	process.				*/
/*						*/
/************************************************/
static void transitr(int action)

{
static pid_t	p_transitr;	/*transitr pid	*/

if (action==true) {
  if (p_transitr==(pid_t)0) {
    p_transitr=trt_cycle();
    }
  else {
    if (prc_checkprocess(p_transitr)==false)
      p_transitr=(pid_t)0;
    }
  }
else {
  PRCTYP *prc;
  
  prc=(PRCTYP *)0;
  prc=prc_addtoprc(prc,p_transitr);
  prc=prc_childkill(prc,5);
  }
}
/*

*/
/************************************************/
/*						*/
/*	procedure to check about cron job to do	*/
/*						*/
/************************************************/
static void docron(SQLFILE *sql)

{
#ifdef	DEBUG
#define	CRONPACE	10	/*each 10 sec	*/
#else
#define	CRONPACE	60	/*each minute	*/
#endif

static	time_t 	lastcron=(time_t)0;
static	pid_t	cronchild=(pid_t)0;

time_t curtime;
char **crontab;
int phase;
int proceed;

curtime=time((time_t *)0);
crontab=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*time to check?	*/
      if ((lastcron+CRONPACE)>curtime)
	proceed=false;	/*no need to go further	*/
      break;
    case 1	:	/*Child still pending?	*/
      if (prc_checkprocess(cronchild)==true) {
	if ((lastcron+(CRONPACE*30))<curtime) 
          (void) jrl_journalling(true,APJRL,"Clement Cron job '%d' still pending (stuck process?)",cronchild);
	phase=999;	/*no need to proceed	*/
	}
      break;
    case 2	:	/*getting cron list	*/
      if ((crontab=dbc_getcrontab(sql,curtime))==(char **)0)
	phase=999;	/*no need to proceed	*/
      break;
    case 3	:	/*something to do	*/
      if (crontab!=(char **)0) {	/*always*/
	GACTYP *gat;

        gat=(GACTYP *)0;
        switch (cronchild=fork()) {
	  case -1	:
            (void) clm_alert(0,"%s Unable to fork cron process (error=<%s>)",
			        appname,strerror(errno));
	    break;
          case  0	:	/*cron process	*/
	    if (getuid()==0) {
	      /*changing forked real ID to be clement		*/
	      (void) setregid(getegid(),getegid());
	      (void) setreuid(geteuid(),geteuid());
	      }
      	    (void) jrl_closejournals();
	    /*module own garbage collector			*/
	    gat=gac_open(2,appname,"cron");
	    (void) gac_start(gat);
            (void) clm_setjournal(APJRL,d_log,appname);
	    while (*crontab!=(char *)0) {
	      char *cmd;
	      int status;

	      (void) asprintf(&cmd,"%s%s",base,*crontab);
	      (void) prc_settitle("%s: cron doing <%s>",appname,cmd);
	      status=system(cmd);
	      (void) jrl_journalling(true,APJRL,"Cron task <%s> done, status='%d'",cmd,status);
	      (void) free(cmd);
	      crontab++;
	      if (status!=0)
		break;		/*trouble trouble	*/
	      }
	    (void) jrl_closejournals();
    	    (void) clm_trashmem(2,gat);
      	    (void) sleep(1);    /*avoiding avalanche    */
      	    (void) exit(0);
	    break;
          default	:	/*main process	*/
	    break;
	  }
        crontab=(char **)vec_freelstlst((void **)crontab,VECFREE);
	}
      break;
    default	:	/*SAFE Guard		*/
      lastcron=curtime;
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect message coming from*/
/*	clement subprocess.			*/
/*	Message type are:			*/
/*		E-mail statistics		*/
/*		Multiple connection checking.	*/
/*						*/
/************************************************/
static void collectmsg(WRKTYP *wrk)

{
#define	OPCO	"clement.c:collectmsg,"

TIMESPEC timeout;
u_long msec;
char *msgref;
char *msg;

timeout.tv_sec=(time_t)0;
timeout.tv_nsec=(long)10000000;/*10ms, short time */
while ((msg=ipc_getpost(wrk->mxc->bptr,&timeout,&msgref,&msec))!=(char *)0) {
  static VOCTYP vocmsg[]={
        {0,STATCOL,(const void *)0},
        {1,(char *)0,(const void *)0}
        };
  VOCTYP *voc;
  char *answ;
 
  answ=(char *)0; 
  if ((voc=clm_getvoca(vocmsg,msg))!=(VOCTYP *)0) {
    switch (voc->code) {
      case   0	:	/*compiling ALL stats	*/
	wrk->stats=collecting(wrk->stats,msg);	
	break;
      default	:	/*unexpected message	*/
	(void) clm_alert(0,"%s Unexpected msg=<%s> (pid='%d' bug?)",OPCO,msg,getpid());
	break;
      }
    }
  if (answ!=(char *)0) {
    if (ipc_postreply(wrk->mxc->bptr,&timeout,msgref,answ)==(u_long)0) {
      (void) clm_alert(0,"%s, timeout to post answer <%s> to msg <%s>",OPCO,answ);
      }
    (void) free(answ);
    }
  msgref=clm_freestr(msgref);
  (void) free(msg);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Main procedure to run clement		*/
/*						*/
/************************************************/
static void cycle(u_int nbrchkr,u_int nbrmnt)

{
#define	MXBOX	16386	/*msg box kBytes size	*/
#define	BOXSPARE    5	/*# of IPC spare chanel	*/

const char *empty;
WRKTYP *wrk;
SQLFILE *sql;
int phase;
int proceed;

(void) clm_settrap(true);
(void) jrl_journalling(true,APJRL,"<%s-%s> starting",appname,clm_getvers());
(void) clm_setjournal(EVJRL,d_log,"events");
(void) clm_setjournal(EVCKR,d_log,"checker");
(void) clm_setjournal(EVTRU,d_log,"trusting");
empty=(const char *)0;
wrk=(WRKTYP *)0;
sql=(SQLFILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase)  {
    case   0	:	/*loop init			*/
      wrk=(WRKTYP *)calloc(1,sizeof(WRKTYP));
      wrk->nbrchkr=nbrchkr;
      wrk->stats=sta_mkstats();
      wrk->chkrlst=(pid_t *)calloc(wrk->nbrchkr,sizeof(pid_t));
      if (isdbdaemonup()==false) {
        (void) clm_alert(0,"%s Unable to find data-base server up",appname);
	phase=999;	/*Trouble, Trouble		*/
        }
      break;
    case  1	:	/*opening data-base	*/
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
        (void) clm_alert(0,"%s:%s Unable to open database",appname,"cycle,");
	phase=999;	/*Trouble, Trouble		*/
	}
      break;
    case   2	:	/*loading the config		*/
      if ((wrk->cnf=cnf_chkconfig(sql,wrk->cnf))==(CNFTYP *)0) {
        (void) clm_alert(0,"%s Unable to load <%s> configuration (exiting)",
			    appname,appname);
	phase=999;
	}
      break;
    case   3	:	/*journalling new config	*/
      if (wrk->cnf->updated==true) 
	(void) cnf_showcnf(wrk->cnf);
      break;
    case   4	:	/*checking if config update	*/
      if (wrk->cnf->updated==true) {
        BININF **binds;

        binds=cnf_getbinds(wrk->cnf);
	wrk->lst=pro_getlst(wrk->lst,pro_smtp,binds);
        binds=dev_freebindinf(binds);
        }
      break;
    case   5	:	/*setting the msg exchange boxe	*/
      if (wrk->cnf->updated==true) {
        (void) dev_mulclose(wrk->lst);
        (void) stopsubproc(wrk);
	if (wrk->mxc!=(MXCTYP *)0)
          (void) ipc_subscribe(wrk->mxc->bptr,(u_long)0,false);
        wrk->mxc=ipc_dropmsgxchg(wrk->mxc);
        if (prepmsgbox(wrk,MXBOX,nbrchkr+nbrmnt+BOXSPARE)==false) {
	  (void) clm_alert(0,"%s Unable to fully prepare shared memory",
			      appname);
	  phase=999;
	  }
	}
      break;
    case   6	:	/*starting needed sub process	*/
      (void) getmentor(wrk);
      (void) checkers(wrk);
      break;
    case   7	:	/*start/check transitr process	*/
      (void) transitr(true);
      break;
    case   8	:	/*check cron action		*/
      (void) docron(sql);
      break;
    case   9	:	/*checking for lost subscriber	*/
      (void) prc_nozombie();
      (void) ipc_cleanbox(wrk->mxc->bptr,10);
      break;
    case  10	:	/*starting/checking proto handler*/
      if (pro_listen(wrk->cnf->updated,wrk->mxc,wrk->lst,pro_doproto)==false)
	phase=999;
      wrk->cnf->updated=false;
      break;
    case  11	:	/*checking for incoming statistic*/
      (void) collectmsg(wrk);
      break;
    case  12	:	/*compiling stats		*/
      (void) sta_movestats(wrk->stats);
      (void) sta_dumpstats(wrk->stats);
      break;
    case  13	:	/*lets wait for event		*/
      phase=1;		/*go back to redoot it again	*/
      (void) sleep(2);
      if (sigterm==true)
        phase=999;
      break;
    default	:
      sql=sql_close(sql);
      (void) dev_mulclose(wrk->lst);
      (void) stopsubproc(wrk);
      (void) transitr(false);
      if (wrk->mxc!=(MXCTYP *)0)
        (void) ipc_subscribe(wrk->mxc->bptr,(u_long)0,false);
      wrk->mxc=ipc_dropmsgxchg(wrk->mxc);
      wrk->lst=pro_freelst(wrk->lst);
      wrk->cnf=cnf_freeconfig(wrk->cnf);
      wrk->stats=sta_freestats(wrk->stats);
      (void) free(wrk->chkrlst);
      (void) free(wrk);
      proceed=false;
      break;
    }
  phase++;
  }
(void) jrl_journalling(true,APJRL,"<%s-%s> stopped",appname,clm_getvers());
(void) clm_settrap(false);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Program Entry.				*/
/*						*/
/************************************************/
int main(int argc,char *argv[])

{
GACTYP *gat;
char *testrqst;
u_int nbrchkr;
char c;

(void) clm_trapsegv(true);
(void) prc_allow_core_dump();
(void) openlog(appname,LOG_PID,0);
base=strdup(base);
gat=(GACTYP *)0;
testrqst=strdup("");
nbrchkr=3;	/*3 checker by default		*/
while ((c=getopt(argc,argv,"b:c:d:fhot:v"))!=EOF) {
  switch(c) {
    case   'b'  :
      (void) free(base);
      base=clm_mkbase(optarg);
      break;
    case   'c'  :
      nbrchkr=atoi(optarg);
      if (nbrchkr==0)
  	nbrchkr=1;	/*default checker process number	*/
      break;
    case   'd'  :
      debug=atoi(optarg);
      break;
    case   'f'  :
      clm_foreground=true;
      break;
    case   'o'  :
      eml_orgn=true;
      break;
    case   't'  :
      clm_foreground=true;
      (void) free(testrqst);
      testrqst=strdup(optarg);
      break;
    case   'v'  :
      verbose=true;
      break;
    case   'h'  :
    default	:
      (void) usage();
      break;
    }
  }
/*application garbage collector		*/
(void) gac_setcurbase(&debug,&verbose,base);
gat=gac_open(2,appname,"");
(void) gac_start(gat);
(void) clm_getmillisec();
(void) gac_setlogdir(clm_appdir(d_log));
(void) clm_chgid(appname,d_sbin);
(void) clm_setjournal(APJRL,d_log,appname);
(void) reportvers();
(void) prc_divedivedive(clm_foreground);
(void) prc_preptitle(argc,argv,environ);
(void) clm_loadsysconf(appname,true);
if (lck_locking(appname,LCK_LOCK,5)==true) {
  if (strlen(testrqst)>0)
    (void) tst_module(testrqst);
  else {
    (void) cycle(nbrchkr,2);
    }
  (void) lck_locking(appname,LCK_UNLOCK,1);
  }
(void) jrl_closejournals();
(void) clm_loadsysconf(appname,false);
(void) prc_cleantitle();
(void) clm_trashmem(2,gat);
(void) free(base);
(void) free(testrqst);
(void) closelog();
(void) clm_trapsegv(false);
return 0;
}
