/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	CLMTRST					*/
/*	Its purpose is to detect to be banned or*/
/*	to be trusted remote IP.		*/
/*	decision is done according data		*/
/*	collected within emlogs table.		*/
/*						*/
/************************************************/
#include	<errno.h>
#include	<string.h>
#include	<stdio.h>
#include	<syslog.h>

#include	"lowtyp.h"
#include	"utlgac.h"
#include	"utlvec.h"
#include	"subprc.h"
#include	"unijrl.h"
#include	"unilck.h"
#include	"gesdbc.h"
#include	"gesclm.h"
#include	"gestru.h"
#include	"lvlcnf.h"

#define	CLMTRST	"clmtrst"
#define	FBAN	0.95
#define	FTRUST	0.05
#define	FSAMPLE 200
/*

*/
/************************************************/
/*						*/
/*	Procedure to display the program usage	*/
/*						*/
/************************************************/
static void usage()

{
(void) fprintf(stderr,"%s Version %s\n",CLMTRST,clm_getvers());
(void) fprintf(stderr,"usage:  %s "
		      "[-b base] "
		      "[-d debug] "
		      "[-T coef] "
		      "[-B coef] "
		      "[-S integer] "
		      "[-h ] "
		      "[-v ] "
		      "[recipient [recipient]....]\n",CLMTRST);
(void) fprintf(stderr,"\t\t-b base      : application working directory\n");
(void) fprintf(stderr,"\t\t-d debug     : debug level\n");
(void) fprintf(stderr,"\t\t-B float     : BELOW this trust level, remote ip is set to be banned,\n");
(void) fprintf(stderr,"\t\t               default value is %3.2f.\n",FBAN);
(void) fprintf(stderr,"\t\t-T float     : ABOVE this trust level, remote ip is set to be trusted,\n");
(void) fprintf(stderr,"\t\t               default value is %3.2f.\n",FTRUST);
(void) fprintf(stderr,"\t\t-S integer   : Minimal number of email received from remote IP to\n");
(void) fprintf(stderr,"\t\t               check for trust level, default value is %03d emails.\n",FSAMPLE);
(void) fprintf(stderr,"\t\t-h           : display usage\n");
(void) fprintf(stderr,"\t\t-v           : verbose debug\n");
(void) exit(-1);
}
/*
^L
*/
/************************************************/
/*						*/
/*	procedure to update a list of 		*/
/*	banned/trusted site.			*/
/*						*/
/************************************************/
static int dbupdate(SQLFILE *sql,float level,int status)

{
#define	OPDB	"dbtrust.c:dbupdate,"

int numbers;
const char *strstat;
char **lists;

numbers=0;
strstat="UKN";
lists=(char **)0;
switch (status) {
  case  TRSTED	:
    lists=dbc_extractrate(sql,level,true);
    strstat=STRTRS;
    break;
  case  RBANNED	:
    lists=dbc_extractrate(sql,level,false);
    strstat=STRBAN;
    break;
  default	:
    (void) clm_alert(0,"%s Unexpected status value='%d' (bug?)",OPDB,status);
    break;
  }
if (lists!=(char **)0) {
  time_t curtime;
  char *origine;

  curtime=time((time_t)0);
  (void) asprintf(&origine,"%s %s %s;",clm_ascsysstamp(curtime),STRSTA,strstat);
  numbers=cnf_updbantrust(sql,lists,status,origine);
  (void) free(origine);
  lists=(char **)vec_freelstlst((void **)lists,VECFREE);
  }
return numbers;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Detect remote IP to be banned or trusted*/
/*						*/
/************************************************/
int cycle(float ftrust,float fban,int fsample)

{
int done;
char **lists;
const char *empty;
SQLFILE *sql;
int phase;
int proceed;

done=false;
lists=(char **)0;
empty=(const char *)0;
sql=(SQLFILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening SQL data-base	*/
      (void) clm_setjournal(APJRL,d_log,appname);
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
	(void) fprintf(stderr,"Unable to open data-base (config default?)\n");
	proceed=false;	/*big big trouble	*/
	}
      break;
    case 1	:	/*computing trust level	*/
      if (dbc_buildbantrust(sql,fsample)==false) {
	(void) jrl_journalling(true,APJRL,"%s is unable to build banned/trusted IP list (db problem?)");
	phase=999;	/*trouble trouble?	*/
        }
      break;
    case 2	:	/*getting trusted rmtip	*/
      (void) dbupdate(sql,ftrust,TRSTED);
      break;
    case 3	:	/*getting banned rmtip	*/
      (void) dbupdate(sql,fban,RBANNED);
      break;
    case 4	:	/*aggregating IP	*/
      (void) dbc_agregate(sql,AF_INET,RBANNED);
      break;
    default	:	/*SAFE Guard		*/
      sql=sql_close(sql);
      (void) jrl_closejournals();
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Program Entry.				*/
/*						*/
/************************************************/
int main(int argc,char *argv[])

{
GACTYP *gat;
int status;
float fban;
float ftrust;
int fsample;
int queuemode;
char *myemail;
char **recipients;
char c;


clm_foreground=true;
gat=(GACTYP *)0;
status=0;
fban=FBAN;
ftrust=FTRUST;
fsample=FSAMPLE;
queuemode=false;
myemail=(char *)0;
recipients=(char **)0;
(void) clm_trapsegv(true);
(void) openlog(CLMTRST,LOG_PID,0);
base=strdup(base);
while ((c=getopt(argc,argv,"B:S:T:b:d:hv"))!=EOF) {
  switch(c) {
    case   'B'  :
      fban=atof(optarg);
      break;
    case   'T'  :
      ftrust=atof(optarg);
      break;
    case   'S'  :
      fsample=atoi(optarg);
      break;
    case   'b'  :
      (void) free(base);
      base=clm_mkbase(optarg);
      break;
    case   'd'  :
      debug=atoi(optarg);
      break;
    case   'v'  :
      verbose=true;
      break;
    case   'h'  :
    default	:
      (void) usage();
      break;
    }
  }
gat=gac_open(2,appname,"");
(void) gac_start(gat);
(void) gac_setcurbase(&debug,&verbose,base);
(void) prc_preptitle(argc,argv,environ);
(void) clm_loadsysconf(CLEMENT,true);
if (lck_locking(CLMTRST,LCK_LOCK,5)==true) {
  status=cycle(ftrust,fban,fsample);
  (void) lck_locking(CLMTRST,LCK_UNLOCK,1);
  }
(void) prc_cleantitle();
(void) clm_trashmem(2,gat);
(void) free(base);
(void) closelog();
(void) clm_trapsegv(false);
return status;
}
