/************************************************/
/*						*/
/*      Copyright:				*/
/*	 Jean-Marc Pigeon <jmp@safe.ca>	 2010	*/
/*						*/
/************************************************/
/* This program is free software; you can 	*/
/* redistribute it and/or modify it under the 	*/
/* terms of the GNU General Public License as	*/
/* published by the Free Software Foundation	*/
/* version 2 of the License			*/
/*						*/
/* This program is distributed in the hope that */
/* it will be useful, but WITHOUT ANY WARRANTY; */
/* without even the implied warranty of		*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR	*/
/* PURPOSE.  See the GNU General Public License	*/
/* for more details.				*/
/*						*/
/* You should have received a copy of the GNU	*/
/* General Public License along with this 	*/
/* program; if not, write to the Free Software	*/
/* Foundation, Inc., 51 Franklin Street,	*/
/* Fifth Floor, Boston, MA  02110-1301, USA.	*/
/************************************************/
/*						*/
/*	PIBSMTP:				*/
/*	Program purpose is to update pop/imap	*/
/*	access with environment data given by	*/
/*	dovecot					*/
/*						*/
/*	List will be used to make a list of	*/
/*	IP used by remote user to read their	*/
/*	mail, this list will be used to allow	*/
/*	relaying.				*/
/*						*/
/************************************************/
#include	<errno.h>
#include	<unistd.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<malloc.h>
#include	<string.h>
#include	<getopt.h>
#
#include	"subclm.h"
#include	"gespbs.h"


/*

*/
/************************************************/
/*						*/
/*	Program usage help			*/
/*						*/
/************************************************/
static void usage()

{
(void) fprintf(stderr,"%s Version %s\n",PIBSMTP,clm_getvers());
(void) fprintf(stderr,"usage: %s [options] [program_to_call]"
		      "[-h "
		      "[-b base] "
		      "[-d debug] "
		      "[-v ] "
		      "\n",PIBSMTP);
(void) fprintf(stderr,"\t\t-d program   : program to call at the end\n");
(void) fprintf(stderr,"\t\t-h           : this help\n");
(void) fprintf(stderr,"\t\t-b base      : application working directory\n");
(void) fprintf(stderr,"\t\t-d debug     : debug level\n");
(void) fprintf(stderr,"\t\t-v           : verbose debug\n");
}
/*
^L
*/
/************************************************/
/*						*/
/*	Program Entry.				*/
/*						*/
/************************************************/
int main(int argc,char *argv[])

{
char c;
char *nextcall;
char **locargv;
int locargc;

base=strdup(base);
nextcall=(char *)0;
locargv=(char **)calloc(argc,sizeof(char *));
locargc=0;
while ((c=getopt(argc,argv,"+b:d:hv"))!=EOF) {
  switch(c) {
    case   'b'  :
      (void) free(base);
      base=clm_mkbase(optarg);
      break;
    case   'd'  :
      debug=atoi(optarg);
      break;
    case   'h'  :
    default	:
      (void) usage();
      break;
    case   'v'  :
      verbose=true;
      break;
    }
  }
if (argc>optind) {
  int i;

  nextcall=argv[optind];
  optind++;
  for (i=optind;i<argc;i++) {
    locargv[locargc]=argv[i];
    locargc++;
    }
  }
(void) clm_chgid(PIBSMTP,d_bin);
(void) pbs_addentry(getenv("USER"),getenv("IP"));
/*let take back our previous ID		*/
if (setegid(getgid())<0) {
  (void) clm_alert(0,"%s Unable to set the Egid to '%d'",getgid());
  }
if (seteuid(getuid())<0) {
  (void) clm_alert(0,"%s Unable to set the Euid to '%d'",getuid());
  }
if (nextcall!=(char *)0) {
  (void) execve(nextcall,locargv,environ);
  }
(void) free(base);
return 0;
}
