/************************************************/
/*						*/
/*	Procedure implementation to manage	*/
/*	MySQL database.				*/
/*						*/
/************************************************/
#include	<unistd.h>
#include	<stdlib.h>
#include	<fcntl.h>
#include	<stdio.h>
#include	<syslog.h>
#include	<string.h>
#include	<time.h>
#include	<errno.h>
#include	<pwd.h>
#include	<sys/types.h>
#include	<sys/stat.h>

#ifdef	with_mysql	/*ONLY with MYSQL	*/
#include	<mysql/mysql.h>
#
#include	"lowtyp.h"
#include	"utlprc.h"
#include	"subclm.h"
#include	"basmys.h"

/*the SQL default port access			*/
static char mysqlport[10]="";
/*

*/
/************************************************/
/*						*/
/*	procedure to find the field number	*/
/*	related to a field name.		*/
/*						*/
/************************************************/
static int mysqlfieldnumber(MYSQL_RES *result,char *fieldname)

{
int position;
uint numfields;
uint i;
MYSQL_FIELD *fields;

position=-1;
numfields=mysql_num_fields(result);
fields=mysql_fetch_fields(result);
for (i=0;i<numfields;i++) {
  if (strcmp(fields[i].name,fieldname)==0) {
    position=i;
    break;
    }
  }
return position;
}
/*

*/
/************************************************/
/*						*/
/*	Procure to extact the ASCII value of	*/
/*	a specific filed within a specific row.	*/
/*						*/
/************************************************/
static char *mysqlgetvalue(MYSQL_RES *result,uint tuple,uint position)

{
MYSQL_ROW row;
char *valeur;

valeur=(char *)0;
(void) mysql_data_seek(result,tuple);
if ((row=mysql_fetch_row(result))!=(MYSQL_ROW)0) {
  valeur=row[position];
  }
return valeur;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a unix time in acsii*/
/*	format (data-base compatible).		*/
/*						*/
/************************************************/
const char *mys_fromunixtime(time_t timestamp)

{
/*Note: 
 *syntax MYSQL ("FROM_UNIXTIME(%ld)",timestamp)
 */
#define TSTAMP   "%Y-%m-%d %H:%M:%S"

static char unixdate[50];

struct tm *tminfo;

tminfo=localtime(&timestamp);
(void) strftime(unixdate,sizeof(unixdate),TSTAMP,tminfo);
return unixdate;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a unix time_t	*/
/*	from a ascii string (data-base 		*/
/*	compatible).				*/
/*						*/
/************************************************/
time_t mys_tounixtime(char *date)

{
#define	DBDTOUNIX	"%Y-%m-%d %H:%M:%S"
#define	STAMPEON	"0000-00-00 00:00:00"

time_t datetime;

datetime=(time_t)0;
if ((date!=(char *)0)&&(strcmp(date,STAMPEON)!=0)) {
  char *ptr;
  struct tm tm;

  if ((ptr=strchr(date,'.'))!=(char *)0) {
    *ptr='\000';
    }
  (void) memset(&tm,'\000',sizeof(struct tm));
  tm.tm_isdst=-1;
  if (strptime(date,DBDTOUNIX,&tm)!=(char *)0) {
    datetime=mktime(&tm);
    }
  else {
    (void) clm_alert(0,"basmys.c,mys_tounixtime: Unable to convert <%s> "
		       "to time_t",date);
    }
  }
return datetime;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to make a connection to	*/
/*	MySQL daemon.				*/
/*						*/
/************************************************/
MYSFILE *mys_open(const char *host,const char *sqlport,const char *dbname)

{
#define OPMY	"basmys.c:mys_open,"
MYSQL *mysql;
struct passwd *pw;

mysql=mysql_init((MYSQL *)0);
if ((pw=getpwuid(geteuid()))==(struct passwd *)0)
  (void) clm_core_dump("%s is unable to find name for uid '%s'",OPMY,geteuid());
if (mysql_real_connect(mysql,host,pw->pw_name,(char *)0,dbname,0,(char *)0,CLIENT_FOUND_ROWS)==(MYSQL *)0){
  (void) clm_alert(0,"%s Connection to database '%s' failed, cause '%s'",
		     OPMY,dbname,mysql_error(mysql));
  (void) free(mysql);
  mysql=(MYSQL *)0;
  }
return (MYSFILE *)mysql;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to disconnect  from a 	*/
/*	MySQL daemon.				*/
/*						*/
/************************************************/
MYSFILE *mys_close(MYSFILE *ms)

{
if (ms!=(MYSFILE *)0) {
  (void) mysql_close((MYSQL *)ms);
  ms=(MYSFILE *)0;
  }
else {
  (void) clm_alert(0,"basmys.c,mys_close: database access pointer empty (bug?)");
  }
return ms;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the default MySQL	*/
/*	access port number.			*/
/*						*/
/************************************************/
const char *mys_getdfltport()

{
if (strlen(mysqlport)==0) {
  (void) snprintf(mysqlport,sizeof(mysqlport),"%d",MYSQL_PORT);
  }
return mysqlport;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do a data-base command	*/
/*	"insert","update", return 0 if success	*/
/*	-1 otherwise.				*/
/*						*/
/************************************************/
int mys_action(MYSFILE *ms,char *command)

{
int status;

status=0;
if ((status=mysql_query((MYSQL *)ms,command))!=0) {
  (void) clm_alert(0,"basmys.c,action, pid='%05d' Unable to carry cmd=<%s>, error=<%s>",
		      getpid(),command,mysql_error((MYSQL *)ms));
  status=-1;
  }
else {
  status=(int)mysql_affected_rows((MYSQL *)ms);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract data from the	*/
/*	data-base				*/
/*	Return MYSRESULT status pointer (can	*/
/*	be null if not successful).		*/
/*						*/
/************************************************/
MYSRESULT *mys_tupple(MYSFILE *ms,char *command)

{
#define	OPTU	"basmys.c:mys_tupple"

MYSQL_RES *mysstatut;

mysstatut=(MYSQL_RES *)0;
if (mysql_query((MYSQL *)ms,command)==0) {
  if ((mysstatut=mysql_store_result((MYSQL *)ms))==(MYSQL_RES *)0)
    (void) clm_alert(3,"%s result empty for cmd=<%s>",OPTU,command);
  }
else {
  (void) clm_alert(0,"%s pid='%05d' Unable to carry cmd=<%s>, error=<%s>",
		      OPTU,getpid(),command,mysql_error((MYSQL *)ms));
  }
return (MYSRESULT *)mysstatut;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the number of tupple*/
/*	related to a previous search.		*/
/*						*/
/************************************************/
int mys_nbrtupple(MYSRESULT *mysresult)

{
return mysql_num_rows((MYSQL_RES *)mysresult);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure clean the result area		*/
/*	previously allocated via pos_tupple	*/
/*	related to a previous search.		*/
/*						*/
/************************************************/
MYSRESULT *mys_dropresult(MYSRESULT *mysresult)

{
(void) mysql_free_result((MYSQL_RES *)mysresult);
return (MYSRESULT *)0;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to retrieve specific field in	*/
/*	specific row, previously located via	*/
/*	pos_tupple.				*/
/*						*/
/************************************************/
char *mys_getvalue(MYSRESULT *mysresult,int row,char *fieldname)

{
int position;

if ((position=mysqlfieldnumber((MYSQL_RES *)mysresult,fieldname))<0) {
  (void) clm_alert(0,"basmys.c:mys_getvalue, field '%s' unknown",fieldname);
  return (char *)0;
  }
return mysqlgetvalue((MYSQL_RES *)mysresult,row,position);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to retrieve specific column in*/
/*	a specific row, previously located via	*/
/*	pos_tupple.				*/
/*						*/
/************************************************/
char *mys_getfield(MYSRESULT *mysresult,int row,int column)

{
return mysqlgetvalue((MYSQL_RES *)mysresult,row,column);
}
#ifdef NEEDED
Removed 2011 Mars 6
/*

*/
/************************************************/
/*						*/
/*	Procedure to set/unset and advisory lock*/
/*	table MUST be a number definition	*/
/*						*/
/************************************************/
static int mslockadvisory(MYSFILE *ms,char *table,int lock)

{
int status;
const char *stract;
MYSRESULT *result;
char command[200];

status=-1;
stract="SELECT GET_LOCK(%s,20)";
if (lock==false)
  stract="SELECT RELEASE_LOCK(%s)";
(void) snprintf(command,sizeof(command),stract,table);
if ((result=mys_tupple(ms,command))!=(MYSRESULT *)0) {
  status=mys_nbrtupple(result);
  result=mys_dropresult(result);
  }
return status;
}
#endif
/*

*/
/************************************************/
/*						*/
/*	Procedure to request a lock to access 	*/
/*	set of table in the MySQL data-base.	*/
/*	Return 0, if everything fine, -1	*/
/*	otherwise.				*/
/*						*/
/************************************************/
int mys_getlock(MYSFILE *ms,LCKTYP **lcklst)

{
int status;
int phase;
int proceed;

status=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch(phase) {
    case 0	:	/*Lets BEGIN		*/
      if ((status=mys_action(ms,"BEGIN"))<0)
	phase=999;	/*trouble trouble	*/	
      break;
    case 1	:	/*lock to establish	*/
      if (lcklst!=(LCKTYP **)0) {
	int i;
	char *cmd;

	cmd=strdup("");
	for (i=0;lcklst[i]!=(LCKTYP *)0;i++) {
	  char *mode;
	  char *addon;
	  char sep;

	  mode="READ";
	  sep=',';
	  if (lcklst[i]->lock==dbd_lockwrite)
	    mode="WRITE";
	  if (lcklst[i+1]==(LCKTYP *)0)
	    sep=' ';
	  (void) asprintf(&addon,"%s%s %s%c",cmd,lcklst[i]->tblname,mode,sep);
	  (void) free(cmd);
	  cmd=addon;
	  }
	if (strlen(cmd)>0) {
	  char *fullcmd;
	  
	  (void) asprintf(&fullcmd,"LOCK TABLE %s",cmd);
	  if ((status=mys_action(ms,fullcmd))<0)
	    phase=999;	/*trouble trouble	*/
	  (void) free(fullcmd);
	  }
	(void) free(cmd);
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to request a lock action 	*/
/*	in case we update table component.	*/
/*						*/
/************************************************/
int mys_lock(MYSFILE *ms)

{
return mys_action(ms,"BEGIN");
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to release pending lock 	*/
/*	set to restrict access on MySQL		*/
/*	data0base tables.			*/
/*	Return 0, if everything fine, -1	*/
/*	otherwise.				*/
/*						*/
/************************************************/
int mys_unlock(MYSFILE *ms,int commit)

{
int status;
char *cmd;

cmd="COMMIT";
if (commit==false)
  cmd="ROLLBACK";
if ((status=mys_action(ms,cmd))==0)
  status=mys_action(ms,"UNLOCK TABLES");
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to make sure the MySQL	*/
/*	daemon is up and running.		*/
/*						*/
/************************************************/
int mys_isdaemonup(const char *host,const char *sqlport,const char *dbname,int tentative)

{
int itisok;
char *vd;
struct passwd *pw;
int i;

itisok=false;
vd=(char *)0;
if ((pw=getpwuid(geteuid()))==(struct passwd *)0)
  (void) clm_core_dump("basmys.c,mys_startdaemon: is unable to find name "
		       "for uid '%s'",geteuid());
for (i=0;i<tentative;i++) {
  MYSQL *sql;

  sql=mysql_init((MYSQL *)0);
  if (mysql_real_connect(sql,vd,pw->pw_name,vd,dbname,0,(char *)0,0)!=(MYSQL *)0){
    (void) clm_alert(1,"MySQL daemon confirmed up and running");
    itisok=true;
    }
  (void) mysql_close(sql);
  if (itisok==true)
    break;
  (void) sleep(1);
  }
if (itisok!=true) {
  (void) clm_alert(0,"MySQL daemon NOT confirmed up and running");
  }
return itisok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract the current MySQl	*/
/*	daemon pid.				*/
/*						*/
/************************************************/
pid_t mys_getdaemonpid(char *basedir)

{
pid_t daemonpid;
char *ptrloc;
FILE *fichier;
char hostname[100];
char command[150];

daemonpid=(pid_t)-1;
(void) memset(hostname,'\000',sizeof(hostname));
(void) gethostname(hostname,sizeof(hostname)-1);
if ((ptrloc=strchr(hostname,'.'))!=(char *)0)
  *ptrloc='\000';
(void) snprintf(command,sizeof(command),"cat %s/dbm/%s.pid",basedir,hostname);
if ((fichier=popen(command,"r"))!=(FILE *)0) {
  char line[200];
                                                                                
  if (fgets(line,sizeof(line)-1,fichier)!=(char *)0) {
    daemonpid=(pid_t)atoi(line);
    (void) pclose(fichier);
    }
  }
return daemonpid;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get the current sequence	*/
/*	number					*/
/*						*/
/************************************************/
u_vlong mys_getseq(MYSFILE *ms,char *sequence)

{
u_vlong seqnum;
MYSRESULT *result;
char cmd[200];

seqnum=(u_vlong)0;
(void) snprintf(cmd,sizeof(cmd),"SELECT num FROM %s",sequence);
if ((result=mys_tupple(ms,cmd))!=(MYSRESULT *)0) {
  seqnum=atoll(mys_getvalue(result,0,"num"));
  (void) mysql_free_result(result);
  }
return seqnum;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to increment the sequence	*/
/*	number					*/
/*						*/
/************************************************/
u_vlong mys_incseq(MYSFILE *ms,char *sequence)

{
u_vlong seqnum;
char cmd[200];

seqnum=(u_vlong)0;
(void) snprintf(cmd,sizeof(cmd),"UPDATE %s SET num=num+1",sequence);
if (mys_action(ms,cmd)==0) 
  seqnum=mys_getseq(ms,sequence);
return seqnum;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to detect and 'clean' any	*/
/*	single quote within a string, this is	*/
/*	mandatory for string variable in the	*/
/*	data-base.				*/
/*	Single quote and double quote are	*/
/*	translated to '\'' and '\"'.		*/
/*						*/
/************************************************/
char *mys_cleanquote(MYSFILE *ms,char *dstr)

{
char *cleanstr;
int taille;
char *ptr;

taille=3;
cleanstr=calloc(3,sizeof(char));
cleanstr[0]='\'';
if (dstr!=(char *)0) {
  char *nptr;

  ptr=dstr;
  while (((nptr=strchr(ptr,'\''))!=(char *)0)||((nptr=strchr(ptr,'\"'))!=(char *)0)) {
    char lit[3];

    (void) strcpy(lit,"\\'");
    if (*nptr=='"')
      lit[1]='"';
    *nptr='\000';	/*to segment string	*/
    taille+=strlen(ptr)+2;
    cleanstr=(char *)realloc(cleanstr,taille);
    (void) strcat(cleanstr,ptr);
    (void) strcat(cleanstr,lit);
    *nptr=lit[1];	/*to glue back string	*/
    ptr=nptr+1;
    }
  if (*ptr!='\000') {
    taille+=strlen(ptr);
    cleanstr=(char *)realloc(cleanstr,taille);
    (void) strcat(cleanstr,ptr);
    }
  }
(void) strcat(cleanstr,"'");
return cleanstr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to format an ipnum to be	*/
/*	stored within an 'varbin' MySQL record	*/
/*						*/
/************************************************/
char *mys_toipbin(MYSFILE *ms,char *ipnum)

{
char *ipbin;
const char *strloc;

strloc="inet_aton";
if (strchr(ipnum,':')!=(char *)0) {
  char *ptr;

  ptr=ipnum;
  strloc="unhex";
  while ((ptr=strchr(ptr,':'))!=(char *)0) {
    (void) memmove(ptr,ptr+1,strlen(ptr+1)+1);
    ptr++;
    }
  }
(void) asprintf(&ipbin,"%s('%s')",strloc,ipnum);
return ipbin;
}
#endif	/*DB_MYSQL ONLY with MYSQL		*/
