/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	COMEML:					*/
/*	Define all routine to manage email 	*/
/*	communication exchange.			*/
/*						*/
/************************************************/
#include	<string.h>
#include	<errno.h>
#
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"subprc.h"
#include	"basafn.h"
#include	"unijrl.h"
#include	"comeml.h"

#define MXTIME	300	/*wait for a 300 sec max*/
/*

*/
/************************************************/
/*						*/
/*	Procedure to report an unexpected	*/
/*	contact trouble.			*/
/*						*/
/************************************************/
static void rpttrouble(CTTYP *contact,char *trouble)

{
if ((contact!=(CTTYP *)0)&&(trouble!=(char *)0)) {
  (void) clm_alert(0,"%s",trouble);
  contact->info=clm_freestr(contact->info);
  contact->info=trouble;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to display connection status	*/
/*	if doing command 'ps ax' from console	*/
/*						*/
/************************************************/
static void emlsettitle(CTTYP *contact)

{
(void) prc_settitle("%s: Reading on %s:%s since %s, from '%s/[%s]'",
                    appname,contact->locip,contact->servport,contact->attime,
		    contact->peername,contact->peerip);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a contact structure*/
/*						*/
/************************************************/
static CTTYP *setcontact(TLSFILE *tls,PROTYP proto,char *port)

{
#define	MXWAIT	300	/*contact maximum idle time	*/
#define	OPSE	"comeml.c:setcontact,"

CTTYP *contact;
char *ip;

contact=(CTTYP *)calloc(1,sizeof(CTTYP));
contact->tls=tls;
contact->locname=tls_getaddrname(tls,pr_locname,"NO_LOCAL_HOST_NAME");
contact->servport=strdup(port);
contact->peername=tls_getaddrname(tls,pr_remname,NORVERS);
ip=tls_getaddrname(tls,pr_locip,UKNIP);
contact->locip=afn_formatip(ip);
(void) free(ip);
ip=tls_getaddrname(tls,pr_remip,UKNIP);
contact->peerip=afn_formatip(ip);
(void) free(ip);
contact->proto=proto;
contact->sequence=0;
contact->lastdata=time((time_t)0);
contact->start=contact->lastdata;
contact->stop=contact->lastdata;
contact->attime=strdup(clm_ascsystime(contact->lastdata));
contact->mtaille=(u_long)0;
contact->maxwait=MXWAIT;
contact->founddata=false;
contact->session=clm_uniquename(contact->sequence);
contact->msession=strdup(contact->session);
if (debug>0) {
  char *dbgname;
  FILE *dbgfile;

  dbgname=clm_uniquename((unsigned int)0);
  contact->bckstdout=stdout;
  (void) asprintf(&(contact->dbgname),"%s/%s.dbgs",clm_appdir(d_dbgs),dbgname);
  if ((dbgfile=fopen(contact->dbgname,"w"))==(FILE *)0)
    (void) clm_core_dump("%s Unable to fopen file <%s> (error=<%s>)",
		          OPSE,contact->dbgname,strerror(errno));
  stdout=dbgfile;	/*now STDOUT is the file	*/
  if (debug>3) {
    char *strloc;

    (void) tls_trace(tls,true);
    (void) asprintf(&strloc,"Opening session '%s'\n",contact->session);
    (void) tls_traceout(contact->tls,"#--",strloc,strlen(strloc));
    (void) free(strloc);
    }
  dbgname=clm_freestr(dbgname);
  }
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a 	*/
/*	contact structure.			*/
/*						*/
/************************************************/
static CTTYP *freecontact(CTTYP *contact)

{
contact->authname=clm_freestr(contact->authname);
contact->attime=clm_freestr(contact->attime);
contact->info=clm_freestr(contact->info);
contact->cipherid=clm_freestr(contact->cipherid);
contact->peerip=clm_freestr(contact->peerip);
contact->servport=clm_freestr(contact->servport);
contact->peername=clm_freestr(contact->peername);
contact->heloname=clm_freestr(contact->heloname);
contact->locip=clm_freestr(contact->locip);
contact->locname=clm_freestr(contact->locname);
contact->session=clm_freestr(contact->session);
contact->msession=clm_freestr(contact->msession);
(void) free(contact);
contact=(CTTYP *)0;
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a string ended with	*/
/*	CRLF on the TCP line			*/
/*						*/
/************************************************/
int printfcrlf(CTTYP *contact,const char *fmt,va_list args)

{
va_list ap;
int written;
int taille;
char stdq[500];

va_copy(ap,args);
if ((taille=vsnprintf(stdq,sizeof(stdq),fmt,ap))>(sizeof(stdq)-5)) {
  /*case of a very long formated line		*/
  char *local;		

  taille=vasprintf(&local,fmt,args);
  local=(char *)realloc(local,taille+10);
  (void) strcpy(local+taille,CRLF);
  taille+=strlen(CRLF);
  written=tls_write(contact->tls,local,taille);
  (void) free(local);
  }
else {
  /*usual length  write				*/
  (void) strcpy(stdq+taille,CRLF);
  taille+=strlen(CRLF);
  written=tls_write(contact->tls,stdq,taille);
  }
va_end(ap);
return written;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send an "HELO" message	*/
/*	if EHLO message is not accepted.	*/
/*						*/
/************************************************/
static int dohelo(CTTYP *contact,char *hostname)

{
int done;
char *buffer;

done=false;
(void) eml_printfcrlf(contact,HELO,contact->locname);
if ((buffer=tls_waitforaction(contact->tls,contact->maxwait))!=(char *)0) {
  int numret;

  (void) clm_cleanstring(buffer);
  if (sscanf(buffer,"%d",&numret)==1) {
    done=true;
    if (numret!=CMDOK) {
      (void) jrl_journalling(false,EVTRT,"Unable to carry HELO with '%s' returned status=<%s>",
		         		  hostname,buffer);
      done=false;
      }
    }
  (void) free(buffer);
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send an "EHLO" message	*/
/*	an collect working parameters		*/
/*						*/
/************************************************/
static int doehlo(CTTYP *contact,char *hostname)

{
#define	OPEH	"comeml.c:doehlo,"

typedef	enum	{
	h_tls,		/*STARTTLS capable	*/
	h_size,		/*SIZE capable		*/
	h_valdted,	/*remote is trusting us	*/
	h_unknown	/*unknown code		*/
	}HTYP;

static VOCTYP vochelo[]={
        {h_tls,STRTTLS,(const void *)0},
        {h_size,MSIZE,(const void *)0},
        {h_valdted,HLOTRS,(const void *)0},
        {h_unknown,(char *)0,(const void *)0}
        };

int done;
int numret;
int sofar;
char *buffer;
int phase;
int proceed;

done=true;
numret=0;
sofar=0;
buffer=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*sending EHLO		*/
      (void) eml_printfcrlf(contact,EHLO,contact->locname);
      if ((buffer=tls_waitforaction(contact->tls,contact->maxwait))==(char *)0) {
	done=false;
	phase=999;	/*trouble! no answer	*/
	}
      break;
    case 1	:	/*getting the answer	*/
      if (sscanf(buffer,"%d%n",&numret,&sofar)!=1) {
        (void) clm_alert(0,"%s (pid=%05d) Unable to carry EHLO with '%s'",
		           OPEH,getpid(),hostname);
        done=false;
        phase=999;	/*trouble! no answer	*/
	(void) free(buffer);
	}
      break;
    case 2	:	/*proper answer		*/
      phase++;
      if (numret!=CMDOK) {
	(void) free(buffer);
	done=dohelo(contact,hostname);/*backup solution	*/
	phase=999;
	}
      break;
    case 3	:	/*lets wait full answer	*/
      if ((buffer=tls_waitforaction(contact->tls,contact->maxwait))==(char *)0) {
	done=false;
	phase=999;	/*trouble! no answer	*/
	}
      break;
    case 4	:	/*lets scanreply	*/
      if (buffer!=(char *)0) {	/*always	*/
	VOCTYP *voc;

        if (buffer[sofar]=='-') { /*EHLO next	*/
	  phase--;	/*next wait for anwser	*/
	  phase--;
	  }
        (void) memmove(buffer,buffer+sofar+1,strlen(buffer+sofar+1)+1);
        (void) clm_cleanstring(buffer);
        if ((voc=clm_getvoca(vochelo,buffer))!=(VOCTYP *)0) {
	  switch (voc->code) {
	    case h_tls		:
	      contact->capable|=CPTLS;
	      break;
	    case h_size		:
	      contact->capable|=CPSIZE;
	      break;
	    case h_valdted	:
	      contact->ctxflag|=VALDTED;
	      break;
	    default	:
	      break;
	    }
	  }
	(void) free(buffer);
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert CTTYP data to an	*/
/*	advice request or a take note message.	*/
/*						*/
/************************************************/
static char *ctttomsg(register CTTYP *cta,char *msghdr)

{
#define	OPCT	"comeml.c:ctttomsg,"

char *msg;
DSPTYP *dsp;
char **dspstr;
char *strloc;
int phase;
int proceed;

msg=(char *)0;
dsp=(DSPTYP *)0;
dspstr=(char **)0;
strloc=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*building DSP		*/
      if ((dsp=eml_contacttodsp(cta))==(DSPTYP *)0) {
	(void) clm_alert(0,"%s Unable to extract dispatch session <%s> (bug?)",
			    OPCT,cta->session);
	proceed=false;	/*BIG trouble		*/
        }
      break;
    case 1	:	/*converting dispatch	*/
      if ((dspstr=eml_setdispatch(dsp))==(char **)0) {
	(void) clm_alert(0,"%s Unable to setdispatch session <%s> (bug?)",
			   OPCT,cta->session);
	phase=999;	/*BIG trouble		*/
        }
      break;
    case 2	:	/*merging dspstr	*/
      if ((strloc=clm_mergestring(dspstr,'\n'))==(char *)0) {
	(void) clm_alert(0,"%s Unable to setdispatch session <%s> (bug?)",
			   OPCT,cta->session);
	phase=999;	/*BIG trouble		*/
	}
      dspstr=(char **)vec_freelstlst((void **)dspstr,VECFREE);
      break;
    case 3	:	/*finale message	*/
      (void) asprintf(&msg,"%s %s",msghdr,strloc);
      (void) free(strloc);
      break;
    default	:	/*SAFE Guard		*/
      dsp=eml_cleandispatch(dsp);
      proceed=false;
      break;
    }
  phase++;
  }
return msg;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert CTTYP data to an	*/
/*	take note message.			*/
/*						*/
/************************************************/
char *eml_cttotknmsg(register CTTYP *cta)

{
return ctttomsg(cta,TKNOTE);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert CTTYP data to an	*/
/*	advice request message.			*/
/*						*/
/************************************************/
char *eml_cttoadvmsg(register CTTYP *cta)

{
return ctttomsg(cta,GVADV);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert an advice answer	*/
/*	message to an RCPT structure		*/
/*						*/
/************************************************/
void eml_advmsgupdrcpt(char *advice,RCPTYP **rcpts)

{
#define	OPAD	"comeml.c:eml_advmsgupdrcpt,"

char **split;
int phase;
int proceed;

split=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*data to scan ?	*/
      if ((advice==(char *)0)||((split=clm_splitstring(advice,'\n'))==(char **)0))
	phase=999;	/*Humm... strange	*/
      break;
    case 1	:	/*proper data		*/
      if (strstr(split[0],MYADV)!=split[0]) {
	(void) clm_alert(0,"%s Unexpected advice <%s>",OPAD,advice);
	phase++;	/*NO! scan		*/
	}
      break;
    case 2	:	/*scanning given advice	*/
      if (split!=(char **)0) {	/*always	*/
        char **ptr;

        ptr=split;
	while (*ptr!=(char *)0) {
	  if (strstr(*ptr,SCHK)==*ptr) {
	    int ind;
	    int status;
	    int pos;

	    ind=0;
	    status=0;
	    if (sscanf(*ptr,"%*s %d %d %n",&ind,&status,&pos)==2) {
	      rcpts[ind]->status=status;
	      if (strlen((*ptr)+pos)>0) {
	        rcpts[ind]->why=clm_freestr(rcpts[ind]->why);
		rcpts[ind]->why=strdup((*ptr)+pos);
		}
	      } 
	    } 
          ptr++;
	  }
        }
      break;
    case 3	:	/*cleanning act		*/
      split=(char **)vec_freelstlst((void **)split,VECFREE);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a string to remote	*/
/*	SMTP channel but add an CRLF at the	*/
/*	of the line.				*/
/*						*/
/************************************************/
int eml_printfcrlf(CTTYP *contact,const char *fmt,...)

{
va_list args;
int written;

va_start(args,fmt);
written=printfcrlf(contact,fmt,args);
va_end(args);
return written;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage CMD exhange with	*/
/*	a remote SMTP server.			*/
/*						*/
/************************************************/
int eml_sendcmd(CTTYP *contact,int expected,char *fmt,...)

{
#define CMDMRK	"comeml.c:eml_sendcmd, "

#define	DBGCM	6	/*routine debug level 		*/

int status;
va_list argsref;
va_list argssend;

status=ERRPROC;
va_start(argsref,fmt);
va_copy(argssend,argsref);
if (printfcrlf(contact,fmt,argssend)>0) {
  contact->info=clm_freestr(contact->info);
  status=tls_getreply(contact->tls,contact->maxwait,pro_smtp,&(contact->info));
  switch (status) {
    case ERRPROC	:
      if (sigterm==true) {
	va_list argsloc;
        char *command;

	va_copy(argsloc,argsref);
	(void) vasprintf(&command,fmt,argsloc);
        (void) clm_alert(0,"%05d %s Got termination signal to sending command <%s> to <%s>",
			       getpid(),CMDMRK,command,contact->peername);
	contact->info=clm_freestr(contact->info);
	(void) free(command);
	va_end(argsloc);
	}
      if (contact->info==(char *)0) {
	va_list argsloc;
	char *command;

	status=NOANSWR;
	va_copy(argsloc,argsref);
	(void) vasprintf(&command,fmt,argsloc);
	(void) asprintf(&(contact->info),"no answer from smtp server <%s/%s:%s> to command=<%s>",
				         contact->peername,contact->peerip,
					 contact->servport,command);
	(void) free(command);
	va_end(argsloc);
	}
      break;
    default		:
      contact->lastdata=time((time_t *)0);
      if ((status!=expected)&&(debug>=DBGCM)) {
	va_list argsloc;
	char *local;

	va_copy(argsloc,argsref);
	(void) vasprintf(&local,fmt,argsloc);
        (void) clm_alert(DBGCM,"%05d %s Unexpected answer <%s> to command <%s> from <%s>",
			       getpid(),CMDMRK,contact->info,local,contact->peername);
	(void) free(local);
	va_end(argsloc);
	}
      break;
    }
  }
va_end(argssend);
va_end(argsref);
return status; 
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to open and SMTP connection	*/
/*	according server information		*/
/*						*/
/************************************************/
CTTYP *eml_opencontact(PROTYP proto,char *hostname,char *port,int maxwait,int *socker)

{
#define	OPSR	"comeml.c:eml_opencontact,"

CTTYP *cta;
TLSFILE *tls;

cta=(CTTYP *)0;
if ((tls=tls_opentcp(hostname,port,maxwait,socker))!=(TLSFILE *)0) {
  int proceed;
  int phase;

  phase=0;
  proceed=true;
  while (proceed==true) {
    switch (phase) {
      case 0	:	/*Lets prepare contact info	*/
	if ((cta=setcontact(tls,proto,port))!=(CTTYP *)0) {
	  char *strloc;

	  (void) asprintf(&strloc,"Contact established with <%s[%s]:%s>, waiting '%d' seconds",
				   cta->peername,cta->peerip,port,maxwait);
	  (void) tls_traceout(cta->tls,"#--",strloc,strlen(strloc));
	  (void) free(strloc);
	  cta->maxwait=maxwait;
	  }
        break;
      case 1	: 	/*Lets get first prompt		*/
        if (tls_waitsignon(cta->tls,cta->maxwait,proto,&(cta->info))!=true) {
	  char *trouble;

	  (void) asprintf(&trouble,"%s Invalid prompt from '%s' (prompt=<%s>)",
			           OPSR,hostname,cta->info);
	  (void) rpttrouble(cta,trouble);
	  proceed=false;	/*contact not successful*/
				/*reporting status info	*/
	  }
        break;
      case 2	:	/*doing Hello if needed		*/
	switch (proto) {
	  case (pro_ment)	:	/*Mentor HELO	*/
	  case (pro_smtp)	:	/*SMTP HELO	*/
	    if (doehlo(cta,hostname)==false) {
	      phase=999;/*trouble trouble		*/
	      }
	    break;
	  default		:	/*nothing to do	*/
	    break;
	  }
	break;
      case 3	:	/*if remote system is TLS, lets go TLS	*/
	if ((cta->capable&CPTLS)!=0) {
	  int statusloc;

	  statusloc=eml_sendcmd(cta,SIGNON,STRTTLS);
	  switch (statusloc) {
	    case SIGNON		:
	      if (tls_gotls(cta->tls,false)<0) {
	        char *trouble;

                (void) asprintf(&trouble,"%s Unable to negociate TLS with site <%s[%s]:%s>)",
				          OPSR,cta->peername,cta->peerip,port);
	  	(void) rpttrouble(cta,trouble);
		phase=999;		/*Big trouble	*/
		break;
	        }
	      switch (proto) {
		case (pro_smtp)	:
	          if (doehlo(cta,hostname)==false) {
	      	    phase=999;		/*big trouble	*/
		    }
		  break;
		default		:	/*nothing to do	*/
		  break;
		}
	      break;
	    case TLSNRDY	:	/*TLS not ready	*/
	      break;
	    default		:
	      if (cta!=(CTTYP *)0) {	/*always	*/
	        char *trouble;

                (void) asprintf(&trouble,"%s Unexpected SIGNON statu under TLS with site <%s[%s]:%s>)",
				          OPSR,cta->peername,cta->peerip,port);
	        (void) rpttrouble(cta,trouble);
		}
	      phase=999;		/*big trouble	*/
	      break;
	    }
	  }
        break;
      case 4	:	/*contact well established	*/
	cta->established=true;
	proceed=false;	/*every thing fine		*/
        break;
      default	:	/*Trouble trouble		*/
        cta=eml_lostcontact(cta);
	proceed=false;
        break;
      }
    phase++;
    }
  }
return cta;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to terminate an SMTP connexion*/
/*						*/
/************************************************/
CTTYP *eml_closecontact(CTTYP *contact)

{
if (contact!=(CTTYP *)0) {
  if (contact->established==true) 
    (void) eml_sendcmd(contact,QUITOK,SQUIT);
  contact=eml_lostcontact(contact);
  }
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait and acknowledge	*/
/*	contact from a remote site.		*/
/*						*/
/************************************************/
CTTYP *eml_acceptcontact(LSTTYP *channel)

{
#define	OPTR	"comeml.c:eml_acceptcontact,"

CTTYP *contact;
TLSFILE *tls;

contact=(CTTYP *)0;
if ((tls=tls_accept(channel))!=(TLSFILE *)0) {
  contact=setcontact(tls,channel->binding->proto,channel->binding->port);
  switch (channel->binding->proto) {
    case pro_ssmtp	:
      if (tls_gotls(contact->tls,true)<0) {
        (void) clm_alert(0,"%s Unable to open SSL connection on port <%s>",
			    OPTR,channel->binding->port);
        contact=eml_lostcontact(contact);	
        break;
	}
				/*NO break	*/
    case pro_smtp	:
      (void) eml_printfcrlf(contact,"%d %s %s %s-%s; %s",
                            SIGNON,contact->locname,"ESMTP",appname,
                            clm_getvers(),clm_ascsysstamp(time((time_t *)0)));
      (void) emlsettitle(contact);
      break;
    case pro_pop3	:
      contact=eml_lostcontact(contact);	
      (void) clm_alert(0,"%s proto pop3 not yet implemented",OPTR);
      break;
    case pro_pop3t	:
      contact=eml_lostcontact(contact);	
      (void) clm_alert(0,"%s proto pop3+tlst not yet implemented",OPTR);
      break;
    case pro_pop3s	:
      contact=eml_lostcontact(contact);	
      (void) clm_alert(0,"%s proto pop3s not yet implemented",OPTR);
      break;
    case pro_imap	:
      contact=eml_lostcontact(contact);	
      (void) clm_alert(0,"%s proto imap not yet implemented",OPTR);
      break;
    case pro_imapt	:
      contact=eml_lostcontact(contact);	
      (void) clm_alert(0,"%s proto imap+tls not yet implemented",OPTR);
      break;
    case pro_imaps	:
      contact=eml_lostcontact(contact);	
      (void) clm_alert(0,"%s proto imaps not yet implemented",OPTR);
      break;
    case pro_ment	:
      (void) eml_printfcrlf(contact,"%d %s %s %s-%s; %s",
                            SIGNON,contact->locname,"MENTOR",appname,
                            clm_getvers(),clm_ascsysstamp(time((time_t *)0)));
      (void) emlsettitle(contact);
      break;
    case pro_ukn	:
      (void) clm_alert(0,"%s undefined protocol",OPTR);
      contact=eml_lostcontact(contact);	
      break;
    }
  }
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to close a contact with a 	*/
/*	remote site.				*/
/*						*/
/************************************************/
CTTYP *eml_lostcontact(CTTYP *contact)

{
if (contact!=(CTTYP *)0) {
  (void) tls_close(contact->tls);
  if (contact->dbgname!=(char *)0) {
    if (fclose(stdout)<0)
      (void) clm_core_dump("%s Unable to fclose stdout (error=<%s>)",
		           "comeml.c:feml_lostcontact",strerror(errno));
    stdout=contact->bckstdout;
    if (debug==0)
      (void) unlink(contact->dbgname);
    contact->dbgname=clm_freestr(contact->dbgname);
    }
  contact=freecontact(contact);
  }
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to format the status report	*/
/*	conveniently for clement logs		*/
/*						*/
/************************************************/
char *eml_forlogs(char *logs,char *note)

{
char *forlogs;
char **lines;

forlogs=(char *)0;
lines=(char **)0;
(void) clm_cutchar(logs,'\r');
if ((lines=clm_splitstring(logs,'\n'))!=(char **)0) {
  int i;

  for (i=0;lines[i]!=(char *)0;i++) {
    if (forlogs!=(char *)0)
      forlogs=clm_astrcat(forlogs,"\n\t  ");
    forlogs=clm_astrcat(forlogs,lines[i]);
    }
  lines=(char **)vec_freelstlst((void **)lines,VECFREE);
  }
if (note!=(char *)0) {
  forlogs=clm_astrcat(forlogs,"\n\t  Note: ");
  forlogs=clm_astrcat(forlogs,note);
  }
return forlogs;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a "MAIL FROM:" line	*/
/*	to a remote SMTP host.			*/
/*						*/
/************************************************/
int eml_sendmfrom(CTTYP *contact,char *originator,u_long size)

{
char *fmsm;

fmsm=FMSM;
if ((contact->capable&CPSIZE)!=0)
  fmsm=FMSM " %s=%lu";
return eml_sendcmd(contact,CMDOK,fmsm,MAILF,originator,MSIZE,size);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert extract a dispatch	*/
/*	structure from a contact structure	*/
/*	according server information		*/
/*						*/
/************************************************/
DSPTYP *eml_contacttodsp(CTTYP *contact)

{
DSPTYP *dispatch;

dispatch=(DSPTYP *)0;
if (contact!=(CTTYP *)0) {
  dispatch=(DSPTYP *)calloc(1,sizeof(DSPTYP));
  dispatch->session=strdup(contact->session);
  dispatch->peerip=strdup(contact->peerip);
  dispatch->peername=strdup(contact->peername);
  dispatch->locip=strdup(contact->locip);
  dispatch->locport=strdup(contact->servport);
  dispatch->auth=contact->autok;
  dispatch->auth|=contact->relayable;
  dispatch->mtaille=contact->mtaille;
  if ((contact->evl!=(EVLTYP *)0)&&(contact->evl->spamlvl!=(char *)0))
    (void) sscanf(contact->evl->spamlvl,"%f/%f",&(dispatch->spmlvl),&(dispatch->spmlim));
  if (contact->frm!=(FRMTYP *)0) {
    dispatch->grpin=contact->frm->grpin;
    (void) asprintf(&(dispatch->efrom),"%s",contact->frm->efrom);
    }
  if (contact->evl!=(EVLTYP *)0) {
    (void) asprintf(&(dispatch->hfrom),"%s",contact->evl->realfrom);
    (void) asprintf(&(dispatch->subject),"%s",contact->evl->subject);
    (void) asprintf(&(dispatch->msgid),"%s",contact->evl->msgid);
    }
  dispatch->rcpts=eml_duprcpts(contact->rcpts);
  }
return dispatch;
}
