/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	Define all routine to manage email 	*/
/*	communication exchange.			*/
/*						*/
/************************************************/
#ifndef COMEML
#define COMEML
#include	"unieml.h"
#include	"gestls.h"
#include	"geshdr.h"

#define	CPTLS	0x0001	/*host STARTTLS capable	*/
#define	CPRQT	0x0002	/*is RQSTID capable	*/
#define	CPSIZE	0x0004	/*host is SIZE capable	*/

#define	DATAEND	"."	/*Email data end marker	*/

/*define 'From: <username@domain>' format	*/
#define	FMSM	"%s: <%s>"

typedef	enum	{
	s_timeout,	/*SMTP exchange timeout	*/
	s_quit,		/*SMTP exch completed	*/
	s_mail,		/*MAIL FROM information.*/
	s_rcpt,		/*RCPT TO request.	*/
	s_data,		/*DATA contents received*/
	s_baddata,	/*data-contents trouble	*/
	s_reset,	/*A reset was requested	*/
	s_auth,		/*An authentication rqst*/
	s_chkid,	/*A Check USER ID rqsted*/
	s_helo,		/*Got an HELO		*/
	s_ehlo,		/*Got an EHLO		*/
	s_id,		/*dfltcode received	*/
	s_goodmail,	/*A goodmail advice	*/
	s_trusting,	/*getting trusting data	*/
	s_unknown	/*SMTP code unknown	*/
	}SCODTYP;

typedef	struct	{
	char *mfrom;	/*MAIL FROM: SMTP proto */
	char *efrom;	/*from Extracted (EHLO)	*/
	char *peerip;	/*orig remote IP number	*/
	char *locip;	/*listening to IP	*/
	char *servport;	/*service listening port*/
	u_long mtaille;	/*announced mail size	*/
	int curbulk;	/*current bulk level	*/
	int maxbulk;	/*accepted bulk level	*/
	int frmstatus;	/*From cumulated status	*/
	int grpin;	/*From found groupid 	*/
	char *stainfo;	/*status info		*/
	}FRMTYP;

typedef struct	{
	TLSFILE *tls;	/*connection reference	*/
	int reset;	/*connection was reseted*/
	int established;/*contact established	*/
	int esmtp;	/*protocol en mode ESMTP*/
	char *locname;	/*local interface name	*/
	char *locip;	/*local interface ipnum	*/
	char *servport;	/*service port		*/
	char *peername;	/*remote interface name	*/
	char *peerip;	/*remote interface ipnum*/
	int maxwait;	/*contact max wait time	*/
	char *heloname;	/*remote helo name.	*/
	int badehlo;	/*flag Badhelo		*/
	char *msession;	/*contact main session	*/
	char *session;	/*contact unique ID	*/
	char *authname;	/*authentication name	*/
	int grpin;	/*auth name group id	*/
	int ctxflag;	/*rmt IP ban/trust flag	*/
	char *dbgname;	/*trace session filename*/
	char *mailname;	/*E-mail file full name	*/
	u_long mtaille;	/*E-mail size		*/
	char *clamvers;	/*clamav version Info	*/
	u_int sequence;	/*contact sequence	*/
	PROTYP proto;	/*contact used protocole*/
	int autok;	/*sender authenticated	*/
	int relayable;	/*sender is relayable	*/
	int founddata;	/*data found in session	*/
	char **headers;	/*header original info	*/
	EVLTYP *evl;	/*E-mail header infos	*/
	FRMTYP *frm;	/*SMTP FROM: information*/
	int totbad;	/*all session bad status*/
	int badrcpt;	/*number of bad RCPT	*/
	int goodrcpt;	/*number of good RCPT	*/
	char *foremail;	/*first good rcpt	*/
	RCPTYP **rcpts;	/*E-mail recipient list	*/
	char *virname;	/*virus name in E-mail	*/
	char *rjctnote;	/*rejection note	*/
	time_t lastdata;/*last data entry time	*/
	time_t start;	/*connection start	*/
	time_t stop;	/*connection stop	*/
	char *attime;	/*contact starting time	*/
	char *cipherid;	/*information about TLS	*/
	int capable;	/*remote host capabilty	*/
	FILE *bckstdout;/*standard stdout	*/
	char *info;	/*data collected so far	*/
	}CTTYP;

extern char *eml_cttoadvmsg(CTTYP *contact);
extern char *eml_cttotknmsg(CTTYP *contact);
extern void eml_advmsgupdrcpt(char *advice,RCPTYP **rcpts);
extern int eml_printfcrlf(CTTYP *contact,const char *fmt,...);
extern int eml_sendcmd(CTTYP *contact,int expected,char *fmt,...);
extern CTTYP *eml_opencontact(PROTYP proto,char *hostname,char *port,int maxwait,int *socker);
extern CTTYP *eml_closecontact(CTTYP *contact);
extern CTTYP *eml_acceptcontact(LSTTYP *channel);
extern CTTYP *eml_lostcontact(CTTYP *contact);
extern char *eml_forlogs(char *logs,char *note);
extern int eml_sendmfrom(CTTYP *contact,char *originator,u_long size);
extern DSPTYP *eml_contacttodsp(CTTYP *contact);
#endif
