/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	GESCNF:					*/
/*	Take care of all clement configuration	*/
/*	structure.				*/
/*						*/
/************************************************/
#include	<string.h>
#include	<errno.h>
#include	<netdb.h>
#include	<arpa/inet.h>
#include	<ctype.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unijrl.h"
#include	"unieml.h"
#include	"gescnf.h"
/*

*/
/************************************************/
/*						*/
/*	Procedure to free the memory used by	*/
/*	config definition.			*/
/*						*/
/************************************************/
CNFTYP *cnf_freeconfig(CNFTYP *cnf)

{
if (cnf!=(CNFTYP *)0) {
  cnf->binds=dev_freebindinf(cnf->binds);
  cnf->relais=rly_freerelais(cnf->relais);
  (void) free(cnf);
  cnf=(CNFTYP *)0;
  }
return cnf;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to dump de config contents	*/
/*	on a file contents (mainly for debug	*/
/*	purpose).				*/
/*						*/
/************************************************/
void cnf_dumpconfig(FILE *fout,CNFTYP *cnf)

{
if (cnf!=(CNFTYP *)0) {
  (void) fprintf(fout,"%s\n","/*");
  (void) fprintf(fout,"Last time scan: %s\n",
		       clm_ascsysstamp(cnf->lastscan));
  (void) fprintf(fout,"%s\n","*/");
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to log in application journal	*/
/*	domain and bind covered by configuration*/
/*						*/
/************************************************/
void cnf_showcnf(CNFTYP *cnf)

{
(void) jrl_journalling(false,APJRL,"Current configuration set at %s",
				    clm_ascsysdatetime(cnf->update));
if (cnf->binds!=(BININF **)0) {
  int i;

  (void) jrl_journalling(false,APJRL,"binding definitions");
  for (i=0;cnf->binds[i]!=(BININF *)0;i++) {
    static char *strpro[]={"smtp","smtps","pop3","ment",(char *)0};
    register BININF *bloc;

    bloc=cnf->binds[i];
    (void) jrl_journalling(false,APJRL,"%-7s %-40s % 5s % 3d",
				        strpro[(int)bloc->proto],
				        bloc->ip,bloc->port,bloc->iteration);
    }
  }
if (cnf->relais!=(RLYTYP *)0) {
  char **strrly;

  if ((strrly=rly_showrelais(cnf->relais))!=(char **)0) {
    char **ptr;

    for (ptr=strrly;*ptr!=(char *)0;ptr++)
       (void) jrl_journalling(false,APJRL,"%s",*ptr);
    strrly=(char **)vec_freelstlst((void **)strrly,VECFREE);
    }
  }
(void) jrl_journalling(true,APJRL,(char *)0);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the list of active	*/
/*	binded IP and port (for all active	*/
/*	contract)				*/
/*						*/
/************************************************/
BININF **cnf_getbinds(CNFTYP *cnf)

{
BININF **binds;

binds=(BININF **)0;
if (cnf->binds!=(BININF **)0) {
  int i;

  for (i=0;cnf->binds[i]!=(BININF *)0;i++) {
    BININF *bindloc;

    bindloc=dev_dupbind(cnf->binds[i]);
    binds=(BININF **)vec_addlstlst((void **)binds,(void *)bindloc);
    }
  }
return binds;
}
