/************************************************/
/*						*/
/*      Copyright:				*/
/*	 Jean-Marc Pigeon <jmp@safe.ca>	 2009	*/
/*						*/
/************************************************/
/* This program is free software; you can 	*/
/* redistribute it and/or modify it under the 	*/
/* terms of the GNU General Public License as	*/
/* published by the Free Software Foundation	*/
/* version 2 of the License			*/
/*						*/
/* This program is distributed in the hope that */
/* it will be useful, but WITHOUT ANY WARRANTY; */
/* without even the implied warranty of		*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR	*/
/* PURPOSE.  See the GNU General Public License	*/
/* for more details.				*/
/*						*/
/* You should have received a copy of the GNU	*/
/* General Public License along with this 	*/
/* program; if not, write to the Free Software	*/
/* Foundation, Inc., 51 Franklin Street,	*/
/* Fifth Floor, Boston, MA  02110-1301, USA.	*/
/************************************************/
/*						*/
/*	GESPBS:					*/
/*	Take care off Pop_Before_Smtp function.	*/
/*						*/
/************************************************/
#include	<errno.h>
#include	<time.h>
#include	<unistd.h>
#include	<string.h>
#include	<malloc.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unilck.h"
#include	"unieml.h"
#include	"gespbs.h"

/*file to record POP/IMAP before SMTP request	*/
#define	PIBSMTP	"pibsmtp"
#define	POPRELY	"poprelay"

#define	PIBFRM	"%ld %s %s  "

#define	TVAL	300	/*Entry validity time	*/

typedef	struct	{	/*relay information	*/
	time_t stamp;
	char *ip;
	char *username;
	}RLYTYP;
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return a previous entry	*/
/*	within a relay list.			*/
/*						*/
/************************************************/
static RLYTYP *srhrely(RLYTYP **rlys,char *ip)

{
RLYTYP *got;

got=(RLYTYP *)0;
if (rlys!=(RLYTYP **)0) {
  int i;

  for (i=0;rlys[i]!=(RLYTYP *)0;i++) {
    if (strcmp(rlys[i]->ip,ip)==0) {
      got=rlys[i];
      break;
      }
    }
  }
return got;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to transfert recent POP/IMAP	*/
/*	access information to poprelay file	*/
/*						*/
/************************************************/
static void dropclean(time_t curtime)

{
#define	ODDP	"gespbs.c:dropclean,"

#define	BGLINE	500

FILE *fpibnew;
FILE *fpiblst;
FILE *fpopnew;
RLYTYP **rlys;
const char *etcdir;
char *pibnew;
char *piblst;
char *popnew;
char *poprly;
int phase;
int proceed;
char line[BGLINE];

fpibnew=(FILE *)0;
fpiblst=(FILE *)0;
fpopnew=(FILE *)0;
etcdir=clm_appdir(d_etc);
pibnew=(char *)0;
piblst=(char *)0;
popnew=(char *)0;
poprly=(char *)0;
rlys=(RLYTYP **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*getting filename	*/
      (void) asprintf(&pibnew,"%s/%s.new",etcdir,PIBSMTP);
      (void) asprintf(&piblst,"%s/%s.lst",etcdir,PIBSMTP);
      (void) asprintf(&popnew,"%s/%s.lst",etcdir,POPRELY);
      (void) asprintf(&poprly,"%s/%s",etcdir,POPRELY);
      break;
    case   1	:	/*opening file		*/
      if ((fpiblst=fopen(piblst,"r"))==(FILE *)0)
	phase=9999;	/*file missing, get out	*/
      else {
	if ((fpibnew=fopen(pibnew,"w"))==(FILE *)0) {
	  (void) fclose(fpiblst);
	  phase=9999;
	  (void) clm_alert(0,"%s Unable to open <%s> (error=<%s>)",
			     ODDP,pibnew,strerror(errno));
          }
	else {
	  if ((fpopnew=fopen(popnew,"w"))==(FILE *)0) {
	    (void) fclose(fpiblst);
	    (void) fclose(fpibnew);
	    phase=9999;
	    (void) clm_alert(0,"%s Unable to open <%s> (error=<%s>)",
			       ODDP,popnew,strerror(errno));
	    }
	  }
	}
      break;
    case   2	:	/*transferring file	*/
      while (fgets(line,sizeof(line)-1,fpiblst)!=(char *)0) {
	time_t stamp;
	char ip[BGLINE];
	char username[BGLINE];

        if (sscanf(line,PIBFRM "\n",&stamp,ip,username)==3) {
	  if ((stamp+TVAL)>curtime) {
	    RLYTYP *prv;

	    if ((prv=srhrely(rlys,ip))!=(RLYTYP *)0) {
	      prv->stamp=stamp;
	      }
	    else {
	      prv=(RLYTYP *)calloc(1,sizeof(RLYTYP));
	      prv->stamp=stamp;
	      prv->ip=strdup(ip);
	      prv->username=strdup(username);
	      rlys=(RLYTYP **)vec_addlstlst((void **)rlys,(void *)prv);
	      }
	    }
          }
	}
      break;
    case   3	:	/*dumping file		*/
      if (rlys!=(RLYTYP **)0) {
	int i;

	for (i=0;rlys[i]!=(RLYTYP *)0;i++) {
	  (void) fprintf(fpibnew,PIBFRM "# %s\n",rlys[i]->stamp,
					       rlys[i]->ip,
					       rlys[i]->username,
					       clm_ascsystime(rlys[i]->stamp));
	  (void) fprintf(fpopnew,"%s/32\n",rlys[i]->ip);
	  rlys[i]->ip=clm_freestr(rlys[i]->ip);
	  rlys[i]->username=clm_freestr(rlys[i]->username);
	  (void) free(rlys[i]);
	  }
        (void) free(rlys);
	rlys=(RLYTYP **)0;
        }
      break;
    case   4	:	/*closing files		*/
      (void) fclose(fpopnew);
      (void) fclose(fpibnew);
      (void) fclose(fpiblst);
      break;
    case   5	:	/*moving PIBSMTP file	*/
      (void) unlink(piblst);
      if (link(pibnew,piblst)<0)
	(void) clm_alert(0,"%s Unable to link file name <%s> to name <%s> (bug?)",
			    ODDP,pibnew,piblst);
      (void) unlink(pibnew);
      break;
    case   6	:	/*moving POPRELY file	*/
      (void) unlink(poprly);
      if (link(popnew,poprly)<0)
	(void) clm_alert(0,"%s Unable to link file name <%s> to name <%s> (bug!?)",
			    ODDP,popnew,poprly);
      (void) unlink(popnew);
      break;
    default	:	/*SAFE guard		*/
      (void) free(pibnew);
      (void) free(piblst);
      (void) free(popnew);
      (void) free(poprly);
      proceed=false;
      break;
    }
  phase++;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to add an entry to the pibsmt	*/
/*	file.					*/
/*						*/
/************************************************/
void pbs_addentry(char *username,char *ip)

{
if (lck_locking(PIBSMTP,LCK_LOCK,5)==true) {
  FILE *fichier;
  char *filename;

  (void) asprintf(&filename,"%s/%s.lst",clm_appdir(d_etc),PIBSMTP);
  if ((fichier=fopen(filename,"a"))!=(FILE *)0) {
  /* "USER" and "IP" are 2 env values provided by 'dovecot' calling   */
    if ((username!=(char *)0)&&(ip!=(char *)0)) {
      time_t curtime;

      curtime=time((time_t *)0);
      (void) clm_alert(0,"Adding <%s/[%s]> to %s",username,ip,PIBSMTP);
      (void) fprintf(fichier,PIBFRM "# %s\n",curtime,ip,
					   username,clm_ascsystime(curtime));
      }
    (void) fclose(fichier);
    }
  else
    (void) clm_alert(0,"%s Unable to open file <%s> (status=<%s>)",
			appname,filename,strerror(errno));
  (void) lck_locking(PIBSMTP,LCK_UNLOCK,1);
  (void) free(filename);
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if file PIBSMTP was	*/
/*	change or due for a little old pop	*/
/*	access list cleaning.			*/
/*						*/
/************************************************/
void pbs_cleanentry()

{
static time_t lastclean=(time_t)0;

time_t curtime;

curtime=time((time_t *)0);
if ((lastclean+15)<curtime) {
  int letsdoit;
  char *filename;

  letsdoit=false;
  (void) asprintf(&filename,"%s/%s.lst",clm_appdir(d_etc),PIBSMTP);
  switch (clm_isnewfile(filename,lastclean)) {
    case  -1	:	/*input file missing	*/
      break;		
    case   0	:	/*file was not updated	*/
      break;		
    case	1	:	/*file has been updated	*/
      letsdoit=true;
      break;
    }
  if (letsdoit==true) {
    if (lck_locking(PIBSMTP,LCK_LOCK,2)==true) {
      (void) dropclean(curtime);
      (void) lck_locking(PIBSMTP,LCK_UNLOCK,1);
      lastclean=curtime;
      }
    }
  (void) free(filename);
  }
}
