/************************************************/
/*						*/
/*	GESSPF:					*/
/*	Take Care of all access to SPF1 	*/
/*	DNS informations.			*/
/*						*/
/************************************************/
#include	<ctype.h>
#include	<stdlib.h>
#include	<syslog.h>
#include	<string.h>
#include	<resolv.h> 
#include	<netdb.h> 
#include	<errno.h> 
#include	<arpa/inet.h>
#
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unidns.h"
#include	"unieml.h"
#include	"gesspf.h"

#define	SPF1	"v=spf1 "
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to expend SPF macro		*/
/*						*/
/************************************************/
static char *getmacro(char *param,AFNTYP *afnnum)

{
#define	OPGE	"gesspf.c:getmacro,"

char *expended;
int itsok;

expended=(char *)0;
itsok=true;
if (param!=(char *)0) {
  char *ptr;

  expended=strdup(param);
  while ((ptr=strstr(expended,"%{"))!=(char *)0) {
    char *cnvstr;
    char *ptr2;
    char *dilate;

    cnvstr=(char *)0;
    if ((ptr2=strchr(ptr+2,'}'))==(char *)0) {
      itsok=false;
      break;
      }
    *ptr2='\000';
    ptr2++;
    dilate=strdup(ptr+2);
    switch (dilate[0]) {
      case 'i'	:	/*take the remote IP num	*/
	switch (dilate[1]) {
	  case '\000':
	    cnvstr=afn_stripnum(afnnum);
	    break;
	  case 'r'   :
	    cnvstr=afn_reversipnum(afnnum);
	    break;
	  default    :
	    (void) clm_alert(0,"%s Unexpected macro parameter <%s> within <%s>",OPGE,dilate,param);
	    itsok=false;
	    break;
	  }
	break;
      default	:
	(void) clm_alert(0,"%s Unexpected macro sequence <%s> within <%s>",OPGE,dilate,param);
	itsok=false;
	break;
      }
    (void) free(dilate);
    if (itsok==false)
      break;
    *ptr='\000';	/*segmenting begore %{		*/
    (void) asprintf(&dilate,"%s%s%s",expended,cnvstr,ptr2);
    (void) free(expended);
    (void) free(cnvstr);
    expended=dilate;
    }
  if (itsok==false) {
    (void) free(expended);
    expended=(char *)0;
    }
  }
return expended;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to locate an SPF sequence	*/
/*	within the SPF string			*/
/*						*/
/************************************************/
static char *getspfseq(char **txt)

{
char *seq;

seq=(char *)0;
if (*txt!=(char *)0) {
  while ((*txt)[0]==' ')
    (*txt)++;
  if ((*txt)[0]!='\000') {
    char *nxt;

    seq=strdup(*txt);
    if ((nxt=strchr(seq,' '))!=(char *)0) {
      char *cidr;

      *nxt='\000';
      cidr=nxt+1;
      (void) clm_skipblank(cidr);
      if (*cidr=='/') {
        (void) memmove(nxt,cidr,strlen(cidr)+1);
	if ((cidr=strchr(nxt,' '))!=(char *)0)
          *cidr='\000';
	}
      }
    }
  }
return seq;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract the CIDR info from	*/
/*	a string sequence.			*/
/*						*/
/************************************************/
static int getcidr(char *param)

{
int cidr;
char *ptr;

cidr=128;
if ((param!=(char *)0)&&((ptr=strchr(param,'/'))!=(char *)0)) {
  *ptr='\000';
  cidr=atoi(ptr+1);
  }
return cidr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return allway true		*/
/*						*/
/************************************************/
static SPFENUM thatsall(SPFENUM cnd,char *param,char *domain,
		        AFNTYP *afnnum,int mask,char **list)

{
SPFENUM newcnd;

newcnd=cnd;
switch (cnd) {
  case spf_neutral	:
    /*should be fail at that stage, but company as telus, doesn't understand	*/
    /*SPF purpose.								*/
    /*newcnd=spf_fail;								*/
    break;
  default		:
    newcnd=cnd;
    break;
  }
return newcnd;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to create a dynamique domain	*/
/*	name, this procedure still need to be	*/
/*	writen.					*/
/*						*/
/************************************************/
static SPFENUM doexists(SPFENUM cnd,char *param,char *domain,
		        AFNTYP *afnnum,int mask,char **list)

{
SPFENUM status;

status=spf_neutral;
if ((param!=(char *)0)&&(strlen(param)>0)) {
  char *hostname;

  if ((hostname=getmacro(param,afnnum))!=(char *)0) {
    char **iplist;

    if ((iplist=dns_getiplist(hostname))!=(char **)0) {
      status=cnd;
      iplist=(char **)vec_freelstlst((void **)iplist,VECFREE);
      } 
    (void) free(hostname);
    }
  else
    (void) clm_alert(0,"gesspf.c:doexists, do not know how to use "
		        "<%s> for domain<%s>",param,domain);
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if address match	*/
/*	the one from the domain.		*/
/*						*/
/************************************************/
static SPFENUM checkaddr(SPFENUM cnd,char *param,char *domain,
			 AFNTYP *afnnum,int mask,char **list)

{
SPFENUM result;
char *hostname;

result=spf_neutral;
hostname=domain;
if ((param!=(char *)0)&&(strlen(param)>0))
  hostname=param;
if (dns_matchIPrec(hostname,afnnum,mask)==true)
  result=cnd;
return result;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if MX list is among	*/
/*	norig IP number				*/
/*						*/
/************************************************/
static SPFENUM checkmx(SPFENUM cnd,char *param,char *domain,
		       AFNTYP *afnnum,int mask,char **list)

{
#define	OPMX	"gesspf.c:checkmx,"

SPFENUM result;
char *mxdomain;
MXTYP **mxlist;

result=spf_neutral;
mxdomain=domain;
if ((param!=(char *)0)&&(strlen(param)>0))
  mxdomain=param;
if ((mxlist=dns_getmx(mxdomain))!=(MXTYP **)0) {
  int found;
  struct addrinfo hints;
  int i;

  found=false;
  (void) memset(&hints,'\000',sizeof(hints));
  hints.ai_family=PF_UNSPEC;
  hints.ai_socktype=SOCK_STREAM;
  hints.ai_flags=HINTFLG;
  for (i=0;(found==false)&&(mxlist[i]!=(MXTYP *)0);i++) {
    int status;
    struct addrinfo *res;
    struct addrinfo *rp;
    int j;

    status=getaddrinfo(mxlist[i]->mxname,"",&hints,&res);
    if (status!=0) {
      (void) clm_alert(0,"%s Unable to find addrinfo for <%s> (error=<%s>)",
			 OPMX,mxlist[i]->mxname,gai_strerror(status));
      continue;
      }
    for (j=0,rp=res;(found==false)&&(rp!=(struct addrinfo *)0);rp=rp->ai_next,j++) {
      AFNTYP *addrnum;

      if ((addrnum=afn_getaddrinfo(rp))==(AFNTYP *)0) {
        (void) clm_alert(0,"%s Found unknown inet family for <%s> IP number",
			    OPMX,mxlist[i]->mxname);
	continue;
	}
      switch (afn_cmpipnum(addrnum,afnnum,mask)) {
	case false	:	/*not found	*/
	  break;
	case true	:	/*found		*/
          found=true;
          result=cnd;
	  break;
	case -1		:	/*trouble	*/
	  (void) clm_alert(0,"%s Unable to compare IP for hostname <%s> (error=<%s>)",
                             OPMX,mxlist[i]->mxname,strerror(errno));
	  break;
	}
      addrnum=afn_freeipnum(addrnum);
      }
    (void) freeaddrinfo(res);
    }
  mxlist=dns_cleanmx(mxlist);
  }
return result;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check domain against the	*/
/*	PTR list				*/
/*						*/
/************************************************/
static SPFENUM checkptr(SPFENUM cnd,char *param,char *domain,
		        AFNTYP *afnnum,int mask,char **list)

{
#define	OPPR	"gesspf.c:checkptr,"

SPFENUM result;
char *target;
char **ptrlist;

result=spf_neutral;
target=domain;
if ((param!=(char *)0)&&(strlen(param)>0))
  target=param;
if ((ptrlist=dns_getptr(afnnum))!=(char **)0) {
  int found;
  struct addrinfo hints;
  int i;

  found=false;
  (void) memset(&hints,'\000',sizeof(hints));
  hints.ai_family=PF_UNSPEC;
  hints.ai_socktype=SOCK_STREAM;
  hints.ai_flags=HINTFLG;
  for (i=0;(found==false)&&(ptrlist[i]!=(char *)0);i++) {
    int status;
    struct addrinfo *res;
    struct addrinfo *rp;
    int j;

    status=getaddrinfo(ptrlist[i],"",&hints,&res);
    if (status!=0) {
      (void) clm_alert(0,"%s Unable to find addrinfo for <%s> (error=<%s>)",
			  OPPR,ptrlist[i],gai_strerror(status));
      continue;
      }
    for (j=0,rp=res;(found==false)&&(rp!=(struct addrinfo *)0);rp=rp->ai_next,j++) {
      AFNTYP *addrnum;

      if ((addrnum=afn_getaddrinfo(rp))==(AFNTYP *)0) {
        (void) clm_alert(0,"%s Found unknown inet family for <%s> IP number",
			    OPMX,ptrlist[i]);
	continue;
	}
      switch (afn_cmpipnum(addrnum,afnnum,mask)) {
	case false	:	/*not found	*/
	  break;
	case true	:	/*found		*/
	  if (strstr(ptrlist[i],target)!=(char *)0) {
            found=true;
            result=cnd;
	    }
	  break;
	case -1		:	/*trouble	*/
	  (void) clm_alert(0,"%s Unable to compare IP for hostname <%s> (error=<%s>)",
                             OPMX,ptrlist[i],strerror(errno));
	  break;
	}
      addrnum=afn_freeipnum(addrnum);
      }
    (void) freeaddrinfo(res);
    }
  ptrlist=(char **)vec_freelstlst((void **)ptrlist,VECFREE);
  }
return result;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to include check with a new	*/
/*	domain name (replacing current sender	*/
/*	domain by a new one)			*/
/*						*/
/************************************************/
static SPFENUM doinclude(SPFENUM cnd,char *param,char *domain,
			 AFNTYP *afnnum,int mask,char **list)

{
SPFENUM result;

result=spf_neutral;
if (param==(char *)0) {
  (void) clm_alert(0,"gesspf.c:domain, Param missing for domain <%s>",domain);
  }
else {
  if (strcmp(domain,param)!=0) {
    result=spf_checkdns(param,afnnum,list);
    }
  }
return result;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to redirect a request to	*/
/*	another domain.				*/
/*						*/
/************************************************/
static SPFENUM doredirect(SPFENUM cnd,char *param,char *domain,
		          AFNTYP *afnnum,int mask,char **list)

{
cnd=spf_neutral;
if ((param!=(char *)0)&&(strchr(param,'{')==(char *)0)) {
  cnd=spf_checkdns(param,afnnum,list);
  }
return cnd;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to swallow the EXP directive,	*/
/*	we don't care about it.			*/
/*	another domain.				*/
/*						*/
/************************************************/
static SPFENUM doexp(SPFENUM cnd,char *param,char *domain,
		          AFNTYP *afnnum,int mask,char **list)

{
return spf_fail;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check an IPV4 or IPV6 	*/
/*						*/
/************************************************/
static SPFENUM checkip(SPFENUM cnd,char *param,char *domain,
		       AFNTYP *afnnum,int mask,char **list)

{
#define	OPCI	"gesspf.c:checkip,"

SPFENUM result;

result=spf_neutral;
if (param==(char *)0) {
  (void) clm_alert(0,"%s Param missing for domain <%s>",OPCI,domain);
  }
else {
  AFNTYP *afnpar;

  if ((afnpar=afn_getipnum(param))!=(AFNTYP *)0) {
    switch (afn_cmpipnum(afnpar,afnnum,mask)) {
      case false	:/*not equal		*/
	break;
      case true		:/*found it equal!	*/
	result=cnd;
	break;
      case -1		:/*trouble trouble	*/
	(void) clm_alert(0,"%s Unable to compare IP for hostname <%s> (error=<%s>)",
			    OPCI,param,strerror(errno));
	break;
      }
    afnpar=afn_freeipnum(afnpar);
    }
  else {
    (void) clm_alert(0,"%s Param=<%s> wrong for domain <%s>",
		       OPCI,param,domain);
    }
  }
return result;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check an SPF sequence and	*/
/*	and return requested status if test	*/
/*	succeed.				*/
/*						*/
/************************************************/
static SPFENUM checkspfstatus(SPFENUM cnd,char *txt,char *domain,
			      AFNTYP *afnnum,char **list)

{
static	struct	{
	char *key;	/*sequence key	*/
	SPFENUM (*checker)(SPFENUM cond,char *param,char *domaine,
		           AFNTYP *afnnumref,int mask,char **liste);
	char sep;
	}chklst[]={
	  {"redirect",doredirect,'='},
	  {"include",doinclude,':'},
	  {"exists",doexists,':'},
	  {"all",thatsall,':'},
	  {"ptr",checkptr,':'},
	  {"ip4",checkip,':'},
	  {"ip6",checkip,':'},
	  {"mx",checkmx,':'},
	  {"a",checkaddr,':'},
	  {"exp",doexp,'='},
	  {(char *)0,thatsall,'\000'}
	};

SPFENUM status;
int i;

status=spf_unknown;
for (i=0;chklst[i].key!=(char *)0;i++) {
  if (strncasecmp(txt,chklst[i].key,strlen(chklst[i].key))==0) {
    char *param;
    int mask;

    mask=0;
    param=(char *)0;
    status=spf_neutral;
    txt+=strlen(chklst[i].key);
    if (txt[0]==chklst[i].sep)
      param=txt+1;
    mask=getcidr(txt);
    status=chklst[i].checker(cnd,param,domain,afnnum,mask,list);
    break;
    }
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check SPF1 information	*/
/*	extracted from DNS			*/
/*						*/
/************************************************/
SPFENUM spf_checkdns(char *domain,AFNTYP *afnnum,char **list)

{
#define	OPCH	"gespf.c:spf_checkdns,"
SPFENUM status;
int found;
int num;
char **txts;

status=spf_neutral;
found=false;
if (list==(char **)0) {
  (void) clm_core_dump("%s 'list' not provisionned for domain <%s>)",
			    OPCH,domain);
  /*no return from core_dump	*/
  }
for (num=0;list[num]!=(char *)0;num++) {
  if (strcmp(list[num],domain)==0) {
    found=true;
    break;
    }
  }
if ((found==true)||(num>MXSPFINC)) 
  return status;
list[num]=domain;
found=false;
if ((txts=dns_getspf(domain))!=(char **)0) {
  int i;

  for (i=0;txts[i]!=(char *)0;i++) {
    char *spf;

    spf=txts[i];
    while (isblank(*spf)!=0)
      spf++; 
    if (strstr(spf,SPF1)==spf) {
      char *start;

      if ((start=strstr(spf,SPF1))!=(char *)0) {
        char *seq;

        start=strchr(start,' ');
        while ((seq=getspfseq(&start))!=(char *)0) {
          SPFENUM cnd;
          char *ptr;

          ptr=seq;
          start+=strlen(seq);
          switch (seq[0]) {
	    case	'?'	:
	      cnd=spf_neutral;
	      ptr++;	
	      break;
	    case	'~'	:
	      cnd=spf_softfail;
	      ptr++;	
	      break;
	    case	'-'	:
	      cnd=spf_fail;
	      ptr++;	
	      break;
	    case	'+'	:
	      ptr++;	/*NO BREAK	*/
	    default		:
	      cnd=spf_pass;
	      break;
            }
          status=checkspfstatus(cnd,ptr,domain,afnnum,list);
          switch (status) {
	    case spf_pass	:
	      found=true;	/*no Break			*/
	    case spf_fail	:
	      break;
	    case spf_unknown:
	      (void) clm_alert(0,"gesspf.c,spf_checkdns, spf sequence <%s> "
		 		 "unknown " "for domain <%s>",ptr,domain);
	      break;
	    default		:
	      break;
	    }
          (void) free(seq);
          if (found==true) 
				/*no need to go futher	SPF is OK*/
	    break;
          }
	if (debug>=6) {		/*definition pour debug SPF	*/ 
	  static char *sspf[]={	/*must be aligned with SPFENUM definition*/
		"spf_pass",	/*OK if condition apply	*/
		"spf_fail",	/*NOK if condition apply*/
		"spf_softfail",	/*Msg origin is dubiuos	*/
		"spf_neutral",	/*same status as no SPF	*/
		"spf_relax",	/*access MUST be relaxed*/
		"spf_unknown"	/*Unknown SPF directive	*/
		};
	  char *strip;

	  strip=afn_stripnum(afnnum);
          (void) clm_alert(6,"spf_checkdns, <%s> for <%s/%s> with <%s>",
	       		      sspf[(int)status],domain,strip,spf);
	  (void) free(strip);
	  }
        }
      if (found==true) {
				/*no need to go futher	SPF is OK*/
        break;
        }
      }
    }
  txts=(char **)vec_freelstlst((void **)txts,VECFREE);
  }
else {
  switch (errno) {	/*trouble with DNS?	*/
    case ETIMEDOUT :	/* Connection timed out	*/
      status=spf_timeout;
      break;
    default	   :
      break;
    }
  }
list[num]=(char *)0;
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check is the remote could	*/
/*	be set with a 'goodspf' trust level.	*/
/*	Return true, false, or bad		*/
/*						*/
/************************************************/
int spf_isgoodspf(char *from,char *peerip)

{
#define	OPBA	"gesspf.c:spf_isgoodspf,"

int goodspf;
char *domain;

goodspf=true;
if ((domain=eml_getdomain(from))!=(char *)0) {
  char **list;
  char *locip;
  char *ptr;
  SPFENUM spfstat;
  AFNTYP *afnnum;

  list=(char **)calloc(MXSPFINC+3,sizeof(char *));
  locip=strdup(peerip);
  if ((ptr=strrchr(locip,'/'))!=(char *)0) {
    *ptr='\000';
    }
  afnnum=afn_getipnum(locip);
  switch (spfstat=spf_checkdns(domain,afnnum,list)) {
    case spf_pass		:	/*OK if condition apply	*/
      break;
    case spf_fail		:	/*NOK if condition apply*/
      goodspf=false;
      break;
    case spf_neutral		:	/*same status as no SPF	*/
      break;
    case spf_softfail		:	/*Msg origin is dubious	*/
    case spf_timeout		:	/*got a domain timeout	*/
      goodspf=maybe;
      break;
    case spf_unknown		:	/*Unknown SPF directive	*/	
    default			:
      (void) clm_alert(0,"%s spf status for  <%s/[%s]spfstat='%d' (bug?)>",
			 OPBA,from,locip,spfstat);
      break;
    }
  afnnum=afn_freeipnum(afnnum);
  (void) free(list);
  (void) free(locip);
  }
return goodspf;
}
