/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	LVLIPC:					*/
/*	Exchange message area			*/
/*						*/
/************************************************/
#include	<sys/poll.h>
#include	<errno.h>
#include	<time.h>
#include	<string.h>
#
#include	"lowtyp.h"
#include	"subclm.h"
#include	"lvlipc.h"
/*

*/
/************************************************/
/*						*/
/*	Procedure to get a message exchange	*/
/*	area.					*/
/*						*/
/************************************************/
MXCTYP *ipc_getmsgxchg(u_int boxsize,u_int mxproc)

{
#define	OPGX	"lvlipc.c:ipc_getmsgxchg,"

MXCTYP *mxc;
int phase;
int proceed;

mxc=(MXCTYP *)calloc(1,sizeof(MXCTYP));
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*setting shared memory		*/
      if ((mxc->bptr=ipc_getbox(boxsize,mxproc))==(BPTR *)0) {
	(void) clm_alert(0,"%s Unable to prepare shared memory",OPGX);
	proceed=false;	/*trouble trouble		*/
	}
      break;
    case   1	:  	/*getting a message pool for ckr*/
      if ((mxc->ckrpool=ipc_openpool(mxc->bptr))==(u_long)0) {
        (void) clm_alert(0,"%s Unable to get checker pool",OPGX);
	phase=999;	/*trouble trouble		*/
	}
      break;
    case   2	:  	/*getting msg pool for mentor	*/
      if ((mxc->mntpool=ipc_openpool(mxc->bptr))==(u_long)0) {
        (void) clm_alert(0,"%s Unable to get mentor pool",OPGX);
	phase=999;	/*trouble trouble		*/
	}
      break;
    case   3	:	/*everything fine		*/
      proceed=false;
      break;
    default	:	/*SAFE guard			*/
      mxc=ipc_dropmsgxchg(mxc);
      proceed=false;
      break;
    }
  phase++;
  }
return mxc;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to release a message exchange	*/
/*	area.					*/
/*						*/
/************************************************/
MXCTYP *ipc_dropmsgxchg(MXCTYP *mxc)

{
if (mxc!=(MXCTYP *)0) {
  (void) ipc_closepool(mxc->bptr,mxc->mntpool);
  (void) ipc_closepool(mxc->bptr,mxc->ckrpool);
  mxc->bptr=ipc_dropbox(mxc->bptr);
  (void) free(mxc);
  mxc=(MXCTYP *)0;
  }
return mxc;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to wait for event, either on	*/
/*	TLS channel or on the message channel	*/
/*						*/
/*	event on TLS channel are response to	*/
/*	previous request and are directly	*/
/*	transmitted back to requester. while	*/
/*	message channel (request) are extracted	*/
/*						*/
/************************************************/
IPEVTYP ipc_waitevent(BPTR *bptr,TLSFILE *tls,u_int millisec)


{
#define	OPWA	"lvlipc.c:ipc_waitevent,"

IPEVTYP action;
sigset_t oldmask;
sigset_t newmask;
register int phase;
register int proceed;

action=ipev_none;
(void) sigemptyset(&newmask);
(void) sigaddset(&newmask,SIGALRM);
(void) sigprocmask(SIG_BLOCK,&newmask,&oldmask);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*do we have a post	*/
      if (ipc_postpending(bptr)>0) {
	action=ipev_post;	/*yes		*/
	phase=999;	/*getting out fast	*/
	}
      break;
    case 1	:	/*is there a TLS	*/
      if (tls==(TLSFILE *)0) {
	TIMESPEC timeout;
	sigset_t oldset;
			/*no lets set a delay	*/
	action=ipev_post;
	timeout.tv_sec=millisec/1000;
        timeout.tv_nsec=(millisec%1000)*1000000;
        (void) sigprocmask(SIG_SETMASK,&oldmask,&oldset);
	if (nanosleep(&timeout,&timeout)==0)
	  action=ipev_none;	/*std timeout	*/
        (void) sigprocmask(SIG_SETMASK,&oldset,(sigset_t *)0);
	phase=999;	
	}
      break;
    case 2	:	/*do we have a line	*/
      if (tls_isready(tls)==true) {
	action=ipev_tls;/*yes			*/
	phase=999;	/*getting out fast	*/
	}
      break;
    case 3	:	/*lets wait for events	*/
      switch (tls_waitforchar(tls,(u_long)(millisec*1000),&oldmask)) {
        case -1	:
          switch(errno) {
            case EINTR	:	/*received a signal, lets see...*/
      	      action=ipev_post;	/*must be an IPS post		*/
      	      break;		/*could be a TERM signal        */
      	    default         :	/*hummm real code fault report  */
      	      (void) clm_core_dump("%s poll error '%s'",OPWA,strerror(errno));
      	      break;
	    }
          break;
        case  0	:       	/*standard time out	*/
          break;
        default	:		/*we got TLS input	*/
          action=ipev_tls;	/*lets flag it		*/
          break;
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false; 
      break;
    }
  phase++;
  }
(void) sigprocmask(SIG_SETMASK,&oldmask,(sigset_t *)0);
return action;
}
