/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	LVLPRO:					*/
/*	Module doing components protocol	*/
/*	exchanges handling.			*/
/*						*/
/************************************************/
#include	<errno.h>
#include	<string.h>
#include	<ifaddrs.h>

#include	"lowtyp.h"
#include	"utlprc.h"
#include	"utlvec.h"
#include	"unidns.h"
#include	"unijrl.h"
#include	"gesclm.h"
#include	"lvlpro.h"
/*

*/
/************************************************/
/*						*/
/*	Procedure to locate a previously 	*/
/*	existing device definition in a list	*/
/*						*/
/************************************************/
static LSTTYP *searchprv(LSTTYP **lst,LSTTYP *prv)

{
LSTTYP *loc;

loc=(LSTTYP *)0;
if (lst!=(LSTTYP **)0) {
  register BININF *bprv;
  register int i;

  bprv=prv->binding;
  for (i=0;lst[i]!=(LSTTYP *)0;i++) {
    register BININF *nprv;

    nprv=lst[i]->binding;
    if (nprv->proto!=bprv->proto)
      continue;
    if (nprv->port!=bprv->port)
      continue;
    if (strcmp(nprv->ip,bprv->ip)!=0)
      continue;
    loc=lst[i];
    break;
    }
  }
return loc;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to scan binding info from	*/
/*	a string, return an LSTTYP* structure	*/
/*						*/
/************************************************/
static LSTTYP *scaninfo(PROTYP proto,BININF *bindinfo)

{
LSTTYP *lstloc;

lstloc=(LSTTYP *)calloc(1,sizeof(LSTTYP)); 
lstloc->handle=-1;
lstloc->binding=dev_dupbind(bindinfo);
if (lstloc->binding->iteration>0)
  lstloc->process=(pid_t *)calloc(lstloc->binding->iteration,sizeof(pid_t));
return lstloc;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to have a list of all		*/
/*	ethernet device available.		*/
/*						*/
/************************************************/
static struct ifaddrs *devethlist()

{
#define	OPVH	"lvlpro.c:devethlist,"

struct ifaddrs *list;
int handle;
int phase;
int proceed;

list=(struct ifaddrs *)0;
handle=-1;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0 	:	/*opening socket	*/
      if ((handle=socket(PF_INET6,SOCK_DGRAM,0))<0) {
        (void) clm_alert(0,"%s Unable to open a socket (error='%s')",
		            OPVH,strerror(errno));
	proceed=false;	/*trouble trouble	*/
	}
      break;
    case 1 	:	/*getting interface list*/
      if (getifaddrs(&list)<0) {
        (void) clm_alert(0,"%s Unable to open interface address (error='%s')",
		            OPVH,strerror(errno));
	}
      break;
    default	:	/*SAFE guard		*/
      (void) close(handle);
      proceed=false;
      break;
    }
  phase++;
  }
return list;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to duplicate a device lst 	*/
/*	contents.				*/
/*						*/
/************************************************/
static LSTTYP *duplst(LSTTYP *lst)

{
LSTTYP *neu;

neu=(LSTTYP *)0;
if (lst!=(LSTTYP *)0) {
  register u_int iteration;

  iteration=lst->binding->iteration;
  neu=(LSTTYP *)calloc(1,sizeof(LSTTYP)); 
  neu->binding=dev_dupbind(lst->binding);
  neu->handle=lst->handle;
  neu->process=(pid_t *)calloc(iteration,sizeof(pid_t));
  (void) memcpy(neu->process,lst->process,iteration*sizeof(pid_t));
  }
return neu;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to locate an ANY_ADDR within	*/
/*	the device list, return a list of	*/
/*	any devices.				*/
/*						*/
/************************************************/
static LSTTYP **grabany(LSTTYP **lst)

{
LSTTYP **lstany;

lstany=(LSTTYP **)0;
if (lst!=(LSTTYP **)0) {
  int i;


  for (i=0;lst[i]!=(LSTTYP *)0;i++) {
    if (strcmp(lst[i]->binding->ip,"*")==0) {
      LSTTYP *ke;

      ke=duplst(lst[i]);
      lstany=(LSTTYP **)vec_addlstlst((void **)lstany,(void *)ke);
      lst[i]->binding->iteration=0;
      }
    }
  }
return lstany;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to locate an ANY_ADDR within	*/
/*	the device list an merge iteration	*/
/*	on it.					*/
/*						*/
/************************************************/
static LSTTYP **mergelst(LSTTYP **lst)

{
LSTTYP **lstany;

if ((lstany=grabany(lst))!=(LSTTYP **)0) {
  int i;

  for (i=0;lstany[i]!=(LSTTYP *)0;i++) {
    register u_int niter;
    register LSTTYP *any;
    register int j;

    any=lstany[i];
    niter=0;
    for (j=0;lst[j]!=(LSTTYP *)0;j++) {
      if (any->binding->proto!=lst[j]->binding->proto)
	continue;
      if (any->binding->port!=lst[j]->binding->port)
	continue;
      if (any->binding->iteration==0)
	continue;
      niter=any->binding->iteration+lst[j]->binding->iteration;
      any->process=(pid_t *)realloc(any->process,niter*sizeof(pid_t));
      (void) memcpy(any->process+any->binding->iteration,lst[j]->process,
		    lst[j]->binding->iteration*sizeof(pid_t));
      lst[j]->binding->iteration=0;
      }
    any->binding->iteration=niter; 
    }
  /*taking care of remaining in the list	*/
  for (i=0;lst[i]!=(LSTTYP *)0;i++) {
    LSTTYP *ke;

    ke=(LSTTYP *)0;
    if (lst[i]->binding->iteration==0)
      continue;
    ke=duplst(lst[i]);
    lstany=(LSTTYP **)vec_addlstlst((void **)lstany,(void *)ke);
    }
  lst=pro_freelst(lst);
  lst=lstany;
  }
return lst;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to build a brand new list of	*/
/*	ip/port/iteration to listen on.		*/
/*	IF binfo is EMPTY, default IP and port	*/
/*	are provided (both localhost and iphost	*/
/*	set on DPORT).				*/
/*						*/
/************************************************/
static LSTTYP **buildlst(PROTYP proto,BININF **binfo)

{
LSTTYP **lst;
BININF **linfo;

lst=(LSTTYP **)0;
linfo=(BININF **)0;
if (binfo==(BININF **)0) {
  struct ifaddrs *list;
  
  if ((list=devethlist())!=(struct ifaddrs *)0) {
    struct ifaddrs *ifa;

    for (ifa=list;ifa!=(struct ifaddrs *)0;ifa=ifa->ifa_next) {
      void *addr;
      char straddr[INET6_ADDRSTRLEN];

      if (strcmp(ifa->ifa_name,"lo")==0)
        continue;
      switch (ifa->ifa_addr->sa_family) {
  	case  AF_INET6	:	/*IPV6		*/
    	  addr=&(((struct sockaddr_in6 *)ifa->ifa_addr)->sin6_addr);
          if (IN6_IS_ADDR_LINKLOCAL(addr)==true)
	    continue;
    	  break;
  	case  AF_INET	:	/*IPV4		*/
  	default		:
    	  addr=&(((struct sockaddr_in *)ifa->ifa_addr)->sin_addr);
    	  break;
	}
      if (inet_ntop(ifa->ifa_addr->sa_family,addr,straddr,sizeof(straddr))==(char *)0) {
        BININF *ninf;

        ninf=dev_mkbindinf(pro_smtp,strdup(straddr),SMTPPORT,1);
        linfo=(BININF **)vec_addlstlst((void **)linfo,(void *)ninf);
	}
      }
    (void) freeifaddrs(list);
    }
  else {
    BININF *ninf;
    /*no eth device found! going on lo		*/
    /*proto smtp, port 25, one binding only	*/
    ninf=dev_mkbindinf(pro_smtp,"127.0.0.1",SMTPPORT,1);
    linfo=(BININF **)vec_addlstlst((void **)linfo,(void *)ninf);
    }
  binfo=linfo; 
  }
if (binfo!=(BININF **)0) {
  int i;

  for (i=0;binfo[i]!=(BININF *)0;i++) {
    LSTTYP *lstloc;

    if ((lstloc=scaninfo(proto,binfo[i]))!=(LSTTYP *)0) {
      LSTTYP *prv;

      if ((prv=searchprv(lst,lstloc))!=(LSTTYP *)0) {
        int niter;

        niter=prv->binding->iteration+lstloc->binding->iteration;
        prv->process=(pid_t *)realloc(prv->process,niter*sizeof(pid_t));
        (void) memset(prv->process+prv->binding->iteration,'\000',
		      lstloc->binding->iteration*sizeof(pid_t));
        prv->binding->iteration=niter;
	lstloc=dev_onefree(lstloc);
        }
      else {
        lst=(LSTTYP **)vec_addlstlst((void **)lst,(void *)lstloc);
        }
      }
    }
  }
lst=mergelst(lst);
linfo=dev_freebindinf(linfo);
return lst;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory related to a	*/
/*	binded devices list.			*/
/*						*/
/************************************************/
LSTTYP **pro_freelst(LSTTYP **lst)

{
if (lst!=(LSTTYP **)0) {
  int i;

  for (i=0;lst[i]!=(LSTTYP *)0;i++) {
   lst[i]=dev_onefree(lst[i]);
    }
  (void) free(lst);
  lst=(LSTTYP **)0;
  }
return lst;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to establish the list of	*/
/*	ip/port/iteration to listen on.		*/
/*						*/
/************************************************/
LSTTYP **pro_getlst(LSTTYP **prv,PROTYP proto,BININF **binfo)

{
LSTTYP **newlst;

if ((newlst=buildlst(proto,binfo))!=(LSTTYP **)0) {
  if (prv!=(LSTTYP **)0) {
    PRCTYP *prc;
    LSTTYP **mrglst;
    int i;

    prc=(PRCTYP *)0;
    mrglst=(LSTTYP **)0;
    for (i=0;newlst[i]!=(LSTTYP *)0;i++) {
      LSTTYP *old;

      if ((old=searchprv(prv,newlst[i]))!=(LSTTYP *)0) {
	int iter;

	iter=newlst[i]->binding->iteration;
	if (old->binding->iteration!=iter) {

	  if (old->binding->iteration<iter) {
	    int taille;
	    int j;

	    taille=sizeof(pid_t)*iter;
	    old->process=(pid_t *)realloc((void *)old->process,taille);
	    for (j=old->binding->iteration;j<iter;j++)
	      old->process[j]=(pid_t)0;
	    }
	  else {
            int j;

	    for (j=iter;j<old->binding->iteration;j++)
	      prc=prc_addtoprc(prc,old->process[j]);
            }
	  }
	old->binding->iteration=iter;
        newlst[i]->binding->iteration=0;
        }
      }
    for (i=0;prv[i]!=(LSTTYP *)0;i++) {
      if (searchprv(newlst,prv[i])==(LSTTYP *)0) {
        /*removing unneeded anymore	*/
        prc=dev_oneclose(prc,prv[i]);
        prv[i]=dev_onefree(prv[i]);
        }
      else {
	/*keeping good connection	*/
        mrglst=(LSTTYP **)vec_addlstlst((void **)mrglst,(void *)prv[i]);
	}
      }
    (void) free(prv);
    for (i=0;newlst[i]!=(LSTTYP *)0;i++) {
      if (newlst[i]->binding->iteration==0) {
        newlst[i]=dev_onefree(newlst[i]);
	}
      else {
	/*brand new connection detected	*/
        mrglst=(LSTTYP **)vec_addlstlst((void **)mrglst,(void *)newlst[i]);
	}
      }
    prc=prc_childkill(prc,5);
    (void) free(newlst);
    newlst=mrglst;
    }
  }
return newlst;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to start listening on each	*/
/*	binded channel.				*/
/*						*/
/************************************************/
int pro_listen(int reload,MXCTYP *mxc,LSTTYP **look,
	       void (*prothdlr)(MXCTYP *mxc,LSTTYP *,int))

{
#define	OPLI	"lvlpro.c:pro_listen,"

int status;

if (reload==true)
  (void) dev_mulclose(look);
if ((status=dev_mullisten(look))==true) {
  int i;

  for (i=0;look[i]!=(LSTTYP *)0;i++) {
    int j;

    if (look[i]->handle<0)
      continue;
    for (j=0;j<look[i]->binding->iteration;j++) {
      if (look[i]->process[j]!=(pid_t)0) {
        if (prc_checkprocess(look[i]->process[j])==false) {
	  (void) clm_alert(1,"lost process='%d'",look[i]->process[j]);
          look[i]->process[j]=(pid_t)0;
	  }
	}
      if (look[i]->process[j]==(pid_t)0) {
	GACTYP *gat;

	gat=(GACTYP *)0;
	switch(look[i]->process[j]=fork()) {
	  case  -1	:
	    (void) clm_alert(0,"%s Unable to fork process (error=<%s>)",
				OPLI,strerror(errno));
	    break;
	  case   0	:
	    (void) jrl_closejournals();
    	    /*module own garbage collector	*/
    	    gat=gac_open(2,appname,"pro_listen");
	    (void) gac_start(gat);
	    (void) clm_setjournal(APJRL,d_log,appname);
	    (void) clm_setjournal(EVJRL,d_log,"events");
	    (void) clm_setjournal(EVTRU,d_log,"trusting");
	    (void) prothdlr(mxc,look[i],j+1);
    	    (void) clm_trashmem(2,gat);
	    (void) sleep(1);	/*avoiding avalanche		*/
	    (void) exit(0);
	    break;
	  default	:
	    break;
	  }
	}
      }
    }
  }
return status;
}
