/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	LVLTRU:					*/
/*	Implement Mid level procedure to manage	*/
/*	banned server list received from trusted*/
/*	remote server.				*/
/*						*/
/************************************************/
#include	<string.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"unidns.h"
#include	"gesdbc.h"
#include	"lvltru.h"
/*

*/
/************************************************/
/*						*/
/*	Procedure to dump a dlg record if	*/
/*	curent debug level is above the debug	*/
/*	target					*/
/*						*/
/************************************************/
static void debugdlgs(int target,char *caller,int depth,DLGTYP **dlgs)

{
#ifdef	DEBUG
if (debug>=target) {
  char **lsts;

  (void) clm_alert(debug,"%s[%d], taille='%d'",caller,depth,vec_sizelstlst((void **)dlgs));
  if ((lsts=tru_dumpdlgs((DLGTYP **)dlgs,4))!=(char **)0) {
    int i;

    for (i=0;lsts[i]!=(char *)0;i++)
       (void) clm_alert(debug,"dlgs[%d]=<%s>",i,lsts[i]);
    lsts=(char **)vec_freelstlst((void **)lsts,VECFREE);
    }
  }
#endif
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to find out if member of one	*/
/*	record exist within a second record	*/
/*						*/
/************************************************/
static int findthesame(DLGTYP **p1,DLGTYP **p2,int taille)

{
int same;
int i;

same=false;
for (i=0;i<taille;i++) {
  DLGTYP *look;
  int j;

  look=p1[i];
  same=false;
  for (j=0;j<taille;j++) {
    if (strcmp(look->orig,p2[j]->orig)==0) {
      if (look->flgt==p2[j]->flgt)
        same=true;
      break;
      }
    }
  if (same==false)
    break;	/*one record not found		*/
  }
return same;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to find out if newone record	*/
/*	is more recent than the new one, the	*/
/*	data record is compared.		*/
/*						*/
/************************************************/
static int comparedlg(DLGTYP *oldone,DLGTYP *newone)

{
#define	OPCO	"lvltru.c:comparedlg,"

register int same;
int taille;
int phase;
int proceed;

same=false;
taille=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*are record same orig	*/
      if (strcmp(oldone->orig,newone->orig)!=0) {
        (void) clm_alert(0,"%s dglrecord not from same origin <%s!=%s> (bug?)",
			   OPCO,oldone->orig,newone->orig);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*are record good type	*/
      if (newone->flgdata!=oldone->flgdata) {
	phase=999;	/*obvioulys not the same*/
	}
      break;
    case 2	:	/*are record same siaze	*/
      if (oldone->flgdata==true) {
        if (oldone->flgt==newone->flgt)
          same=true;
	phase=999;	/*check done		*/
	}
      break;
    case 3	:	/*are record same siaze	*/
      taille=vec_sizelstlst((void **)(oldone->frwdata.frw));
      if (taille!=vec_sizelstlst((void **)(newone->frwdata.frw)))
	phase=999;	/*not the same		*/
      break;
    case 4	:	/*searching newone	*/
      if ((same=findthesame(newone->frwdata.frw,oldone->frwdata.frw,taille))==false) 
	phase=999;	/*no need to go further	*/
      break;
    case 5	:	/*searching oldone	*/
      if ((same=findthesame(oldone->frwdata.frw,newone->frwdata.frw,taille))==false) 
	phase=999;	/*no need to go further	*/
      break;
    case 6	:	/*bingo! they are same	*/
      same=true; 
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return same;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare and merge 2 streams*/
/*	info.					*/
/*						*/
/************************************************/
static STRMTY *mergestrm(STRMTY *oldstrm,char *ip,int mask,DLGTYP *data)

{
#define	OPME	"lvltru.c:mergestrm"

STRMTY *nstrm;
DLGTYP *ndata;
int phase;
int proceed;

nstrm=(STRMTY *)0;
ndata=(DLGTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*avail data?		*/
      if (data==(DLGTYP *)0)
	phase=999;	/*nothing to do		*/
      break;
    case 1	:	/*no local data?	*/
      ndata=tru_dupdlg(data);
      if (oldstrm==(STRMTY *)0) {
	nstrm=(STRMTY *)calloc(1,sizeof(STRMTY));
        nstrm->ip=strdup(ip);
	nstrm->mask=mask;	
	nstrm->data=(DLGTYP **)calloc(2,sizeof(DLGTYP *));
	nstrm->data[0]=ndata;
	phase=999;	/*record created	*/
	}
      break;
    case 2	:	/*locating old data	*/
      nstrm=tru_dupstrm(oldstrm);
      if (nstrm->data!=(DLGTYP **)0) {
	int i;

	for (i=0;nstrm->data[i]!=(DLGTYP *)0;i++) {
	  if (strcmp(nstrm->data[i]->orig,data->orig)==0) {
	    nstrm->data[i]=tru_freedlg(nstrm->data[i]);
	    nstrm->data[i]=ndata;
	    phase=999;	/*replaced		*/
	    break;
	    }
	  }
	}
      break;
    case 3	:	/*adding data		*/
      nstrm->data=(DLGTYP **)vec_addlstlst((void **)nstrm->data,(void *)ndata);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return nstrm;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract banned IP info	*/
/*						*/
/************************************************/
static DLGTYP *getcurdata(STRMTY *strm,char *orig)

{
DLGTYP *data;

data=(DLGTYP *)0;
if (strm->data!=(DLGTYP **)0) {
  int i;

  for (i=0;strm->data[i]!=(DLGTYP *)0;i++) {
    if (strcmp(strm->data[i]->orig,orig)==0) {
      data=tru_dupdlg(strm->data[i]);
      break;	/*record found			*/
      }
    }
  }
return data;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract banned IP info	*/
/*						*/
/************************************************/
static STRMTY *getcurstrm(SQLFILE *sql,char *ip,int mask)

{
#define	OPTR	"lvltru.c:getcurstrm,"

STRMTY *strm;
TRUTYP *trust;
char **info;
int phase;
int proceed;

strm=(STRMTY *)0;
trust=(TRUTYP *)0;
info=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting IP status	*/
      if ((info=dbc_getbaninfo(sql,ip))==(char **)0)
	phase=999;	/*no record		*/
      break;
    case 1	:	/*parsing info		*/
      if ((trust=tru_parsetrust("","",info))==(TRUTYP *)0) {
        (void) clm_alert(0,"%s Unable to parse <%s/%d> info=<%s.....> (data corruption?)",
			   "lvltru.c:getcurdata,",ip,mask,info[0]);
	phase=999;	/*no parsing		*/
	}
      info=(char **)vec_freelstlst((void **)info,VECFREE);
      break;
    case 2	:	/*searching in record	*/
      if (trust->strms==(STRMTY **)0) {
	(void) clm_alert(0,"%s Problem with current <%s/%d> (data-base contents corruption?)",OPTR,ip,mask);
	phase++;	/*no strm to extract	*/
	}
      break;
    case 3	:	/*extracting strm	*/
      strm=trust->strms[0];
      trust->strms[0]=(STRMTY *)0;
      break;
    case 4	:	/*cleaning trust	*/
      trust=tru_freetrusts(trust);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return strm;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check banned/std status	*/
/*	within the stream sequence.		*/
/*						*/
/************************************************/
static int finddlgstatus(DLGTYP **dlgs,int *std,int *banned,int depth)

{
int found;

found=false;
if (dlgs!=(DLGTYP **)0) {
  for (;(found==false)&&(*dlgs!=(DLGTYP *)0);dlgs++) {
    if ((*dlgs)->flgdata==true) {
      if ((strlen((*dlgs)->orig)==0)&&(depth==0))
	found=true;
      switch ((*dlgs)->frwdata.status) {
	case BANBAN	:
	  (*banned)++;
	  if (found==true)
	    (*std)=0;
	  break;
	default		:
	  (*std)++;
	  if (found==true)
	    (*banned)=0;
	  break;
	}
      continue;
      }
    found=finddlgstatus((*dlgs)->frwdata.frw,std,banned,depth+1);
    }
  }
return found;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a new DLG is	*/
/*	already within a LDG array		*/
/*	return the number of DLG within the	*/
/*	list.					*/
/*						*/
/************************************************/
static int insertsdlg(DLGTYP **lists,DLGTYP *tocheck,register int num)

{
int i;
int found;

found=false;
for (i=0;i<num;i++) {
  if (strcmp(lists[i]->orig,tocheck->orig)==0) {
    found=true;
    if (lists[i]->flgt<tocheck->flgt) {
      DLGTYP *ptr;

      ptr=lists[i];
      lists[i]=tocheck;
      tocheck=ptr;
      }
    break;
    }
  }
if (found==false) {
  lists[num]=tocheck;
  num++;
  }
else
 tocheck=tru_freedlg(tocheck);
return num;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to find out a specifi record	*/
/*	and remove them from the tree while	*/
/*	returning the most recent record time.	*/
/*						*/
/************************************************/
static time_t fdestroy(char *nameip,int status,DLGTYP **fdata,time_t stamp,int depth)

{
if (fdata!=(DLGTYP **)0) {
  int i;

  (void) debugdlgs(4,"fdestroy",depth,fdata);
  for (i=0;fdata[i]!=(DLGTYP *)0;i++) {
    DLGTYP *ptr;
    int toremove;

    ptr=fdata[i];
    toremove=false;
    if (ptr->flgdata==true) {
      if (ptr->frwdata.status!=status)
	continue;
      if (strcmp(nameip,ptr->orig)==0) {
	if (ptr->flgt>stamp)
	  stamp=ptr->flgt;
	toremove=true;
	}
      }
    else {
      stamp=fdestroy(nameip,status,ptr->frwdata.frw,stamp,depth+1);
      if (ptr->frwdata.frw[0]==(DLGTYP *)0) {
	toremove=true;
	}
      }
    if (toremove==true) {
      DLGTYP **mvr;

      mvr=fdata+i;
      (void) tru_freedlg(ptr);
      while (*mvr!=(DLGTYP *)0) {
	*mvr=*(mvr+1);
	mvr++;
	}
      }
    }
  }
return stamp;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove duplicate entry	*/
/*	sinthin a banned/trusted data stream.	*/
/*						*/
/************************************************/
static void rmduplicate(char *nameip,DLGTYP **toclean)

{
int i;

for (i=0;toclean[i]!=(DLGTYP *)0;i++) {
  if (strcmp(nameip,toclean[i]->orig)==0)
    (void) strcpy(toclean[i]->orig,"");
  }
for (i=0;toclean[i]!=(DLGTYP *)0;i++) {
  DLGTYP *p1;
  int j;
   

  p1=toclean[i];
  for (j=i+1;toclean[j];j++) {
    DLGTYP *p2;

    p2=toclean[j];
    if (strcmp(p1->orig,p2->orig)==0) {
      int k;

      if (p1->flgt<p2->flgt)
	p1->flgt=p2->flgt;
      k=j;
      do {
	toclean[k]=toclean[k+1];
	k++;
	}
      while (toclean[k]!=(DLGTYP *)0);
      j--;
      p2=tru_freedlg(p2);
      }
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to compact entry		*/
/*	return the entry side			*/
/*						*/
/************************************************/
static int compacted(DLGTYP **fdata)

{
int i;

for (i=0;fdata[i]!=(DLGTYP *)0;i++) {
  if (fdata[i]->flgdata==false) { /*always	*/
    DLGTYP **ptr;
    
    ptr=fdata[i]->frwdata.frw;
    switch (vec_sizelstlst((void **)ptr)) {
      case 0	:	/*empty	*/
	if (fdata[i]!=(DLGTYP *)0) {	/*always*/
	  int k;

	  k=i;
	  (void) tru_freedlg(fdata[k]);
	  do {
	    fdata[k]=fdata[k+1];
	    k++;
	    }
	  while (fdata[k]!=(DLGTYP *)0);
	  }
	break;
      case 1	:	/*one record, data?	*/
	if (ptr[0]->flgdata==true) {
	  DLGTYP *toc;

	  toc=ptr[0];
	  if ((strlen(toc->orig)==0)||(strcmp(toc->orig,fdata[i]->orig)==0)) {
	    fdata[i]->flgdata=true;
	    fdata[i]->frwdata.status=toc->frwdata.status;
	    ptr[0]=tru_freedlg(ptr[0]);
	    (void) free(ptr);
	    } 
	  }
	break;
      default	:	/*too big to be compacted	*/
	break;
      }
    }
  }
return vec_sizelstlst((void **)fdata);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to locate all similare entry	*/
/*	below that level destroy them.		*/
/*						*/
/************************************************/
static void getmostrecent(char *nameip,DLGTYP **pdata,DLGTYP **fdata)

{
int i;

for (i=0;pdata[i]!=(DLGTYP *)0;i++) {
  char *ptr;
  
  ptr=nameip;
  if (strlen(pdata[i]->orig)>0)
    ptr=pdata[i]->orig;
  pdata[i]->flgt=fdestroy(ptr,pdata[i]->frwdata.status,fdata,pdata[i]->flgt,1);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the deepest status	*/
/*	available in the tree pertaining to	*/
/*	a specific name ip. Move up the status	*/
/*	and remove it from thetree.		*/
/*						*/
/************************************************/
static DLGTYP *findeepest(char *nameip,DLGTYP **toclean)

{
DLGTYP *deepest;

deepest=(DLGTYP *)0;
if (toclean!=(DLGTYP **)0) {	/*always	*/
  int i;

  for (i=0;toclean[i]!=(DLGTYP *)0;i++) {
    DLGTYP *canbe;

    canbe=toclean[i];
    if (canbe->flgdata==true) {
      if (strcmp(nameip,canbe->orig)==0) {
	int k;

	k=i;
	if (deepest!=(DLGTYP *)0) {
	  if (deepest->flgt<canbe->flgt) 
	    deepest=tru_freedlg(deepest);
	  else  {
	    canbe=tru_freedlg(canbe);
	    canbe=deepest;
	    }
	  }
	deepest=canbe;
	do {
	  toclean[k]=toclean[k+1];
	  k++;
	  }
	while (toclean[k]!=(DLGTYP *)0);
	break;	/*no need to search further	*/
	}
      }
    else
      if ((deepest=findeepest(nameip,canbe->frwdata.frw))!=(DLGTYP *)0)
	break;	/*no need to search further	*/
    }
  }
return deepest;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to read DLGTYP structure and	*/
/*	optimize contents			*/
/*						*/
/************************************************/
static DLGTYP **optimizesdlgs(char *nameip,DLGTYP **toclean)

{
#define	OPOP	"lvltru.c:optimizesdlg,"
int phase;
int proceed;

phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*do we have data?	*/
      if (toclean==(DLGTYP **)0) {
	(void) clm_alert(0,"%s Unexpected empty dlgrecord to clean (bug?)",OPOP);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*find deepest status	*/
      if (toclean!=(DLGTYP **)0){	/*always*/
	DLGTYP *deep;

	if ((deep=findeepest(nameip,toclean))!=(DLGTYP *)0)
	  toclean=(DLGTYP **)vec_addlstlst((void **)toclean,(void *)deep);
	}
      break;
    case 2	:	/*compressing data	*/
      if (toclean!=(DLGTYP **)0){	/*always*/
	int i;

	for (i=0;toclean[i]!=(DLGTYP *)0;i++) {
	  if (toclean[i]->flgdata==true)
	    continue;
	  if (toclean[i]->frwdata.frw==(DLGTYP **)0) {
	    (void) clm_alert(0,"%s Unexpected empty data-record for index '%d', orig=<%s>",
			       OPOP,i,toclean[i]->orig);
	    continue;
	    }
	  toclean[i]->frwdata.frw=optimizesdlgs(toclean[i]->orig,toclean[i]->frwdata.frw);
	  if (strcmp(nameip,toclean[i]->orig)==0) {
	    DLGTYP *ref;
	    int j;

	    ref=toclean[i];
	    for (j=i;toclean[j]!=(DLGTYP *)0;j++)
	      toclean[j]=toclean[j+1];
	    for (j=0;ref->frwdata.frw[j]!=(DLGTYP *)0;j++)
	      toclean=(DLGTYP **)vec_addlstlst((void **)toclean,(void *)ref->frwdata.frw[j]);
	    (void) free(ref->frwdata.frw);
	    ref->frwdata.frw=(DLGTYP **)0;
	    ref=tru_freedlg(ref);
	    i--;	/*lets start again	*/
	    }
	  }
	}
      break;
    case 3	:	/*ordering data		*/
      if (toclean!=(DLGTYP **)0){	/*always*/
	register int taille;
	register int p;	/*pdata current index	*/
	DLGTYP **pdata;	/*pure data		*/
	register int f;	/*pdata current index	*/
	DLGTYP **fdata;	/*forwarded data	*/
	register DLGTYP **ptr;
	register int i;

	taille=vec_sizelstlst((void **)toclean);	
	p=0;
	f=0;
	pdata=(DLGTYP **)calloc(taille+1,sizeof(DLGTYP *));
	fdata=(DLGTYP **)calloc(taille+1,sizeof(DLGTYP *));
	for (i=0,ptr=toclean;i<taille;i++,ptr++) {
	  if ((*ptr)->flgdata==true)  {
	    p=insertsdlg(pdata,*ptr,p);
	    }
	  else
	    {
	    fdata[f]=*ptr;
	    f++;
	    }
	  }
	(void) memset(toclean,'\000',sizeof(DLGTYP *)*taille);
	if (p>0) {
	  getmostrecent(nameip,pdata,fdata);
	  (void) memcpy(toclean,pdata,sizeof(DLGTYP *)*p);
	  }
	if ((f=compacted(fdata))>0) 
	  (void) memcpy(toclean+p,fdata,sizeof(DLGTYP *)*f);
	(void) free(fdata);
	(void) free(pdata);
	(void) rmduplicate(nameip,toclean); 
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return toclean;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to optimize one DLGTYP only.	*/
/*						*/
/************************************************/
static DLGTYP *optimizesdlg(char *nameip,DLGTYP *toclean)

{
DLGTYP **dlgs;

dlgs=(DLGTYP **)calloc(2,sizeof(DLGTYP *));
dlgs[0]=toclean;
dlgs=optimizesdlgs(nameip,dlgs);
toclean=dlgs[0];
(void) free(dlgs);
return toclean;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract current banned	*/
/*	status and update data-base		*/
/*	accordingly.				*/
/*	Return true if record have been updated	*/
/*						*/
/************************************************/
static char *trumngdbd(SQLFILE *sql,char *myname,char *myip,STRMTY *strm)

{
#define	OPTU	"lvltru.c:trumngdbd,"

char *report;
DLGTYP *olddata;
DLGTYP *newdata;
STRMTY *oldstrm;
STRMTY *newstrm;
char **infos;
char *info;
char *nameip;
int newstat;
int phase;
int proceed;

report=(char *)0;
olddata=(DLGTYP *)0;
newdata=(DLGTYP *)0;
oldstrm=(STRMTY *)0;
newstrm=(STRMTY *)0;
infos=(char **)0;
info=(char *)0;
(void) asprintf(&nameip,"%s/[%s]",myname,myip);
newstat=BANSTD;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*check received record	*/
      if (strm==(STRMTY *)0) {
	(void) clm_alert(0,"%s Unexpected empty stream input (bug?)",OPTU);
	phase=999;	/*no need to go further	*/
	}
      break;
    case 1	:	/*check received record	*/
      newdata=tru_dupdlg(strm->data[0]);
      if ((newdata=optimizesdlg(nameip,newdata))==(DLGTYP *)0) {
	(void) clm_alert(0,"%s no data in stream record (bug?)",OPTU);
	phase=999;	/*no need to go further	*/
	}
      break;
    case 2	:	/*getting current stream*/
      if ((oldstrm=getcurstrm(sql,strm->ip,strm->mask))!=(STRMTY *)0) {
        if ((olddata=getcurdata(oldstrm,newdata->orig))!=(DLGTYP *)0)
          olddata=optimizesdlg(nameip,olddata);
        if (oldstrm->mask!=strm->mask) {
	  (void) asprintf(&report,"IP mask mismatch for banned IP <%s/%d>, droping submission",
				   strm->ip,strm->mask);
	  phase=999;	/*ban IP mask not same	*/
	  }
	}
      break;
    case 3	:	/*comparing old and new	*/
      if (olddata!=(DLGTYP *)0) {
        if (comparedlg(olddata,newdata)==true) 
	  phase=999;	/*no need to go further	*/
        }
      break;
    case 4	:	/*comparing old and new	*/
      if ((newstrm=mergestrm(oldstrm,strm->ip,strm->mask,newdata))==(STRMTY *)0) {
	(void) clm_alert(0,"%s Unable to merge <%s/%d> (data-base contents corruption?)",
			    OPTU,oldstrm->ip,oldstrm->mask);
	phase=999;	/*merge empty?		*/
	}
      break;
    case 5	:	/*set banned status	*/
      if (newstrm!=(STRMTY *)0) {	/*always*/
	int std;
	int banned;

	std=0;
	banned=0;
	(void) finddlgstatus(newstrm->data,&std,&banned,0);
	if (banned>std)
	  newstat=BANBAN;
	}
      break;
    case 6	:	/*extracting new infos	*/
      if (newstrm!=(STRMTY *)0) {/*always	*/
        if ((infos=tru_dumpstrm(myname,myip,newstrm))==(char **)0) {
	  (void) clm_alert(0,"%s, No stream to dump (Bug?)",OPTU);
	  phase=999;	/*getting out now	*/
	  }
        newstrm=tru_freestrm(newstrm);
        }
      break;
    case 7	:	/*triming data contents	*/
      if (infos!=(char **)0) { 	/*always	*/
	int num;
	
	if ((num=vec_sizelstlst((void **)infos))>2) {
	  infos[0]=clm_freestr(infos[0]);
	  infos[1]=clm_freestr(infos[1]);
	  (void) memmove(infos,infos+2,(num-1)*sizeof(char *));
	  info=clm_mergestring(infos,'\n');
	  }
        else 
	  (void) clm_alert(0,"%s, No stream data found (Feed default?)",OPTU);
        infos=(char **)vec_freelstlst((void **)infos,VECFREE);
	}
      break;
    case 8	:	/*Updating database	*/
      if (info!=(char *)0) {
        (void) dbc_updban(sql,strm->ip,strm->mask,newstat,info);
        info=clm_freestr(info);
	}
      break;
    default	:	/*SAFE Guard		*/
      oldstrm=tru_freestrm(oldstrm);
      olddata=tru_freedlg(olddata);
      newdata=tru_freedlg(newdata);
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(nameip);
return report;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to update database bantrust	*/
/*	table, according report from friendly	*/
/*	and trusted remote server.		*/
/*						*/
/************************************************/
char *tru_upddb(SQLFILE *sql,TRUTYP *trust)

{
char *report;

report=(char *)0;
if ((trust!=(TRUTYP *)0)&&(trust->strms!=(STRMTY **)0)) {
  char *myname;
  char *myip;
  int i;

  myname=dns_getmyname();
  myip=dns_getfirstip(myname);
  for (i=0;trust->strms[i]!=(STRMTY *)0;i++) {
    char *str;

    if ((str=trumngdbd(sql,myname,myip,trust->strms[i]))!=(char *)0) {
      char *sep;
      char *old;

      sep="\n";
      old=report;
      if (old==(char *)0) {
	sep="";
	old=strdup("");
	}
      (void) asprintf(&report,"%s%s%s",old,sep,str);
      (void) free(old);
      (void) free(str);
      }
    }
  myip=clm_freestr(myip);
  myname=clm_freestr(myname);
  }
return report;
}
