/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	MODCKR:					*/
/*	Module doing mail checking process	*/
/*						*/
/************************************************/
#include	<errno.h>
#include	<math.h>
#include	<string.h>

#include	"lowtyp.h"
#include	"utlprc.h"
#include	"utlvec.h"
#include	"subprc.h"
#include	"subclm.h"
#include	"bascnv.h"
#include	"unieml.h"
#include	"unidns.h"
#include	"unisql.h"
#include	"unijrl.h"
#include	"gesspf.h"
#include	"gestls.h"
#include	"gesbls.h"
#include	"geseml.h"
#include	"gestru.h"
#include	"gesque.h"
#include	"lvlipc.h"
#include	"lvlcnf.h"
#include	"lvlspm.h"
#include	"lvleml.h"
#include	"lvlsta.h"
#include	"lvlbls.h"
#include	"lvlckr.h"
#include	"lvltrt.h"
#include	"lvltru.h"
#include	"modckr.h"

typedef	struct	{	/*working modckr struct	*/
	time_t grabed;	/*last collected time	*/
	CNFTYP *cnf;	/*configuration struct	*/
	STATYP *stats;	/*statistics collection */
	SQLFILE *sql;	/*acces to data-base	*/
	CLAMTYP *clamav;/*virus DB reference	*/
	SPAMTYP *spamd;	/*spamassassin scanner	*/
	BPTR *bptr;	/*reference to IPC box	*/
	u_long mstpool;	/*master pool number	*/
	u_long mntpool;	/*mentor pool number	*/
	}WRKTYP;

typedef	struct	{	/*Sorting and dispatch	*/
	int status;	/*Channel status line	*/
	char *smtpmx;	/*MX to use to send	*/
	char *smtpport;	/*Port to use to send	*/
	char *from;	/*email sender		*/
	RCPTYP **rcpts;	/*dispatched user	*/
	}AIGTYP;


typedef	struct	{	/*dispatching struct	*/
	char *efrom;	/*SMTP from		*/
	char *inbehalf;	/*sent inbehalf origin	*/
	char *rejectto;	/*where to send reject	*/
	RCPTYP **rcpts;	/*recipient data	*/
	}DISTYP;
/*

*/
/************************************************/
/*						*/
/*	Procedure to set the same status to an	*/
/*	RCPT list.				*/
/*						*/
/************************************************/
static void alignstatus(RCPTYP **rcpts,int status)

{
if (rcpts!=(RCPTYP **)0) {
  int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++)
    rcpts[i]->status=status;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if an email address	*/
/*	was already within list.		*/
/*						*/
/************************************************/
static int alreadyseen(char **seen,char *email)

{
int already;

already=false;
if (seen!=(char **)0) {
  int i;

  for (i=0;seen[i]!=(char *)0;i++) {
    if (strcmp(seen[i],email)==0) {
      already=true;
      break;		/*found it		*/
      }
    }
  }
return already;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	sorted dispatched user list		*/
/*						*/
/************************************************/
static AIGTYP **freeaig(AIGTYP **aig)

{
if (aig!=(AIGTYP **)0) {
  int i;

  for (i=0;aig[i]!=(AIGTYP *)0;i++) {
    aig[i]->rcpts=eml_freercpts(aig[i]->rcpts);
    aig[i]->smtpport=clm_freestr(aig[i]->smtpport);
    aig[i]->smtpmx=clm_freestr(aig[i]->smtpmx);
    aig[i]->from=clm_freestr(aig[i]->from);
    (void) free(aig[i]);
    }
  (void) free(aig);
  aig=(AIGTYP **)0;
  }
return aig;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if mail need to be	*/
/*	store in quarantine area.		*/
/*						*/
/************************************************/
static int tobequarantined(SQLFILE *sql,DSPTYP *dsp,USRTYP *usr,char **note)

{
register int quarantined;
PRFTYP *prf;
char locnot[100];
int phase;
int proceed;

quarantined=false;
prf=(PRFTYP *)0;
(void) strcpy(locnot,"");
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*global spam limit	*/
      if (dsp->spmlvl>dsp->spmlim) {
 	(void) snprintf(locnot,sizeof(locnot),"%s: ratio=%.0f%%","Default Spam-lvl",(dsp->spmlvl/dsp->spmlim)*100);
	quarantined=true;
	}
      break;
    case 1	:	/*mail to external user	*/
      if (usr->prfid==0) 
	phase=999;	/*no need to go further	*/
      break;
    case 2	:	/*user profile		*/
      if ((prf=dbc_getprofile(sql,usr->prfid))==(PRFTYP *)0)
	phase=999;	/*trouble trouble?	*/
      break;
    case 3	:	/*user spam level	*/
      if (isnormal(prf->spamlim)==true) {
        (void) strcpy(locnot,"");
        quarantined=false;
        if (dsp->spmlvl>prf->spamlim) {
 	  (void) snprintf(locnot,sizeof(locnot),"%s: %.2f, ratio=%.0f%%","Profile Spam-limit",
						prf->spamlim,(dsp->spmlvl/prf->spamlim)*100);
          quarantined=true;
	  }
        }
      prf=dbc_freeprofile(prf);
      break;
    default	:	/*SAFE Guard		*/
      *note=clm_freestr(*note);
      *note=strdup(locnot);
      proceed=false;
      break;
    }
  phase++;
  }
return quarantined;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add a message if an error 	*/
/*	in processing occur, letting answ with	*/
/*	a null value.				*/
/*						*/
/************************************************/
static char *is_errproc(char *answ,const char *fmt,...)

{
if (answ==(char *)0) {
  va_list args;
  char *strloc;

  va_start(args,fmt);
  vasprintf(&strloc,fmt,args);
  answ=eml_ascstat(ERRPROC,strloc);
  (void) free(strloc);
  va_end(args);
  }
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add one stat to stats	*/
/*	collection				*/
/*						*/
/************************************************/
static void addstat(WRKTYP *wrk,char *fromgrp,char *fromdom,char *from,
		    u_long *in,u_long *out,char *msg)

{
#define	OPAD	"modckr.c:addsta,"

MENTYP mstatus;
char *ptr;
char *rcpt;
char *rcptgrp;
char *rcptdom;
char *dfltgrpname;
char *dfltfromdom;
GRPTYP *grp;
int phase;
int proceed;

mstatus=m_undefined;
ptr=(char *)0;
rcpt=(char *)calloc(strlen(msg),sizeof(char));
rcptgrp=(char *)0;
rcptdom=(char *)0;
dfltgrpname=(char *)0;
dfltfromdom=(char *)0;
grp=(GRPTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch(phase) {
    case 0	:	/*looking for data	*/
      if ((ptr=strchr(msg,' '))==(char *)0) {
	(void) clm_alert(0,"%s missing space in msg <%s> (bug?)",OPAD,msg);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*extracting data	*/
      ptr++;
      if (sscanf(ptr,"%d %s\n",(int *)&mstatus,rcpt)!=2) {
	(void) clm_alert(0,"%s unable to scan msg <%s> (bug?)",OPAD,ptr);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*getting the rpct data	*/
      if ((grp=dbc_getgroup(wrk->sql,rcpt))!=(GRPTYP *)0) {
        rcptgrp=strdup(grp->grpname);
	rcptdom=eml_getdomain(rcpt);
        grp=dbc_freegroup(grp);
	}
      break;
    case 3	:	/*no group at all?	*/
      if ((rcptgrp==(char *)0)&&(fromgrp==(char *)0)) {
	dfltgrpname=dbc_getdfltgrpname(wrk->sql);
        fromgrp=dfltgrpname;
	if (fromdom==(char *)0) {
	  dfltfromdom=dbc_getdfltdom(wrk->sql);
	  fromdom=dfltfromdom;
	  }
        if (fromgrp==(char *)0) {
	  (void) clm_alert(0,"%s originator and recipient are both unknown group rcpt=<%s>,from=<%s>",OPAD,rcpt,from);
	  phase=999;	/*trouble trouble	*/
	  }
	}
      break;
    case 4	:	/*adding all data	*/
      if (fromgrp==(char *)0) 
        wrk->stats=sta_collationne(wrk->stats,rcptgrp,"extern",mstatus,*in,*out);
      else {
        wrk->stats=sta_collationne(wrk->stats,fromgrp,fromdom,mstatus,*in,*out);
        wrk->stats=sta_collationne(wrk->stats,fromgrp,ALLDOM,mstatus,*in,*out);
	}
      if (rcptgrp==(char *)0) 
        wrk->stats=sta_collationne(wrk->stats,fromgrp,"intern",mstatus,*in,*out);
      else {
        wrk->stats=sta_collationne(wrk->stats,rcptgrp,rcptdom,mstatus,*in,*out);
        wrk->stats=sta_collationne(wrk->stats,rcptgrp,ALLDOM,mstatus,*in,*out);
	}
      wrk->stats=sta_collationne(wrk->stats,ALLDOM,ALLDOM,mstatus,*in,*out);
      switch (mstatus) {
	case m_toresend	:
	case m_direct	:
	  *in=(u_long)0; 
	  break;
	case m_relayed	:
	  *out=(u_long)0; 
	  break;
	default		:	/*No action	*/
	  break;
	}
      (void) clm_freestr(rcptgrp);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
dfltfromdom=clm_freestr(dfltfromdom);
dfltgrpname=clm_freestr(dfltgrpname);
rcpt=clm_freestr(rcpt);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add an entry within the	*/
/*	override data-base for each relayed	*/
/*	email, such if a reply is sent, it will	*/
/*	accepted regardless.			*/
/*						*/
/************************************************/
static void forceoverride(SQLFILE *sql,char *from,char *to)

{
#define	OPFO	"modckr.c:forceoverride,"

char *orgname;
char *orgdom;
int phase;
int proceed;

orgname=strdup(to);
orgdom=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*extracting orgdom	*/
      if ((orgdom=strchr(orgname,'@'))==(char *)0) {
        (void) clm_alert(0,"%s orgname=<%s> format incorrect (bug?)",OPFO,orgname);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*updating overrides	*/
      *orgdom='\000';
      orgdom++;
      if (dbc_setoverride(sql,from,orgname,orgdom,"0.0.0.0",0,tr_accepted)==false) {
        (void) clm_alert(0,"%s Unable to set override recipient <%s> origin=<%s> (bug?)",
			   OPFO,from,to);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*recovering quarantine	*/
      (void) dbc_purgequar(sql,to,from);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(orgname);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add aliases information	*/
/*	to sorting data.			*/
/*						*/
/************************************************/
static RCPTYP **expendalias(SQLFILE *sql,char *locip,char *efrom,RCPTYP *alias,int auth,char ***seen)

{
#define	OPEX	"modckr.c:expendalias,"

RCPTYP **rcpts;

rcpts=(RCPTYP **)0;
if ((alias!=(RCPTYP *)0)&&(alias->rcpt!=(char *)0)) {	/*always		*/
  int dfltgrpid;
  char **aliases;

  dfltgrpid=dbc_getgroupid(sql,alias->rcpt);
  if ((aliases=dbc_getaliaslist(sql,alias->rcpt))!=(char **)0) {
    int i;
    const char *discarded;

    
    discarded=eml_strmntstatus(m_discarded);
    for (i=0;aliases[i]!=(char *)0;i++) {
      USRTYP *usr;

      if (alreadyseen(*seen,aliases[i])==true)
	continue;
      if ((strchr(aliases[i],'@')==(char *)0)&&(strchr(alias->rcpt,'@')!=(char *)0)) {
	char *newal;

	(void) asprintf(&newal,"%s@%s",aliases[i],eml_getdomain(alias->rcpt));
	(void) free(aliases[i]);
        aliases[i]=newal;
	}
      usr=ckr_getuserinfo(sql,aliases[i],locip);
      if (usr==(USRTYP *)0) {
	(void) clm_alert(0,"%s unable to getuserinfo for  <%s> extracted from <%s> (bug?)",
				OPEX,aliases[i],alias->rcpt);
        continue;
	}
      else {
	MENTYP m_type;

        m_type=m_undefined;
	if (usr->grpid==0)
	  usr->grpid=dfltgrpid;
	switch (usr->rcptflg) {
      	  case rcpt_notchecked	:	/*remote not responding!	*/
      	  case rcpt_forward	:	/*mail to relay			*/
	    m_type=m_relayed;
	    break;
      	  case rcpt_islocalias	:	/*local alias			*/
      	  case rcpt_islocal	:	/*mail to store			*/
	    m_type=m_acceptable;
	    if (auth!=false)
	      m_type=m_direct;
	    break;
      	  case rcpt_isalias	:	/*aliasing name			*/
	    (void) clm_alert(0,"%s Unexpected alias found <%s> extracted from <%s> (bug?)",
				OPEX,aliases[i],alias->rcpt);
	    break;
      	  case rcpt_resend	:	/*mail to resend		*/
	    m_type=m_toresend;
	    break;
	  case rcpt_foreign	:	/*external email		*/
            (void) forceoverride(sql,efrom,aliases[i]);
	    m_type=m_relayed;
	    break;
      	  case rcpt_mailtb	:	/*email rouble	*/
  	    (void) jrl_journalling(false,EVCKR,"%*s  - %-10s <%s> Mailbox full",
						2,"",discarded,aliases[i]);
	    break;
      	  case rcpt_nomx	:	/*email rouble	*/
  	    (void) jrl_journalling(false,EVCKR,"%*s  - %-10s <%s> Unknown MX",
						2,"",discarded,aliases[i]);
	    break;
      	  case rcpt_inactive	:	/*inactive email*/
  	    (void) jrl_journalling(false,EVCKR,"%*s  - %-10s <%s> Inactive account",
					  	2,"",discarded,aliases[i]);
	    break;
      	  case rcpt_unknown	:	/*unknown email*/
  	    (void) jrl_journalling(false,EVCKR,"%*s  - %-10s <%s> Unknown email",
						2,"",discarded,aliases[i]);
	    break;
	  default		:	/*building list	*/
	    (void) jrl_journalling(true,EVCKR,"Going to core Dump from <%s> (pid='%05d'), "
					       "Email <%s> unexped rcptflg value '%d'",
						OPEX,getpid(),usr->username,usr->rcptflg);
	    (void) clm_core_dump("%s Unexpected <%s> rcptflg status '%d'",
				  OPEX,getpid(),usr->username,usr->rcptflg);
	    break;
	  }
	if (m_type!=m_undefined) {
	  RCPTYP *neu;

	  neu=(RCPTYP *)calloc(1,sizeof(RCPTYP));
	  neu->rcpt=strdup(usr->username);
	  neu->status=m_type;
	  neu->grpid=usr->grpid;
	  rcpts=(RCPTYP **)vec_addlstlst((void **)rcpts,(void *)neu);
	  *seen=(char **)vec_addlstlst((void **)(*seen),(void *)strdup(aliases[i]));
	  }
        usr=dbc_freeuserinfo(usr);
	}
      }
    aliases=(char **)vec_freelstlst((void **)aliases,VECFREE);
    }
  }
return rcpts;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to update user status		*/
/*	aliases.				*/
/************************************************/
static void updatestatus(SQLFILE *sql,DSPTYP *dsp,RCPTYP *rcp)

{
#define 	OPUP	"modckr.c:updatestatus,"

USRTYP *usr;
char *from;
char *note;
int phase;
int proceed;

usr=(USRTYP *)0;
note=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting from		*/
      from=dsp->hfrom;
      if (from==(char *)0)
	from=dsp->efrom;
      if (from==(char *)0) {
	(void) clm_alert(0,"%s Unable to find from on session <%s>",OPUP,dsp->session);
	phase=999;	/*getting out fast		*/
	}
      break;
    case 1	:	/*getting user status	*/
      if ((usr=ckr_finduserstatus(sql,from,rcp->rcpt,dsp->peerip,dsp->locip,dsp->auth,&note))==(USRTYP *)0) {
	(void) clm_alert(0,"%s we should be able to get a status sessions <%s> recipient <%s>",
			    OPUP,dsp->session,rcp->rcpt);
	phase=999;	/*getting out fast		*/
	}
      break;
    case 2	:	/*set the action status	*/
      rcp->note=clm_freestr(rcp->note);
      rcp->note=note;
      switch (usr->rcptflg) {
	case rcpt_resend	:
	  rcp->status=m_toresend;
	  break;
	case rcpt_forward	:
	  rcp->status=m_toforward;
	  break;
	case rcpt_forcefwd	:
	  rcp->status=m_mustfwd;
	  break;
	case rcpt_foreign	:
	  rcp->status=m_relayed;
	  if (tobequarantined(sql,dsp,usr,&(rcp->note))==true)
	    rcp->status=m_quarantine;
	  break;
	case rcpt_islocal	:
	case rcpt_islocalias	:
	  rcp->status=m_acceptable; 
	  if (tobequarantined(sql,dsp,usr,&(rcp->note))==true)
	    rcp->status=m_quarantine;
	  if (dsp->auth==true)
	    rcp->status=m_direct; 
	  break;
	case rcpt_banned	:
	  rcp->status=m_quarantine;
	  break;
	case rcpt_alidirect	:
	case rcpt_direct	:
	  rcp->status=m_acceptable;
	  if (dsp->auth==true)
	    rcp->status=m_direct;
	  break;
	default			:
	  (void) clm_alert(0,"%s session <%s> unexpected <%s> status='%d'",
			      OPUP,dsp->session,rcp->rcpt,usr->rcptflg);
	  break;
	}
      break;
    case 3	:	/*freeing memory	*/
      usr=dbc_freeuserinfo(usr);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do one rcpt advise		*/
/*	within data base			*/
/*						*/
/************************************************/
static int advise(SQLFILE *sql,char *peerip,char *hfrom,RCPTYP *rcpt)

{
register int myadv;
register OVRENUM override;

myadv=true;
override=ckr_getoverride(sql,hfrom,rcpt->rcpt,peerip);
switch (override) {
  case tr_banned	:	/*user to be rejected	*/
    rcpt->status=m_quarantine;
    (void) jrl_journalling(false,EVCKR,"Quarantine Override from <%s> for <%s>",hfrom,rcpt->rcpt);
    break;
  case tr_bypass	:	/*user must be a bypass	*/
    (void) jrl_journalling(false,EVCKR,"Direct Override from <%s> for <%s>",hfrom,rcpt->rcpt);
    rcpt->status=m_direct;
    break;
  case tr_accepted	:	/*user to be accepted	*/
    (void) jrl_journalling(false,EVCKR,"Accepted Override from <%s> for <%s>",hfrom,rcpt->rcpt);
    rcpt->status=m_acceptable;
    break;
  case tr_maybe		:	/*no advice to give	*/
    myadv=false;		
    break;
  }
return myadv;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to give advise about user, 	*/
/*	check about special override or spam	*/
/*	detection				*/
/*						*/
/************************************************/
static void ckradvise(SQLFILE *sql,DSPTYP *dsp,RCPTYP *rcp)

{
#define	OPCR	"modckr.c:ckradvise,"

char *from;
char *note;
USRTYP *usr;
int phase;
int proceed;

from=(char *)0;
note=(char *)0;
usr=(USRTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*special directive	*/
      if (rcp->status==m_indata) {
        if (advise(sql,dsp->peerip,dsp->hfrom,rcp)==true) 
	  rcp->smtpstatus=RETADV;
	}
      break;
    case 1	:	/*need to check spam lvl*/
      switch (rcp->status) {
	case m_indata	 :
	  rcp->status=m_acceptable;
	  rcp->smtpstatus=RETADV;
	  break;
	case m_relayed	 :
	  break;	
	case m_acceptable:/*already accepted	*/
	case m_toresolv	 :
        default		 :/*nothing to do	*/
	  phase=999;	  /*not going further	*/
	  break;
	}
      break;
    case 2	:	/*getting user status	*/
      from=dsp->hfrom;
      if (from==(char *)0)
	from=dsp->efrom;
      if ((usr=ckr_finduserstatus(sql,from,rcp->rcpt,dsp->peerip,dsp->locip,dsp->auth,&note))==(USRTYP *)0) {
	(void) clm_alert(0,"%s we should be able to get a status sessions <%s> recipient <%s>",
			    OPCR,dsp->session,rcp->rcpt);
	phase=999;	/*getting out fast		*/
	}
      break;
    case 3	:	/*getting user spamlim	*/
      rcp->note=clm_freestr(rcp->note);
      rcp->note=note;
      if (tobequarantined(sql,dsp,usr,&(rcp->note))==true) {
	rcp->status=m_quarantine;
	rcp->smtpstatus=RETADV;
	}
      usr=dbc_freeuserinfo(usr);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add a destination to a	*/
/*	sorted entry.				*/
/*						*/
/************************************************/
static int addsort(AIGTYP ***aig,char *smtpmx,char *smtpport,char *from,const RCPTYP *rcpt)

{
AIGTYP *aigloc;

aigloc=(AIGTYP *)0;
if ((*aig)!=(AIGTYP **)0) {
  int j;

  for (j=0;(*aig)[j]!=(AIGTYP *)0;j++) {
    register AIGTYP *ptr;

    ptr=(*aig)[j];
    if (ptr->status==rcpt->status) {
      if ((strcmp(ptr->smtpmx,smtpmx)==0)&&strcmp(ptr->smtpport,smtpport)==0) {
        aigloc=ptr;
        break;
	}
      }
    }
  }
if (aigloc==(AIGTYP *)0) {
  aigloc=(AIGTYP *)calloc(1,sizeof(AIGTYP));
  aigloc->from=strdup(from);
  aigloc->smtpmx=strdup(smtpmx);
  aigloc->smtpport=strdup(smtpport);
  aigloc->status=rcpt->status;
  *aig=(AIGTYP **)vec_addlstlst((void **)*aig,(void *)aigloc);
  }
aigloc->rcpts=(RCPTYP **)vec_addlstlst((void **)aigloc->rcpts,(void *)eml_duprcpt(rcpt));
return 1;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to store Email in to relay 	*/
/*	queue					*/
/*	NOTE:					*/
/*		procedure with future provision	*/
/*		to set outip according the 	*/
/*		email originator domain.	*/
/*						*/
/************************************************/
static u_int torelay(SQLFILE *sql,DSPTYP *dsp,AIGTYP *aig)

{
char *outip;
char *infile;
u_int status;

outip=ANYADDR;
(void) asprintf(&infile,"%s/%s",clm_appdir(d_queue),dsp->session);
status=trt_transit(sql,outip,aig->smtpmx,aig->smtpport,aig->from,dsp,infile,aig->rcpts);
(void) free(infile);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	dispatch according origin list		*/
/*						*/
/************************************************/
static DISTYP **freedispatchorigin(DISTYP **disporig)

{
if (disporig!=(DISTYP **)0) {
  int i;

  for (i=0;disporig[i]!=(DISTYP *)0;i++) {
    disporig[i]->rcpts=eml_freercpts(disporig[i]->rcpts);
    disporig[i]->efrom=clm_freestr(disporig[i]->efrom);
    disporig[i]->inbehalf=clm_freestr(disporig[i]->inbehalf);
    disporig[i]->rejectto=clm_freestr(disporig[i]->rejectto);
    (void) free(disporig[i]);
    }
  (void) free(disporig);
  disporig=(DISTYP **)0;
  }
return disporig;
}

/*

*/
/************************************************/
/*						*/
/*	Procedure to sort user list to dispatch	*/
/*						*/
/************************************************/
static DISTYP **getdispatchorigin(SQLFILE *sql,DSPTYP *dsp)

{
#define	OPTD	"modckr.c:getdispatchorigin,"

DISTYP **displst;
DISTYP *disploc;
char **seen;
int phase;
int proceed;

displst=(DISTYP **)0;
disploc=(DISTYP *)0;
seen=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*do we have data	*/
      if ((dsp==(DSPTYP *)0)||(dsp->rcpts==(RCPTYP **)0)) {
	(void) clm_alert(0,"%s, No data available (bug?)",OPTD);
	proceed=false;
	}
      break;
    case 1	:	/*get std recipient	*/
      disploc=(DISTYP *)calloc(1,sizeof(DISTYP));
      disploc->efrom=strdup(dsp->efrom);
      disploc->inbehalf=strdup(dsp->efrom);
      (void) asprintf(&(disploc->rejectto),"%s",dsp->efrom);
      displst=(DISTYP **)vec_addlstlst((void **)0,(void *)disploc);
      break;
    case 2	:	/*Looking for std user	*/
      if (dsp->rcpts!=(RCPTYP **)0) {/*always	*/
	int i;

	for (i=0;dsp->rcpts[i]!=(RCPTYP *)0;i++) {
	  register RCPTYP *rcp;

	  rcp=dsp->rcpts[i];
	  switch (rcp->status) {
	    case m_toresolv	:/*Alias	*/
	      /*to be done after standard rcpt	*/
	      break;
	    default		:/*std name	*/
	      if (alreadyseen(seen,rcp->rcpt)==false) {
	  	RCPTYP *add;

	        add=(RCPTYP *)calloc(1,sizeof(RCPTYP));
	        add->rcpt=strdup(rcp->rcpt);
	        add->status=rcp->status;
	        add->grpid=dbc_getgroupid(sql,add->rcpt);
	        if (rcp->why!=(char *)0)
      		  add->why=strdup(rcp->why);
	        displst[0]->rcpts=(RCPTYP **)vec_addlstlst((void **)displst[0]->rcpts,(void *)add);
	        seen=(char **)vec_addlstlst((void **)seen,(void *)strdup(rcp->rcpt));
		}
	      break;
	    }
	  }
	}
      break;
    case 3	:	/*looking for aliases	*/
      if (dsp->rcpts!=(RCPTYP **)0) {/*always	*/
	int i;

	for (i=0;dsp->rcpts[i]!=(RCPTYP *)0;i++) {
	  register RCPTYP *rcp;

	  rcp=dsp->rcpts[i];
	  switch (rcp->status) {
	    case m_toresolv	:/*Alias	*/
	      if (alreadyseen(seen,rcp->rcpt)==true)
		break;
	      (void) ckr_checkdomover(sql,&(rcp->rcpt),EVCKR);
              disploc=(DISTYP *)calloc(1,sizeof(DISTYP));
              disploc->efrom=strdup(dsp->efrom);
              disploc->inbehalf=strdup(rcp->rcpt);
	      (void) asprintf(&(disploc->rejectto),"%s@%s",POSTM,eml_getdomain(rcp->rcpt));
	      if ((disploc->rcpts=expendalias(sql,dsp->locip,dsp->efrom,rcp,dsp->auth,&seen))!=(RCPTYP **)0) {
	  	int j;

	        for (j=0;disploc->rcpts[j]!=(RCPTYP *)0;j++) 
	    	  (void) updatestatus(sql,dsp,disploc->rcpts[j]);
	        }
              displst=(DISTYP **)vec_addlstlst((void **)displst,(void *)disploc);
	      seen=(char **)vec_addlstlst((void **)seen,(void *)strdup(rcp->rcpt));
	      break;
	    default		:/*standard name*/
	      /*already done in phase 2		*/
	      break;
	    }
	  }
	}
      break;
    default	:	/*SAFE Guard		*/
      seen=(char **)vec_freelstlst((void **)seen,VECFREE);
      proceed=false;
      break;
    }
  phase++;
  }
return displst;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to sort user list to dispatch	*/
/*						*/
/************************************************/
static int sortdispatch(SQLFILE *sql,DSPTYP *dsp,AIGTYP ***aig)

{
#define	OPSO	"modckr.c:sortdispatch,"
#define	MFOR	"Unable to retrieve domain <%s> forwarder (configuration problem?)"

int todo;

todo=0;
if ((dsp!=(DSPTYP *)0)&&(dsp->rcpts!=(RCPTYP **)0)) {
  int i;

  for (i=0;dsp->rcpts[i]!=(RCPTYP *)0;i++) {
    register RCPTYP *rcploc;
    GRPTYP *group;

    rcploc=dsp->rcpts[i];
    group=(GRPTYP *)0;
    switch (rcploc->status) {
      case m_stayput	:
      case m_quarantine	:
	todo+=addsort(aig,LOCALQUAR,SMTPPORT,dsp->efrom,rcploc);
	break;
      case m_acceptable	:
      case m_direct	:
	todo+=addsort(aig,LOCALDEL,SMTPPORT,dsp->efrom,rcploc);
	break;
      case m_toforward	:
      case m_mustfwd	:
        if ((group=dbc_getgroup(sql,rcploc->rcpt))!=(GRPTYP *)0) {
          if ((group->smtprmt!=(char *)0)&&(group->smtpport!=0)) {
	    todo+=addsort(aig,group->smtprmt,group->smtpport,dsp->efrom,rcploc);
	    }
	  else
            (void) jrl_journalling(false,EVCKR,MFOR,eml_getdomain(rcploc->rcpt));
          group=dbc_freegroup(group);
	  }
	break;
      case m_toresend	:
      case m_relayed	:
	if (rcploc->rcpt!=(char *)0) {	/*	always		*/
	  if (dns_gotmxorip(rcploc->rcpt)==true) {
	    char *mxname;

	    (void) asprintf(&mxname,"%s %s",WITHMX,eml_getdomain(rcploc->rcpt));
	    todo+=addsort(aig,mxname,SMTPPORT,dsp->inbehalf,rcploc);
	    (void) free(mxname);
	    }
	  else
	    (void) clm_alert(0,"%s Unable to find one MX for email address <%s> (trouble?, forgery?)",
			        OPSO,eml_getdomain(rcploc->rcpt));
	  }
	break;
      case m_virusfound	:	/*we do not dispatch virus	*/
      case m_rejected	:	/*we do not dispatch rejected	*/
      case m_noaddress	:	/*we do not dispatch no address	*/
	break;
      default		:
        (void) jrl_journalling(true,EVCKR,"Going to core Dump from <%s> (pid='%05d'), "
					  "Unexpected <%s> rcpt status '%d'",
					  OPSO,getpid(),rcploc->rcpt,rcploc->status);
	(void) clm_core_dump("%s Unexpected <%s> rcpt status '%d'",
			      OPSO,rcploc->rcpt,rcploc->status);
	break;
      }
    }
  }
return todo;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do email dispatching	*/
/*	according dsp structure			*/
/*						*/
/************************************************/
static int dodispatching(SQLFILE *sql,DSPTYP *dsp,int *todo,char *infile)

{
#define	OPDI	"modckr.c:dodispatching,"

int numdone;
AIGTYP **aig;

numdone=0;
aig=(AIGTYP **)0;
*todo+=sortdispatch(sql,dsp,&aig);
if (aig!=(AIGTYP **)0) {
  int i;



  for (i=0;aig[i]!=(AIGTYP *)0;i++) {
    register AIGTYP *aigloc;

    aigloc=aig[i];
    switch (aigloc->status) {
      case m_stayput	:
      case m_quarantine	:
      case m_acceptable	:
      case m_direct	:
	(void) dbc_setinbehalf(sql,aigloc->rcpts,dsp->efrom);
        numdone+=trt_transit(sql,ANYADDR,aigloc->smtpmx,aigloc->smtpport,aigloc->from,dsp,infile,aigloc->rcpts);
	break;
      case m_toresend	:
      case m_relayed	:
	(void) dbc_setinbehalf(sql,aigloc->rcpts,dsp->inbehalf);
        numdone+=torelay(sql,dsp,aigloc);
        (void) alignstatus(aigloc->rcpts,rcpt_foreign);
	break;
      case m_toforward	:
      case m_mustfwd	:
	(void) dbc_setinbehalf(sql,aigloc->rcpts,dsp->inbehalf);
        numdone+=torelay(sql,dsp,aigloc);
        (void) alignstatus(aigloc->rcpts,rcpt_forward);
	break;
      case m_rejected	:	/*nothing to do	*/
      case m_noaddress	:	/*nothing to do	*/
	(void) dbc_setinbehalf(sql,aigloc->rcpts,dsp->efrom);
	break;
      default		:	/*REAL BUG	*/
        (void) clm_core_dump("%s Unexpected aig[%d] status '%d' (bug!)",
                              OPDI,i,aigloc->status);
	break;
      }
    (void) dbc_setlogs(sql,dsp,aigloc->rcpts);
    }
  aig=freeaig(aig);
  }
return numdone;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check about remote, return	*/
/*	a status string if touble.		*/
/*						*/
/************************************************/
static char *isremotegood(SQLFILE *sql,int auth,CKPTYP *scn,char *rcpt)

{
char *answ;
PRFTYP  *profile;
int phase;
int proceed;

answ=(char *)0;
profile=(PRFTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting user profile		*/
      if ((profile=dbc_getemailprofile(sql,rcpt))==(PRFTYP *)0) {
	(void) jrl_journalling(false,EVCKR,"No profile available for recipient <%s> (config?)",rcpt);
	phase=999;	/*no need to go further		*/
	}
      break;
    case 1	:	/*is remote profile wrong	*/
      if (profile->active==false) {
        answ=eml_ascstat(EXPIRED,rcpt);
	phase=999;	/*can't write to		*/
	}
      break;
    case 2	:	/*are we in forced relay mode	*/
      if ((scn->ctxflag&FRELAY)!=0) {
	phase=999;	/*no need to check	*/
	}
      break;
    case 3	:	/*is originator MX known	*/
      if (((scn->ctxflag&NOMX)!=0)&&(profile->checkrmtmx==1)) {
        answ=eml_ascstat(BADMX,eml_getdomain(scn->efrom));
	phase=999;	/*flagged mean originator, OUT	*/
	}
      if (auth==true)	/*no need to check further	*/
	phase=999;
      break;
    case 4	:	/*BADEHLO		*/
      if (((scn->ctxflag&BADEHLO)!=0)&&(profile->checkrmtmx==1)) {
        answ=eml_ascstat(NODNS,scn->heloname);
	phase=999;
	}
      break;
    case 5	:	/*Checking SPF		*/
      if (((scn->ctxflag&SPFFAIL)!=0)&&(profile->checkrmtspf==1)) {
        answ=eml_ascstat(BADSPF,scn->peerip,eml_getdomain(scn->efrom));
	phase=999;	/*No SPF?!			*/
        }
      break;
    case 6	:	/*Check Reverse-address needed	*/
      if (((scn->ctxflag&FNOVERS)!=0)&&(profile->needreverse==1)) {
        answ=eml_ascstat(BADRVRS,scn->peerip,eml_getdomain(scn->efrom));
	phase=999;	/*No REVERSE?!			*/
        }
      break;
    case 7	:	/*is originator black listed	*/
      if (profile->blsid!=0) {
	char *note;

        if ((note=bls_isblslisted(sql,scn->peerip,profile->blsid))!=(char *)0) {
          answ=eml_ascstat(BLCKLST,note,scn->peerip);
	  (void) free(note);
	  phase=999;	/*flagged mean spammer,out fast	*/
	  }
	}
      break;
    case 8	:	/*comparing email max size	*/
      if (scn->taille>0) {
	if ((scn->taille/(1024*1024))>profile->emailsz) {
          answ=eml_ascstat(ISBIG,rcpt);
	  phase=999;	/*flagged email is too big	*/
	  }
	}
      break;
    default	:	/*SAFE Guard			*/
      profile=dbc_freeprofile(profile);
      proceed=false;
      break;
    }
  phase++;
  }
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if user is a known 	*/
/*	user.					*/
/*						*/
/************************************************/
static int isknownuser(SQLFILE *sql,char **from)

{
int status;
RCPTFLG rcptflg;
int phase;
int proceed;

status=false;
rcptflg=rcpt_uknstatus;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*find out about user	*/
      if (*from!=(char *)0) {	/*always	*/
	USRTYP *usr;

        usr=dbc_mkuserinfo(*from,rcptflg);
        (void) ckr_isuserknown(sql,usr);
	rcptflg=usr->rcptflg;
        usr=dbc_freeuserinfo(usr);
	if (rcptflg!=rcpt_uknstatus)
	  phase++;	/*we found a status	*/
	}
      break;
    case 1	:	/*could be a multipath	*/
      if (*from!=(char *)0) {	/*always	*/
	char *idpath;

	if ((idpath=dbc_getidpath(sql,*from))!=(char *)0) {
	  USRTYP *usr;

          usr=dbc_mkuserinfo(idpath,rcptflg);
          (void) ckr_isuserknown(sql,usr);
	  rcptflg=usr->rcptflg;
          usr=dbc_freeuserinfo(usr);
          (void) free(*from);
	  *from=idpath;
	  }
	}
      break;
    case 2	:
      switch (rcptflg) {
	case rcpt_forward      :
	case rcpt_islocalias   :
	case rcpt_islocal      :
	  status=true;
          break;	
	default		       :/*bad status	*/
				/*nothing to do	*/
          break;	
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check about FROM		*/
/*	information.				*/
/*						*/
/************************************************/
static char *docheckfrom(CNFTYP *cnf,SQLFILE *sql,char *request)

{
#define	DOCHECKFROM	"docheckfrom,"

char *answ;
char *info;
PRFTYP *profile;
int bulk;
int status;
int groupid;
char *from;
char *peerip;
char *locip;
int locport;
int aut;
int ctxflags;
int phase;
int proceed;

answ=(char *)0;
info=strdup("");
profile=(PRFTYP *)0;
bulk=15;
status=0;
groupid=0;
from=(char *)malloc(strlen(request));
peerip=(char *)malloc(strlen(request));
locip=(char *)malloc(strlen(request));
locport=0;
aut=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*extracting data	*/
      if (sscanf(request,"%s %s %s %d %d %d",from,peerip,locip,&locport,&aut,&ctxflags)!=6){
        (void) jrl_journalling(true,EVCKR,"%s bad from request <%s> (bug?)",
					  DOCHECKFROM,request);
	status|=BADFROM;/*big big trouble	*/
	phase=999;	/*gesting out fast	*/
	}
      break;
    case  1	:	/*could from be an IP	*/
      status=ctxflags&RBANNED;	/*keep ban flag	*/
      if (afn_ispureip(eml_getdomain(from))==true) {
	AFNTYP *afnnum;

        if ((afnnum=afn_getipnum(eml_getdomain(from)))!=(AFNTYP *)0) {
	  afnnum=afn_freeipnum(afnnum);
	  status|=BADFROM;/*not RFC compliant	*/
	  phase=999;	/*gesting out fast	*/
	  }
	}
      break;
    case  2	:	/*good Authentication	*/
      if (aut!=0) {
	status|=AUTH;
	phase++;	/*no need to check relay*/
	}
      break;
    case  3	:	/*checking RELAYABLE	*/
      switch (rly_isrelayable(cnf->relais,from,peerip,&groupid)) {
	case rly_forcedrelay	:
	  status|=FRELAY;
	  break;
	case rly_smartrelay	:
	  status|=RELAY;
	  break;
	default			:
	  break;
	}
      break;
    case  4	:	/*is domain responding	*/
      if (dns_gooddomdns(eml_getdomain(from))==false) {
	status|=NOMX;
	status|=SPFFAIL;
	phase=999;   /*no need to check further	*/
	}
      break;
    case  5	:	/*checking MX value	*/
      if (dns_gotmxorip(from)==false) {
	status|=NOMX;
	}
      break;
    case  6	:	/*checking if good spf	*/
      if (spf_isgoodspf(from,peerip)==false)
	status|=SPFFAIL;
      break;
    case  7	:	/*relayed and user exist?*/
      if ((status&(RELAY|AUTH))!=0) {
	if (isknownuser(sql,&from)==false) {
	  status|=BADUSR;
	  phase=999;   /*no need to check further	*/
	  }
	(void) free(info);
        info=strdup(from);
	}
      break;
    case  8	:	/*is it a relayed user	*/
      if ((status&(RELAY|AUTH))!=0) {
	GRPTYP *grp;

	status|=RLYUSR;
        if ((grp=dbc_getgroup(sql,from))==(GRPTYP *)0) {
	  USRTYP *usr;	/*lets locat the group	*/

          if ((usr=ckr_getuserinfo(sql,from,locip))!=(USRTYP *)0) {
            if ((usr->prfid!=0)&&((profile=dbc_getprofile(sql,usr->prfid))!=(PRFTYP *)0)) {
              bulk=profile->bulk;
              profile=dbc_freeprofile(profile);
	      phase=999;/*no need to go further	*/
	      }
            usr=dbc_freeuserinfo(usr);
	    }
	  }
        grp=dbc_freegroup(grp);
	}
      break;
    case  9	:	/*originator bulk value	*/
      if ((profile=dbc_getemailprofile(sql,from))!=(PRFTYP *)0) {
        bulk=profile->bulk;
        profile=dbc_freeprofile(profile);
	}
      break;
    default	:	/*SAFE guard		*/
      (void) asprintf(&answ,"%d %d %d %s",status,bulk,groupid,info);
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(locip);
(void) free(peerip);
(void) free(from);
(void) free(info);
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check a foreign sending	*/
/*						*/
/************************************************/
static char *checkforeign(CNFTYP *cnf,SQLFILE *sql,CKPTYP *scn)

{
char *answ;
int sent;
int groupid;
int phase;
int proceed;

answ=(char *)0;
sent=0;
groupid=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*are we authenticated?	*/
      if ((scn->ctxflag&(AUTH|RELAY|FRELAY))==0){
        answ=eml_ascstat(NORELAY,scn->efrom);
	phase=999;	/*we do not relay	*/
	}
      break;
    case 1	:	/*extracting group id	*/
      if ((scn->ctxflag&(RELAY|FRELAY))!=0)
        (void) rly_isrelayable(cnf->relais,scn->efrom,scn->peerip,&groupid);
      break;
    case 2	:	/*too many sending	*/
      if (dbc_toomanymail(sql,scn->efrom,scn->peerip,groupid,&sent)==true) {
        answ=eml_ascstat(MAXDAY,scn->efrom,sent);
	phase=999;	/*we do not relay	*/
	}
      break;
    case 3	:	/*lest send email	*/
      (void) forceoverride(sql,scn->efrom,scn->rcpt);
      answ=eml_ascstat(RELAYOK,scn->rcpt);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check about RCPT request	*/
/*						*/
/************************************************/
static char *docheckrcpt(CNFTYP *cnf,SQLFILE *sql,char *request)

{
#define	MRRC	"modckr.c:docheckrcpt,"

char *answ;
CKPTYP *scn;
USRTYP *usr;
int auth;
int phase;
int proceed;

answ=(char *)0;
scn=(CKPTYP *)0;
usr=(USRTYP *)0;
auth=false;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*find about request contents	*/
      if ((scn=eml_scanckrrqst(request))==(CKPTYP *)0) {
        (void) jrl_journalling(false,EVCKR,"%s, Wrong request <%s> (bug?)",
					   MRRC,request);
        answ=eml_ascstat(ERRPROC,"Wrong request format (bug?)");
	proceed=false;
        }
      break;
    case   1	:	/*find about RCPT domain	*/
      if ((strchr(scn->rcpt,'@')==(char *)0)&&(strcmp(scn->rcpt,POSTM)!=0)) {
        (void) jrl_journalling(false,EVCKR,"%s, Wrong email address <%s> (bug?)",
					   MRRC,scn->rcpt);
        answ=eml_ascstat(SYNERR,"Wrong email format (bug?)");
	phase=999;	/*getting out fast		*/
	}
      break;
    case   2	:	/*is from OK			*/
      if ((scn->ctxflag&BADFROM)!=0) {
	(void) asprintf(&answ,"%d Invalid format for \"MAIL FROM: <%s>\"",BADPAR,scn->efrom);
	phase=999;
	}
      break;
    case   3	:	/*Is remote IP banned		*/
      if ((scn->ctxflag&BADUSR)!=0) {
	/*Oct 4, 2012					*/
	/*BADUSR will be tested after receiving data	*/
	/*part, against previous email Message-Id,	*/
	/*This to detect a forward.			*/
	/*Previous Code.
        answ=eml_ascstat(UKNUSR,scn->efrom);
	phase=999;
	*/
	}
      break;
    case   4	:	/*Is remote IP banned		*/
      if ((scn->ctxflag&RBANNED)!=0) {
        answ=eml_ascstat(IPBAN,scn->peerip,scn->rcpt);
	phase=999;
	}
      break;
    case   5	:	/*find out originator group	*/
      if ((usr=ckr_finduserstatus(sql,scn->efrom,scn->rcpt,scn->peerip,scn->locip,auth,(char **)0))==(USRTYP *)0) {
        answ=eml_ascstat(ERRPROC,"Wrong user name (bug?)");
	phase=999;	/*getting out fast		*/
	}
      break;
    case   6	:	/*checking forced relay		*/
      if ((scn->ctxflag&(FRELAY|BADUSR))!=0) {
	phase++;/*no need to check is originator is good*/
	}
      break;
    case   7	:	/*lets get the usr infomation	*/
      if (((scn->ctxflag&(AUTH|RELAY))!=0)&&((usr->rcptflg!=rcpt_fforeign))) {
        AFNTYP *afnnum;

	auth=true;
        if ((afnnum=dns_isipaddress(eml_getdomain(scn->efrom)))==(AFNTYP *)0) {
	  GRPTYP *grp;

          if ((grp=dbc_getgroup(sql,scn->efrom))==(GRPTYP *)0) {
            answ=eml_ascstat(WRGRLY,scn->efrom);
	    phase=999;	/*getting out fast		*/
	    }
          grp=dbc_freegroup(grp);
	  }
        afnnum=afn_freeipnum(afnnum);
        }
      break;
    case   8	:	/*Is remote a trouble maker	*/
      switch (usr->rcptflg) {
	case rcpt_forward	:
	case rcpt_islocal	:
	case rcpt_islocalias	:
	case rcpt_isalias	:
	case rcpt_resend	:
	  if ((answ=isremotegood(sql,auth,scn,usr->username))!=(char *)0) {
	    phase=999;	/*no getting out fast		*/
	    }
	  break;
	default			:/*nothing to check	*/
	  break;
	}
      break;
    case   9	:	/*what is rcpt flag		*/
      switch (usr->rcptflg) {
	case rcpt_notchecked	:
          answ=eml_ascstat(CHKUSR,"");
	  break;
	case rcpt_forward	:
          answ=eml_ascstat(FWRDOK,"");
	  break;
	case rcpt_islocal	:
          (void) asprintf(&answ,"%d %s",CMDOK,"Address accepted");
	  break;
	case rcpt_alidirect	:/*alias accep remote	*/
          (void) asprintf(&answ,"%d Alias (#%d) accept Originator by special agreement",
				 ALIOK,usr->numalias);
	  break;
	case rcpt_forcefwd	:/*Remote email MUst be forwarded	*/
          (void) asprintf(&answ,"%d %s",MUSTFWD,"Recipient request originator forwarding by special agreement");
	  break;
	case rcpt_direct	:/*Remote accepted	*/
          (void) asprintf(&answ,"%d %s",OVEROK,"Recipient accept originator by special agreement");
	  break;
	case rcpt_islocalias	:/*aliases		*/
	case rcpt_isalias	:
          (void) asprintf(&answ,"%d Address will be aliased (#%d)",ALIOK,usr->numalias);
	  break;
	case rcpt_resend	:
          answ=eml_ascstat(SENDOK,"");
	  break;
	case rcpt_foreign	:
	  answ=checkforeign(cnf,sql,scn);
	  break;
	case rcpt_fforeign	:
	  answ=checkforeign(cnf,sql,scn);
	  (void) free(answ);
          (void) asprintf(&answ,"%d %s",RELAYOK,"Recipient very special agreement to relay Email");
	  break;
	case rcpt_loopback	:
          (void) asprintf(&answ,"%d MX for relayed address <%s> is a loop back to me!!",MISSMX,scn->rcpt);
	  break;
	case rcpt_banned	:
          answ=eml_ascstat(BANNED,scn->efrom);
	  break;
	case rcpt_noalias	:
          answ=eml_ascstat(REJECT,"Account without recipient");
	  break;
	case rcpt_unknown	:
          answ=eml_ascstat(NOLOC,scn->rcpt);
	  break;
        case rcpt_inactive	:
          answ=eml_ascstat(EXPIRED,scn->rcpt);
	  break;
        case rcpt_mailtb	:
        case rcpt_boxtb		:
          answ=eml_ascstat(MBFULL,scn->rcpt);
	  break;
	case rcpt_nomx		:
          answ=eml_ascstat(MISSMX,eml_getdomain(scn->rcpt));
	  break;
	case rcpt_uknstatus	:
          answ=eml_ascstat(ERRPROC,"Unknown user status");
	  break;
	default			:
	  (void) clm_alert(0,"%s Unexpected status '%d' for user <%s> (bug?)",
			     MRRC,usr->rcptflg,scn->rcpt);
	  break;
	}
      break;
    case  10	:	/*lets check for trouble	*/
      if (answ==(char *)0) {
	char *strloc;

	(void) asprintf(&strloc,"Very unexpected internal status for user <%s>",scn->rcpt);
        answ=eml_ascstat(ERRPROC,strloc);
	(void) free(strloc);
	}
      break;
    default	:	/*SAFE guard lets get out	*/
      usr=dbc_freeuserinfo(usr);
      scn=eml_freeckrrcpt(scn);
      proceed=false;
      break;
    }
  phase++;
  }
(void) jrl_journalling(true,EVCKR,(char *)0);
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to request to a remote	*/
/*	authentication server about a user ID	*/
/*						*/
/************************************************/
static char *checkremoteuser(SQLFILE *sql,char *username,char *password,GRPTYP *grp)

{
#define	OPCK		"modckr.c:checkremoteuser"

#define	WAITOPEN	30	/*authenticator rsp time*/

static VOCTYP vocprot[]={
	{pro_pop3,POP3,(void *)0},	/*pop3 plain link	*/
	{pro_pop3t,POP3TLS,(void *)0},	/*pop3 secured TLS link	*/
	{pro_pop3s,POP3S,(void *)0},	/*pop3 secured SSL link	*/
	{pro_imap,IMAP,(void *)0},	/*IMAP plain link	*/
	{pro_imapt,IMAPTLS,(void *)0},	/*IMAP secured TLS link	*/
	{pro_imaps,IMAPS,(void *)0},	/*IMAP secured TLS link	*/
	{0,(char *)0,(void *)0}
	};

char *answ;
VOCTYP *voc;
TLSFILE *tlsfile;
char *buffer;
int socker;
int phase;
int proceed;

answ=(char *)0;
voc=(VOCTYP *)0;
tlsfile=(TLSFILE *)0;
buffer=(char *)0;
socker=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*do we know the protocol	*/
      if (((voc=clm_getvoca(vocprot,grp->authprot))==(VOCTYP *)0)||(voc->code==0)) {
        answ=eml_ascstat(CMDABRT,"Remote server unknown protocol! (config?)");
	phase=999;	/*trouble trouble		*/
	}
      break;
    case 1	:	/*establishing link		*/
      if ((tlsfile=tls_opentcp(grp->authrmt,grp->authport,WAITOPEN,&socker))==(TLSFILE *)0) {
        answ=eml_ascstat(CMDABRT,"Unable to contact authentication server");
	phase=999;	/*trouble trouble		*/
	}
      break;
    case 2	:	/*encoding link			*/
      switch (voc->code) {
	case pro_imaps	:	/*establishing TLS	*/
	case pro_pop3s	:	/*establishing TLS	*/
	  if (tls_gotls(tlsfile,false)<0) {
            answ=eml_ascstat(CMDABRT,"Unable to establish secure link with authenticator");
            (void) tls_close(tlsfile);
	    phase=999;	/*trouble trouble		*/
	    }
	  break;
	default	:	/*nothing special to do		*/
	  break;
	}
      break;
    case 3	:	/*waiting for signon		*/
      switch (voc->code) {
	case pro_imap	:	/*imap			*/
	case pro_imapt	:	/*imaptls		*/
	case pro_imaps	:	/*imaps			*/
	case pro_pop3	:	/*pop3			*/
	case pro_pop3t	:	/*pop3tls		*/
	case pro_pop3s	:	/*pop3s			*/
	  if (tls_waitsignon(tlsfile,WAITOPEN,voc->code,&buffer)==false) {
	    answ=eml_ascstat(CMDABRT,"Unable to get positive acknowledge from authentication server");
	    phase=999;	/*trouble trouble		*/
	    }
	  break;
	default	:	/*unexpected protocol		*/
	  (void) clm_alert(0,"%s Authentication signon, Unexpected Protocol (bug?)",OPCK);
          answ=eml_ascstat(CMDABRT,"Authentication Protocol not properly supported (Bug?)");
	  break;
	}
      buffer=clm_freestr(buffer);
      break;
    case 4	:	/*establishing TLS if requested	*/
      switch (voc->code) {
	case pro_pop3t	:
	  if (eml_setpop3tls(tlsfile)==false)
	    (void) clm_alert(0,"%s, Unable to establish a TLS connection on pop-3 protocol with <%s>",OPCK,grp->authrmt);
	  break;
	case pro_imapt	:
	  if (eml_setimaptls(tlsfile)==false)
	    (void) clm_alert(0,"%s, Unable to establish a TLS connection on imap protocol with <%s>",OPCK,grp->authrmt);
	  break;
        default		:	/*nothing to do		*/
	  break;
	}
      break;
    case 5	:	/*checking ID			*/
      switch (voc->code) {
	case pro_pop3	:	/*pop3			*/
	case pro_pop3t	:	/*pop3+tls		*/
	case pro_pop3s	:	/*pop3s			*/
	  if (eml_checkviapop3(tlsfile,grp->authrmt,grp->authport,username,password)==true)
            answ=eml_ascstat(IDOK);
          else
            answ=eml_ascstat(BADID);
	  break;
	case pro_imap	:	/*plain imap		*/
	case pro_imapt	:	/*imapunder tls		*/
	case pro_imaps	:	/*crypted imaps		*/
	  if (eml_checkviaimap(tlsfile,grp->authrmt,grp->authport,username,password)==true)
            answ=eml_ascstat(IDOK);
          else
            answ=eml_ascstat(BADID);
	  break;
	default	:
	  (void) clm_alert(0,"%s Authentication link, Unexpected Protocol (bug?)",OPCK);
          answ=eml_ascstat(CMDABRT,"Authentication Protocol unknown (Bug?)");
	  break;
	}
      break;
    case 6	:	/*closing connexion		*/
      (void) tls_close(tlsfile);
      break;
    default	:	/*SAFE Guard			*/
      proceed=false;
      break;
    }
  phase++;
  }
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check a LOCAL user ID	*/
/*						*/
/************************************************/
static char *checklocaluser(SQLFILE *sql,char *username,char *password)

{
char *answ;
USRTYP *localuser;
int phase;
int proceed;

answ=(char *)0;
localuser=(USRTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*Is local user known	*/
      localuser=dbc_mkuserinfo(username,rcpt_uknstatus);
      (void) dbc_getlocaluser(sql,localuser);
      if (localuser->rcptflg==rcpt_unknown) {
        answ=eml_ascstat(BADID);/*user not existing	*/
	phase=999;
	}
      break;
    case  1	:	/*is user an active one		*/
      if (localuser->active==false) {
        answ=eml_ascstat(EXPIRED,"Not current account");
	phase=999;	/*user not accepted		*/
	}
      break;
    case  2	:	/*is password empty		*/
      if (localuser->password==(char *)0) 
	localuser->password=strdup("");
      if (strlen(localuser->password)==0) {
        (void) jrl_journalling(true,EVCKR,"Warning! user <%s> password is empty (check)",username);
        answ=eml_ascstat(IDOK);
	phase=999;
	}
      break;
    case  3	:	/*lets check password		*/
      if (localuser->password!=(char *)0) {	/*always*/
	char *crypted;

	crypted=crypt(password,localuser->password);
        if ((crypted!=(char *)0)&&(strcmp(crypted,localuser->password)==0))
          answ=eml_ascstat(IDOK);
        else
          answ=eml_ascstat(BADID);
	phase=999;	/*flagged local			*/
	}
      break;
    default	:	/*SAFE Guard			*/
      localuser=dbc_freeuserinfo(localuser);
      proceed=false;
      break;
    }
  phase++;
  }
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check a user ID		*/
/*						*/
/************************************************/
static char *docheckid(CNFTYP *cnf,SQLFILE *sql,char *request)

{
#define MRID	"docheckid,"
#define	TRBJRL	"%s seems to be missing within request <%s> (bug?)"

char *answ;
char *locip;
int grpin;
int locport;
char *username;
char *password;
GRPTYP *grp;
int phase;
int proceed;

answ=(char *)0;
locip=(char *)0;
grpin=0;
locport=0;
username=(char *)0;
password=(char *)0;
grp=(GRPTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*lets be sur about request	*/
      if ((request==(char *)0)||(strlen(request)==0)) {
        (void) jrl_journalling(true,EVCKR,"%s, bad chkid request <%s> (bug?)",
					  MRID,request);
        answ=eml_ascstat(ERRPROC,"Wrong format 'docheckid_1' (bug?)");
        proceed=false;
        }
      break;
    case  1	:	/*extracting variable		*/
      username=(char *)calloc(strlen(request),sizeof(char));
      password=(char *)calloc(strlen(request),sizeof(char));
      locip=(char *)calloc(strlen(request),sizeof(char));
      if (sscanf(request,"%s %s %s %d",username,password,locip,&locport)!=4) {
        (void) jrl_journalling(true,EVCKR,"%s bad chkid format <%s> (bug?)",
					  MRID,request);
        answ=eml_ascstat(ERRPROC,"Wrong format 'docheckid_2' (bug?)");
	phase=999;	/*going out fast		*/
	}
      break;
    case  2	:	/*converting username from B64 	*/
      if (strlen(username)==0) {
        (void) jrl_journalling(true,EVCKR,TRBJRL,"username",MRID,request);
        answ=eml_ascstat(BADPAR,request);
	phase=999;	/*going out fast		*/
        }
      break;
    case  3	:	/*converting password from B64 	*/
      if (strlen(password)==0) {
        (void) jrl_journalling(true,EVCKR,TRBJRL,"password",MRID,request);
        answ=eml_ascstat(BADPAR,request);
	phase=999;	/*going out fast		*/
	}
      break;
    case  4	:	/*is it a know domain		*/
      if ((grp=dbc_getgroup(sql,username))==(GRPTYP *)0) {
        answ=eml_ascstat(CMDABRT,"Domain Unknown");
	phase=999;	/*going out fast		*/
	}
      break;
    case  5	:	/*is domain active		*/
      grpin=grp->groupid;
      if (grp->active==false) {
        answ=eml_ascstat(CMDABRT,"Domain Inactive");
	phase=999;	/*going out fast		*/
	}
      break;
    case  6	:	/*is it a remote domain		*/
      if (((grp->smtprmt!=(char *)0)&&(strlen(grp->smtprmt)>0))&&
	  ((grp->authrmt!=(char *)0)&&(strlen(grp->authrmt)>0))) {
        answ=checkremoteuser(sql,username,password,grp);
	phase++;	/*no need for a local user check*/
	}
      break;
    case  7	:	/*user is a known one?		*/
      answ=checklocaluser(sql,username,password);
      break;
    default	:	/*SAFE guard			*/
      if (answ!=(char *)0) {	/*always		*/
	int retstat;
	int pos;

        if (sscanf(answ,"%d%n",&retstat,&pos)==1) {
	  char *ptr;

	  (void) asprintf(&ptr,"%d %d%s",retstat,grpin,answ+pos);
	  (void) free(answ);
	  answ=ptr;
	  }
	}
      grp=dbc_freegroup(grp);
      (void) free(password);
      (void) free(username);
      (void) free(locip);
      proceed=false;
      break;
    }
  phase++;
  }
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check virus within E-mail	*/
/*						*/
/************************************************/
static char *dochkvir(CLAMTYP **clamav,char *request)

{
char *answ;
const char *virname;
char *mailname;

answ=(char *)0;
virname=(char *)0;
(void) asprintf(&mailname,"%s/%s",clm_appdir(d_queue),request);
switch (cla_checkvirus(clamav,mailname,&virname)) {
  case cla_clean	:
    (void) asprintf(&answ,"%d",cla_clean);
    break;
  case cla_virus	: 
    (void) asprintf(&answ,"%d %s",cla_virus,virname);
    break;
  case cla_nocheck	:
    (void) asprintf(&answ,"%d",cla_nocheck);
    break;
  }
(void) free(mailname);
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get the spam level		*/
/*						*/
/************************************************/
static char *dogetspam(SPAMTYP **spamd,char *request)

{
#define	DOGETSPAM	"dogetspam,"

char *answ;
char *locip;
int locport;
char *email;
char *session;
int phase;
int proceed;

answ=(char *)0;
locip=(char *)calloc(strlen(request),sizeof(char));
email=(char *)calloc(strlen(request),sizeof(char));
session=(char *)calloc(strlen(request),sizeof(char));
phase=1;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*trouble trouble	*/
      answ=strdup("NAN");
      phase=999;
      break;
    case 1	:	/*extracting data	*/
      if (sscanf(request,"%s %d %s %s",locip,&locport,email,session)!=4) {
        (void) jrl_journalling(true,EVCKR,"Get spam request wrong format <%s>",
					   request);
	phase=-1;
        }
      break;
    case 2	:	/*getting spam level	*/
      if ((answ=spm_checklevel(spamd,session))==(char *)0)
	phase=-1;	/*spamassassin not avail*/
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(session);
(void) free(email);
(void) free(locip);
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to execute Email dispatching	*/
/*						*/
/************************************************/
static char *dodispatch(CNFTYP *cnf,SQLFILE *sql,char *session)

{
#define	OPDO	"modckr.c:dodispatch,"

char *answ;
char **dspstr;
FILE *fichier;
DSPTYP *dsp;
int todo;
int numdone;
int phase;
int proceed;

answ=(char *)0;
dspstr=(char **)0;
fichier=(FILE *)0;
dsp=(DSPTYP *)0;
todo=0;
numdone=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*extracting Q file	*/
      if (session!=(char *)0) {
        char *dspname;
	
	(void) asprintf(&dspname,"%s/%s-%s",clm_appdir(d_queue),session,RCPTEXT); 
        if ((fichier=fopen(dspname,"r"))==(FILE *)0) {
  	  (void) clm_alert(0,"%s Unable to open file <%s> (error=<%s>)",
			      OPDO,dspname,strerror(errno));
	  phase=999;	/*trouble trouble	*/
	  }
	(void) free(dspname);
        }
      break;
    case  1	:	/*extracting dsp data	*/
      if (fichier!=(FILE *)0) {	/*always true	*/
	char bu[2048];

        while (fgets(bu,sizeof(bu)-1,fichier)!=(char *)0) {
	  (void) clm_cleanstring(bu);
	  dspstr=(char **)vec_addlstlst((void **)dspstr,(void *)strdup(bu));
	  }
	(void) fclose(fichier);
	}
      break;
    case  2	:	/*converting *str to DSP*/
      if (dspstr!=(char **)0) {
	dsp=eml_getdispatch(dspstr);
	dspstr=(char **)vec_freelstlst((void **)dspstr,VECFREE);
	}
      break;
    case  3	:	/*checking everything OK*/
      if (dsp==(DSPTYP *)0) {
        (void) clm_alert(0,"%s Unable get a DSP struct for session <%s> (bug?)",
			   OPDO,session);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case   4	:	/*sorting destination	*/
      numdone=ckr_dispatch(sql,dsp,&todo);
      if (todo==numdone)
	(void) asprintf(&answ,"%d Message delivered to each recipient",CMDOK);
      else
	(void) asprintf(&answ,"%d session <%s> %d Emails not delivered (expected=%d)",
			      ERRPROC,dsp->session,todo-numdone,todo);
      break;
    default	:	/*SAFE guard		*/
      dsp=eml_cleandispatch(dsp);
      proceed=false;
      break;
    }
  phase++;
  }
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect all E-mail stats	*/
/*	for a specific session, if everything	*/
/*	fine, will return CMDOK, if not will	*/
/*	return ERRPROC.				*/
/************************************************/
static char *dostatfrom(WRKTYP *wrk,char *buffer)

{
#define	OPDS	"modck.c:dostatfrom,"

int status;
char **split;
u_long in;
u_long out;
char *from;
char *fromdom;
char *fromgrp;
GRPTYP *ingrp;
int phase;
int proceed;

status=ERRPROC;
split=(char **)0;
in=(u_long)0;
out=(u_long)0;
from=(char *)0;
fromdom=(char *)0;
fromgrp=(char *)0;
ingrp=(GRPTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*extracting message	*/
      if ((split=clm_splitstring(buffer,'\n'))==(char **)0) {
	(void) clm_alert(0,"%s Unable to split buffer <%s>",OPDS,buffer);
	phase=999;	/*trouble trouble		*/
	}
      break;
    case 1	:	/*from information	*/
      from=(char *)calloc(strlen(split[0]),sizeof(char));
      if (sscanf(split[0],"%s %lu",from,&in)!=2) {
	(void) clm_alert(0,"%s Unable to scan split[0] <%s>",OPDS,split[0]);
	phase=999;	/*trouble trouble		*/
	}
      break;
    case 2	:	/*getting group name	*/
      if ((ingrp=dbc_getgroup(wrk->sql,from))!=(GRPTYP *)0) {
        fromgrp=strdup(ingrp->grpname);
	fromdom=eml_getdomain(from);
        ingrp=dbc_freegroup(ingrp);
	}
      break;
    case 3	:	/*scanning information	*/
      out=in;
      if (split[1]!=(char *)0) {
	int i;
        for (i=1;split[i]!=(char *)0;i++) {
	  (void) addstat(wrk,fromgrp,fromdom,from,&in,&out,split[i]);
	  }
        }
      status=CMDOK;
      break;
    default	:	/*SAFE guard		*/
      fromgrp=clm_freestr(fromgrp);
      from=clm_freestr(from);
      split=(char **)vec_freelstlst((void **)split,VECFREE);
      proceed=false;
      break;
    }
  phase++;
  }
return eml_ascstat(status,buffer);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to give checker advice about	*/
/*	user status.				*/
/*	the SQL logs.				*/
/************************************************/
static char *giveckradv(SQLFILE *sql,char *msg)

{
#define OPGI	"modckr.c:giveckradv,"

char *answ;
char **split;
DSPTYP *dsp;
int phase;
int proceed;

answ=(char *)0;
split=(char **)0;
dsp=(DSPTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*splitting the message	*/
      if ((split=clm_splitstring(msg,'\n'))==(char **)0) {
	const char *trouble;

	trouble="Unable to split advice rqst (bug?)";
	(void) clm_alert(0,"%s %s",OPGI,trouble);
	answ=eml_ascstat(CMDBAD,trouble);
	proceed=false;
	}
      break;
    case 1	:	/*extracting the advice */
      if ((dsp=eml_getdispatch(split))==(DSPTYP *)0) {
	const char *trouble;

	trouble="Unable to make advice from request (bug?)";
	(void) clm_alert(0,"%s %s",OPGI,trouble);
	answ=eml_ascstat(CMDBAD,trouble);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*give checker advice	*/
      if (dsp!=(DSPTYP *)0) { /*always		*/
	int i;

	for (i=0;dsp->rcpts[i]!=(RCPTYP *)0;i++) 
	 (void) ckradvise(sql,dsp,dsp->rcpts[i]);
	}
      break;
    case 3	:	/*building msg advice	*/
      answ=dsp_advicemsg(dsp);
      dsp=eml_cleandispatch(dsp);
      break;
    default	:	/*SAFE Guard		*/
      if (answ==(char *)0)
        (void) asprintf(&answ,"%s %s\n.",MYADV,split[0]);
      split=(char **)vec_freelstlst((void **)split,VECFREE);
      proceed=false;
      break;
    }
  phase++;
  }
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to store all session data to	*/
/*	the SQL logs.				*/
/************************************************/
static char *doseslog(SQLFILE *sql,char *msg)

{
int status;
char **split;
DSPTYP *dsp;

status=ERRPROC;
split=clm_splitstring(msg,'\n');
if ((dsp=eml_getdispatch(split))!=(DSPTYP *)0) {
  status=CMDOK;
  (void) dbc_setinbehalf(sql,dsp->rcpts,dsp->efrom);
  if (dsp->rcpts!=(RCPTYP **)0) {
    RCPTYP **ptr;

    for (ptr=dsp->rcpts;*ptr!=(RCPTYP *)0;ptr++) 
      (*ptr)->grpid=dbc_getgroupid(sql,(*ptr)->rcpt);
    }
  (void) dbc_setlogs(sql,dsp,dsp->rcpts);
  dsp=eml_cleandispatch(dsp);
  }
split=(char **)vec_freelstlst((void **)split,VECFREE);
return eml_ascstat(status,"");
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to store final session report	*/
/*	within the seslogs table.		*/
/************************************************/
static char *dorptlogs(SQLFILE *sql,char *msg)

{
#define	OPRP	"modckr.c:dorptlogs,"

char *answ;
char *ptr;
int status;
int phase;
int proceed;

answ=(char *)0;
ptr=(char *)0;
status=ERRPROC;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*extracting session ID	*/
      if ((ptr=strchr(msg,' '))==(char *)0) {
	(void) clm_alert(0,"%s Unable to find sessionID within <%s>",OPRP,msg);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*storing data		*/
      *ptr='\000';	/*splitting line	*/
      ptr++;
      if (dbc_rptsession(sql,msg,ptr)==false) {
	(void) clm_alert(0,"%s Unable to store sessionlogs for session: %s <%s>",OPRP,msg,ptr);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*everything OK		*/
      status=CMDOK;
      break;
    default	:	/*SAFE Guard		*/
      answ=eml_ascstat(status,"");
      proceed=false;
      break;
    }
  phase++;
  }
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the remote IP	*/
/*	ban/trust status.			*/
/*						*/
/************************************************/
static char *getbantrust(SQLFILE *sql,char *buffer)

{
char *answ;
register int val;

answ=(char *)0;
if (dbc_gettrust(sql,buffer,TRSTRS)==true) 
  val=TRSTED;
else {
  switch (val=dbc_getban(sql,buffer)) {
    case BANUKN	:	/*not in bantrust table	*/
    case BANSTD	:	/*present without status*/
      val=0;
      break;
   case  BANBAN	:	/*banned status		*/
      val=RBANNED;
      break;
    default	:
      (void) clm_alert(0,"%s Unexpected status='%d' in table (DB problem?)","modckr.c:getbantrust,");
      val=0;
      break;
    }
  }
(void) asprintf(&answ,"%d %d",CMDOK,val);
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check about the trust data	*/
/*	stream received from trusted server.	*/
/*						*/
/************************************************/
static char *chktrustdata(SQLFILE *sql,char *buffer)

{
#define	OPDA	"modckr.c:chktrustdata,"

char *answ;
char *peername;
char *peerip;
char *trustfile;
char **trustlst;
TRUTYP *trusts;
int phase;
int proceed;

answ=(char *)0;
peername=(char *)calloc(strlen(buffer),sizeof(char));
peerip=(char *)calloc(strlen(buffer),sizeof(char));
trustfile=(char *)calloc(strlen(buffer),sizeof(char));
trustlst=(char **)0;
trusts=(TRUTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*extracting data*/
      if (sscanf(buffer,"%s %s %s",peername,peerip,trustfile)!=3) {
        (void) clm_alert(0,"%s, Unable to parse command line <%s> (bug?)",OPDA,buffer);
        answ=eml_ascstat(ERRPROC,"Unable to parse command line");
	proceed=false;
	}
      break;
    case 1	:	/*getting the session ID*/
      if ((trustlst=eml_gettrust(trustfile))==(char **)0) {
        (void) clm_alert(0,"%s, Unable to get trust data for session <%s> (bug?)",
			    OPDA,trustfile);
        answ=eml_ascstat(ERRPROC,"Unable to retrieve trust data stream");
	proceed=false;
	}
      break;
    case 2	:	/*getting data stream	*/
      if ((trusts=tru_parsetrust(peername,peerip,trustlst))==(TRUTYP *)0) {
	(void) asprintf(&answ,"%d Unable to parse data stream (session=<%s>)",ERRPROC,buffer);
	phase=999;	/*no need to go further	*/	
	}
      trustlst=(char **)vec_freelstlst((void **)trustlst,VECFREE);
      break;
    case 3	:	/*managing data stream	*/
      (void) tru_setorigin(trusts);
      if ((answ=tru_upddb(sql,trusts))!=(char *)0) {
	char *report;

	report=answ;
	(void) asprintf(&answ,"%d %s",OVEROK,report);
	(void) free(report);
	}
      trusts=tru_freetrusts(trusts);
      break;
    case 4	:	/*managing data stream	*/
      if (answ==(char *)0)
        answ=eml_ascstat(CMDOK,"OK");
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(trustfile);
(void) free(peerip);
(void) free(peername);
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check about previous 	*/
/*	received Email MSG-ID.			*/
/*						*/
/************************************************/
static char *chkprvmessageID(SQLFILE *sql,char *buffer)

{
char *answ;
char *session;
char *msgid;

answ=(char *)0;
session=calloc(strlen(buffer)+1,sizeof(char));
msgid=calloc(strlen(buffer)+1,sizeof(char));
if (sscanf(buffer,"%s %s",session,msgid)==2) {
  if (dbc_findmsgid(sql,session,msgid)==true)
    (void) asprintf(&answ,"%d <%s>",CMDOK,buffer);
  else
    (void) asprintf(&answ,"%d <%s>",FAILED,buffer);
  }
else
  (void) clm_alert(0,"modckr.c:chkprvmessageID, Invalid formart for buffer=<%s> (bug?)",buffer);
(void) free(msgid);
(void) free(session);
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to answer to an event (request*/
/*	coming from the protocol module), this	*/
/*	is NOT always responding to request as	*/
/*	some result are delayed and transmited	*/
/*	later.					*/
/*						*/
/************************************************/
static char *traitevent(WRKTYP *wrk,char *buffer,char *msgref)

{
#define	OPTR	"modckr.c:traitevent,"

/*known Checking command			*/
static VOCTYP vocckr[]={
	{ 1,CKFROM,(void *)0},
	{ 2,CKRCPT,(void *)0},
	{ 3,CKID,(void *)0},
	{ 4,CLAMVS,(void *)0},
	{ 5,CHKVIR,(void *)0},
	{ 6,GETSPM,(void *)0},
	{ 7,RQSTSND,(void *)0},
	{ 8,STATFRM,(void *)0},
	{ 9,DFLTDOM,(void *)0},
        {10,GVADV,(const void *)0},
        {11,REPTLOG,(const void *)0},
        {12,SESLOG,(const void *)0},
        {13,BANTRST,(const void *)0},
        {14,CHKTRST,(const void *)0},
        {15,CHKEMID,(const void *)0},
	{ 0,(char *)0,(void *)0}
	};

char *answ;
VOCTYP *voc;

answ=(char *)0;
if ((voc=clm_getvoca(vocckr,buffer))!=(VOCTYP *)0) {
  switch (voc->code) {
    case  1	:	/*request to check FROM	*/
      answ=docheckfrom(wrk->cnf,wrk->sql,buffer);
      break;
    case  2	:	/*Request to check RCPT	*/
      answ=docheckrcpt(wrk->cnf,wrk->sql,buffer); 
      break;
    case  3	:	/*Request to check AUTH	*/
      answ=docheckid(wrk->cnf,wrk->sql,buffer); 
      break;
    case  4	:	/*virus checker version	*/
      answ=cla_getversdate(wrk->clamav);
      break;
    case  5	:	/*checking Email for vir*/
      answ=dochkvir(&(wrk->clamav),buffer); 
      break;
    case  6	:	/*get E-mail spam level	*/
      answ=dogetspam(&(wrk->spamd),buffer); 
      break;
    case  7	:	/*dispatching email	*/
      answ=dodispatch(wrk->cnf,wrk->sql,buffer); 
      break;
    case  8	:	/*compile E-mail stats	*/
      answ=dostatfrom(wrk,buffer); 
      break;
    case  9	:	/*get default domain	*/
      answ=dbc_getdfltdom(wrk->sql);
      break;
    case 10	:	/*get default domain	*/
      answ=giveckradv(wrk->sql,buffer);
      break;
    case 11	:	/*reporting to logs	*/
      answ=doseslog(wrk->sql,buffer);
      break;
    case 12	:	/*reporting to logs	*/
      answ=dorptlogs(wrk->sql,buffer);
      break;
    case 13	:	/*get ban/trust status	*/
      answ=getbantrust(wrk->sql,buffer);
      break;
    case 14	:	/*get ban/trust status	*/
      answ=chktrustdata(wrk->sql,buffer);
      break;
    case 15	:	/*get ban/trust status	*/
      answ=chkprvmessageID(wrk->sql,buffer);
      break;
    default	:
      (void) jrl_journalling(true,EVCKR,"Unknown request <%s>",buffer);
      (void) clm_alert(0,"%s PID=%d unexpected msg <%s> reg=<%s>",OPTR,getpid(),buffer,msgref);
      answ=eml_ascstat(ERRPROC,"Syntaxe Error");
      break;
    }
  }
else {
  (void) jrl_journalling(true,EVCKR,"Unable to scan request <%s>",buffer);
  answ=eml_ascstat(ERRPROC,"Syntaxe Error");
  }
return is_errproc(answ,"modckr.c:traitevent, no answer to <%s>",buffer);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect an IPC message	*/
/*	dispatch answer once checking is done	*/
/*						*/
/************************************************/
static int dobounce(WRKTYP *wrk)

{
#define	OPDU	"modckr.c:dobounce"

int count;
TIMESPEC timeout;
char *msgref;
u_long msec;
char *msg;

count=0;
timeout.tv_sec=(time_t)0;
timeout.tv_nsec=(long)1000000;/*1ms, short time */
if ((msg=ipc_getpost(wrk->bptr,&timeout,&msgref,&msec))!=(char *)0) {
  char *answ;
 
  if ((answ=traitevent(wrk,msg,msgref))!=(char *)0) {
    timeout.tv_sec=(time_t)2;	/*2 sec max to send answer	*/
    timeout.tv_nsec=(u_long)0;
    if (ipc_postreply(wrk->bptr,&timeout,msgref,answ)==(u_long)0) {
      (void) clm_alert(0,"%s, timeout to post answer <%s> to msg <%s>",
			    OPDU,answ,msg);
      }
    (void) free(answ);
    count++;
    }
  msgref=clm_freestr(msgref);
  (void) free(msg);
  }
return count;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by wrk	*/
/*						*/
/************************************************/
static WRKTYP *freewrk(WRKTYP *wrk)

{
if (wrk!=(WRKTYP *)0) {
  wrk->stats=sta_freestats(wrk->stats);
  wrk->spamd=spm_closespamd(wrk->spamd);
  wrk->clamav=cla_closescanner(wrk->clamav);
  wrk->sql=sql_close(wrk->sql);
  wrk->cnf=cnf_freeconfig(wrk->cnf);
  (void) free(wrk);
  wrk=(WRKTYP *)0;
  }
return wrk;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if data need to be	*/
/*	collected.				*/
/*						*/
/************************************************/
static void tocollect(WRKTYP *wrk,int now)

{
#define	OPTO	"modckr.c:tocollect,"

#define COLTIME	5	/*collationne data every*/
			/*5 seconds		*/
#define	MXTIME	(COLTIME/2)

time_t curtime;
char **stlst;
int phase;
int proceed;

curtime=time((time_t)0);
stlst=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*need to be done	*/
      if (((wrk->grabed+COLTIME)>curtime)&&(now==false))
	proceed=false;	/*no need to go further	*/
      break;
    case 1	:	/*getting message list	*/
      stlst=sta_statstomsg(curtime,wrk->stats,0);
      wrk->stats=sta_freestats(wrk->stats);
      wrk->stats=sta_mkstats();
      break;
    case 2	:	/*sending messages	*/
      if (stlst!=(char **)0) {
	int i;

	for (i=0;stlst[i]!=(char *)0;i++) {
	  if (eml_sendmsg(wrk->bptr,wrk->mstpool,(time_t)MXTIME,"%s %s",STATCOL,stlst[i])==false) 
	    (void) clm_alert(0,"%s Unable to send STATCOL data in due time (Overload?)",OPTO);
	  }
	}
      break;
    default	:	/*SAFE guard		*/
      stlst=(char **)vec_freelstlst((void **)stlst,VECFREE);
      wrk->grabed=curtime; 
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to locate a spamd scanner	*/
/*						*/
/************************************************/
static SPAMTYP *getspamd(SQLFILE *sql,SPAMTYP *spamd)

{
#define	OPGD	"modckr.c:getspamd,"

static	time_t	lastchk=(time_t)0;


time_t curtime;
SRVLST **srvlst;
int phase;
int proceed;

curtime=time((time_t *)0);
srvlst=(SRVLST **)0;
phase=0;
if (lastchk==(time_t)0)
  phase=2;		/*direct start		*/
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*checking spamtype	*/
      if (((lastchk+15)<curtime)&&(spamd==(SPAMTYP *)0)) {
        (void) clm_alert(0,"Initiate reconnect to spamassassin daemon");
	lastchk=curtime;
	phase++;	/*need spamd contact!	*/
	}
      break;
    case 1	:	/*get service list	*/
      if ((lastchk+(60*5))>curtime) {
	proceed=false;	/*no need to check again*/
	}
      break;
    case 2	:	/*get service list	*/
      spamd=spm_closespamd(spamd);
      if ((srvlst=dbc_getsrvlst(sql,SPAMASSASSIN))==(SRVLST **)0) {
        (void) clm_alert(0,"%s Unable to find to spamassassin scanner service",OPGD);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*lets open spamassassin*/
      if (srvlst!=(SRVLST **)0){	/*always*/
	int i;

	for (i=0;srvlst[i]!=(SRVLST *)0;i++) {
	  register SRVLST *loc;

	  loc=srvlst[i];
          if ((spamd=spm_openspamd(loc->srvhost,loc->srvport,loc->srvaccess))==(SPAMTYP *)0) {
            (void) clm_alert(0,"%s Unable contact spamassassin scanner of priority %d",OPGD,i+1);
	    continue;
	    }
	  break;	/*found one ready!	*/
	  }
	}
      break;
    case 4	:	/*freeing service list	*/
      srvlst=dbc_freesrvlst(srvlst);
      break;
    default	:	/*SAFE Guard		*/
      lastchk=curtime;
      proceed=false;
      break;
    }
  phase++;
  }
return spamd;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check auxilary services	*/
/*						*/
/************************************************/
static void chkservices(WRKTYP *wrk,int jrlflush)

{
#define	OPCS	"modckr.c:chkservices,"

static int prvclmstat=-1;
static int prvspmstat=-1;

int phase;
int proceed;

phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*checking clamav	*/
      if (wrk!=(WRKTYP *)0) { /*always		*/
	int status;
	const char *strstat;

	status=false;
	strstat="down";
	wrk->clamav=ckr_getclamav(wrk->sql,wrk->clamav);
	if (wrk->clamav!=(CLAMTYP *)0) {
	  status=true;
	  strstat="up and running";
	  }
	if (prvclmstat!=status) {
	  (void) jrl_journalling(jrlflush,APJRL,"Virus Detector Daemon found %s (pid='%05d')",strstat,getpid());	
	  prvclmstat=status;
	  }
	if (status==false)
          (void) clm_alert(0,"%s Unable to access to a virus scanner (pid=%05d)",OPCS,getpid());
        }
      break;
    case 1	:	/*checking clamav	*/
      if (wrk!=(WRKTYP *)0) { /*always		*/
	int status;
	const char *strstat;

	status=false;
	strstat="down";
        wrk->spamd=getspamd(wrk->sql,wrk->spamd);
	if (wrk->spamd!=(SPAMTYP *)0) {
	  status=true;
	  strstat="up and running";
	  }
	if (prvspmstat!=status) {
	  (void) jrl_journalling(jrlflush,APJRL,"SPAM Detector Daemon found %s (pid='%05d')",strstat,getpid());	
	  prvspmstat=status;
	  }
	if (status==false)
          (void) clm_alert(0,"%s Unable to access to a spam scanner (pid=%05d)",OPCS,getpid());
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a working area	*/
/*						*/
/************************************************/
static WRKTYP *mkwrk(BPTR *bptr,pid_t father)

{
#define	OPWR	"modckr.c:mkwrk, Aborting ->"

WRKTYP *wrk;
const char *empty;
int phase;
int proceed;

wrk=(WRKTYP *)0;
empty=(const char *)0;
phase=1;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*trouble handling	*/
      wrk=freewrk(wrk);	/*cleaning		*/
      (void) sleep(5);	/*relaxing		*/
      phase=999;	/*exiting		*/
      break;
    case  1	:	/*trouble handling	*/
      wrk=(WRKTYP *)calloc(1,sizeof(WRKTYP));
      wrk->bptr=bptr;
      wrk->grabed=time((time_t)0);
      break;
    case  2	:	/*opening data-base	*/
      if ((wrk->sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
        (void) clm_alert(0,"%s Unable to open database",OPWR);
	phase=-1;	/*TROUBLE		*/
	}
      break;
    case  3	:	/*getting config	*/
      if ((wrk->cnf=cnf_chkconfig(wrk->sql,(CNFTYP *)0))==(CNFTYP *)0) {
	(void) clm_alert(0,"%s unable to read config",OPWR);
	phase=-1;	/*TROUBLE		*/
	}
      break;
    case  4	:	/*checking services	*/
      (void) chkservices(wrk,false);
      break;
    case  5	:	/*getting master pool	*/
      if ((wrk->mstpool=ipc_getpool(bptr,father))==(u_long)0) {
        (void) clm_alert(0,"%s Unable to get master pool number",OPWR);
	phase=-1;	/*TROUBLE		*/
	}
      break;
    case  6	:	/*preparing stats area	*/
      wrk->stats=sta_mkstats();
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return wrk;
}
/*

*/
/************************************************/
/*						*/
/* 	Routine to log rcpt transaction with	*/
/*	the rejected or impossible status.	*/
/*						*/
/************************************************/
static void logallrejected(SQLFILE *sql,DSPTYP *dsp)

{
if ((dsp!=(DSPTYP *)0)&&(dsp->rcpts!=(RCPTYP **)0)) {
  RCPTYP **rejected;
  int i;

  rejected=(RCPTYP **)0;
  for (i=0;dsp->rcpts[i]!=(RCPTYP *)0;i++) {
    switch (dsp->rcpts[i]->status) {
      case m_noaddress	:
      case m_rejected	:
        dsp->rcpts[i]->grpid=dbc_getgroupid(sql,dsp->rcpts[i]->rcpt);
	if (dsp->rcpts[i]->inbehalf==(char *)0)
	  dsp->rcpts[i]->inbehalf=strdup(dsp->efrom);
	rejected=(RCPTYP **)vec_addlstlst((void **)rejected,(void *)eml_duprcpt(dsp->rcpts[i]));
	break;
      default		:	/*nothing to do	*/
	break;
      }
    }
  if (rejected!=(RCPTYP **)0) {
    (void) dbc_setlogs(sql,dsp,rejected);
    rejected=eml_freercpts(rejected);
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Main process waiting checking request	*/
/*	from protocol handler.			*/
/*						*/
/************************************************/
static void ckrworking(MXCTYP *mxc,u_int nbrchkr,u_int iteration)

{
#define	OPCW	"modckr.c:ckrworking,"
#define	TITLE	"%s Checker module (%02d/%02d/%09lu)"

u_long cycle;
int locterm;
pid_t father;
WRKTYP *wrk;

cycle=(u_long)0;
locterm=true;
father=getppid();
(void) ipc_wakemeup(mxc->bptr,true);
(void) jrl_journalling(false,APJRL,"Starting Clement Checker %d (pid='%05d')",iteration,getpid());
(void) prc_settitle(TITLE,appname,iteration,nbrchkr,cycle);
if ((wrk=mkwrk(mxc->bptr,father))!=(WRKTYP *)0) {
  wrk->mntpool=mxc->mntpool;
  locterm=false;
  }
(void) jrl_journalling(true,APJRL,(char *)0);
sigterm=locterm;
while (sigterm==false) {
  IPEVTYP reveil;

  (void) tocollect(wrk,false);
  (void) chkservices(wrk,true);
  wrk->cnf=cnf_chkconfig(wrk->sql,wrk->cnf);
  reveil=ipc_waitevent(wrk->bptr,(TLSFILE *)0,5000);
  switch (reveil) {
    case ipev_none	:
      break;
    case ipev_tls	:
      break;
    case ipev_post	:
      cycle+=dobounce(wrk); 
      (void) prc_settitle(TITLE,appname,iteration,nbrchkr,cycle);
      break;
    default		:	/*bigre!			*/
      (void) clm_core_dump("%s unexpected event",OPCW);
      break;
    }
  if (prc_checkprocess(father)==false)
    break;
  }
(void) ipc_wakemeup(mxc->bptr,false);
(void) jrl_journalling(true,APJRL,"Terminating Clement Checker(pid='%05d')",getpid());
wrk=freewrk(wrk);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check and dispatch DSP	*/
/*						*/
/************************************************/
int ckr_dispatch(SQLFILE *sql,DSPTYP *dsp,int *todo)

{
int numdone;
char *infile;
DISTYP **disporig;

numdone=0;
(void) asprintf(&infile,"%s/%s",clm_appdir(d_queue),dsp->session);
(void) jrl_journalling(false,EVCKR,"Dispatching session <%s>",dsp->session);
if ((disporig=getdispatchorigin(sql,dsp))!=(DISTYP **)0) {
  int i;
  char *efrom;
  char *inbehalf;
  char *rejectto;
  RCPTYP **orgrcpts;

  efrom=dsp->efrom;
  inbehalf=dsp->inbehalf;
  rejectto=dsp->rejectto;
  orgrcpts=dsp->rcpts;
  for (i=0;disporig[i]!=(DISTYP *)0;i++) {
    dsp->rejectto=disporig[i]->rejectto;
    dsp->inbehalf=disporig[i]->inbehalf;
    dsp->efrom=disporig[i]->efrom;
    dsp->rcpts=disporig[i]->rcpts;
    numdone+=dodispatching(sql,dsp,todo,infile);
    }
  disporig=freedispatchorigin(disporig);
  dsp->rcpts=orgrcpts;
  dsp->rejectto=rejectto;
  dsp->inbehalf=inbehalf;
  dsp->efrom=efrom;
  }
(void) logallrejected(sql,dsp);
(void) jrl_journalling(true,EVCKR,(char *)0);
(void) free(infile);
return numdone;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract user real status	*/
/*						*/
/************************************************/
USRTYP *ckr_finduserstatus(SQLFILE *sql,char *from,char *rcpt,char *peerip,char *locip,int auth,char **note)

{
USRTYP *usr;
char *locnote;
int phase;
int proceed;

usr=(USRTYP *)0;
locnote=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting user pure info*/
      if ((usr=ckr_getuserinfo(sql,rcpt,locip))==(USRTYP *)0) 
	phase=999;	/*trouble trouble	*/
      break;
    case 1	:	/*is user authenticated	*/
      if (auth==true) {
	if (usr->rcptflg==rcpt_foreign) {
	  (void) ckr_isuserknown(sql,usr);
	  if (usr->rcptflg==rcpt_unknown)
	    /*not a localy know user, to relay	*/
	    usr->rcptflg=rcpt_foreign;
	  }
	}
      break;
    case 2	:	/*let double check	*/
      switch (usr->rcptflg) {
	case rcpt_nomx		:
	case rcpt_resend 	:
	case rcpt_forward 	:
	  (void) asprintf(&locnote,"Override set by <%s>/[%s]",from,peerip);
	  switch (ckr_getoverride(sql,from,rcpt,peerip)) {
	    case tr_banned	:
	      usr->rcptflg=rcpt_banned;
	      break;
	    case tr_accepted	:
	      switch (usr->rcptflg) {
	        case rcpt_resend	:	/* better MX			*/
	        case rcpt_forward	:
	          usr->rcptflg=rcpt_forcefwd;
		  break;
	        case rcpt_nomx		:	/*we want a rebounce later!	*/
	          usr->rcptflg=rcpt_fforeign;
		  break;
		default			:	/*nothing to do			*/
		  break;
	        }
	      break;
	    default		:	/*nothing to do	*/
	      locnote=clm_freestr(locnote);
	      break;
	    }
	  break;
	case rcpt_isalias	:
	case rcpt_islocalias	:
	case rcpt_islocal	:
	  (void) asprintf(&locnote,"Override set by <%s>/[%s]",from,peerip);
	  switch (ckr_getoverride(sql,from,rcpt,peerip)) {
	    case tr_banned	:
	      usr->rcptflg=rcpt_banned;
	      break;
	    case tr_accepted	:
	      switch (usr->rcptflg) {
		case rcpt_islocal	:
		  usr->rcptflg=rcpt_direct;
		  break;
		case rcpt_isalias	:
		case rcpt_islocalias	:
	          usr->rcptflg=rcpt_alidirect;
		  break;
		default			:
        	  (void) clm_alert(0,"%s unexpected user <%s> status",
				     "modckr.c:ckr_finduserstatus",rcpt);
		  break;
		}
	      break;
	    default		:	/*nothing to do	*/
	      locnote=clm_freestr(locnote);
	      break;
	    }
	  break;
	default		  :	/*nothing to do	*/
	  break;
	}
      if ((note!=(char **)0)&&(locnote!=(char *)0))
	*note=strdup(locnote);
      locnote=clm_freestr(locnote);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return usr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to locate a clamav scanner	*/
/*						*/
/************************************************/
CLAMTYP *ckr_getclamav(SQLFILE *sql,CLAMTYP *clamav)

{
#define	OPGT	"modckr.c:ckr_getclamav,"
#define	CLMRTRY 55 /*try contact clam again	*/
#define	FRQRLD	30 /*virus DB reload frequency	*/

static	time_t	lastchk=(time_t)0;

time_t curtime;
SRVLST **srvlst;
int phase;
int proceed;

curtime=time((time_t *)0);
srvlst=(SRVLST **)0;
phase=0;
if (lastchk==(time_t)0)
  phase=2;		/*direct start		*/
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*checking clamav	*/
      if ((clamav==(CLAMTYP *)0)&&((lastchk+CLMRTRY)<curtime)) {
        (void) clm_alert(0,"Initiate reconnect to clamd daemon");
	lastchk=curtime;
	phase++;	/*need clamd contact!	*/
	}
      break;
    case 1	:	/*get service list	*/
      if ((lastchk+(FRQRLD*60))>curtime) {
	proceed=false;	/*no need to check again*/
	}
      break;
    case 2	:	/*get service list	*/
      clamav=cla_closescanner(clamav);
      if ((srvlst=dbc_getsrvlst(sql,CLAMAV))==(SRVLST **)0) {
        (void) clm_alert(0,"%s Unable to find to virus scanner service",OPGT);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*lets open clamav	*/
      if (srvlst!=(SRVLST **)0){	/*always*/
	int i;

	for (i=0;srvlst[i]!=(SRVLST *)0;i++) {
          if ((clamav=cla_openscanner(srvlst[i]->srvhost,srvlst[i]->srvport))==(CLAMTYP *)0) {
            (void) clm_alert(0,"%s Unable contact clamav scanner of priority %d",OPGT,i+1);
	    continue;
	    }
	  break;	/*found one ready!	*/
	  }
	}
      break;
    case 4	:	/*freeing service list	*/
      srvlst=dbc_freesrvlst(srvlst);
      break;
    default	:	/*SAFE Guard		*/
      lastchk=curtime;
      proceed=false;
      break;
    }
  phase++;
  }
return clamav;
}
/*

*/
/************************************************/
/*						*/
/*	checker process prepartion and starting	*/
/*	Set the message pool value to wait about*/
/*	request from brother process.		*/
/*						*/
/************************************************/
void ckr_checker(MXCTYP *mxc,u_int nbrchkr,u_int iteration)

{
#define	OPTL	"modckr.c:ckr_checker,"

int phase;
int proceed;

phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*subscribing to message pool		*/
      if (ipc_subscribe(mxc->bptr,mxc->ckrpool,true)==false) {
        (void) clm_alert(0,"%s Unable to subscribe to message pool",OPTL);
	phase=999;
	}
      break;
    case   1	:	/*doing real work			*/
      (void) ckrworking(mxc,nbrchkr,iteration);
      break;
    default	:	/*SAFE guard			*/
      (void) ipc_subscribe(mxc->bptr,(u_long)0,false);
      proceed=false;
      break;
    }
  phase++;
  }
}
