/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	MODMEN:					*/
/*	Module doing RCPT validation and stats	*/
/*	collection to be requested/provided to	*/
/*	a remote mentor				*/
/*						*/
/************************************************/
#include	<string.h>

#include	"lowtyp.h"
#include	"utlprc.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"subprc.h"
#include	"comeml.h"
#include	"unisql.h"
#include	"unijrl.h"
#include	"gesdbc.h"
#include	"gescnf.h"
#include	"lvlipc.h"
#include	"lvleml.h"
#include	"modmen.h"
typedef enum	{
	a_advice,	/*advise request	*/
	a_tknote,	/*take note feed	*/
	a_unknown	/*unknown		*/
	}ASKTYP;

static VOCTYP vocable[]={
        {a_advice,GVADV,(const void *)0},
        {a_tknote,TKNOTE,(const void *)0},
        {a_unknown,(char *)0,(const void *)0}
        };
/*

*/
/************************************************/
/*						*/
/*	Processus to create a local autoreply	*/
/*	to mentor advice request in case mentor */
/*	is not connected.			*/
/*						*/
/************************************************/
static char *autoreply(char *msg,const char *cmd) 

{
#define	OPAU	"modmen.c:autoreply,"

char *reply;
char **split;

reply=(char *)0;
if ((split=clm_splitstring(msg,'\n'))!=(char **)0) {
  VOCTYP *voc;

  voc=clm_getvoca(vocable,split[0]);
  switch(voc->code) {
    case a_advice	:	/*request remote advice	*/
      (void) asprintf(&reply,"%s %s\n.",MYADV,split[0]);
      break;
    default		:	/*really unexpected	*/
      (void) clm_alert(0,"%s Unexpected command <%s %s>",OPAU,cmd,split[0]);
      reply=eml_ascstat(ERRPROC,cmd);
      break;
    }
  split=(char **)vec_freelstlst((void **)split,VECFREE);
  }
else {
  (void) clm_alert(0,"%s Unable to split command <%s>",OPAU,msg);
  reply=eml_ascstat(ERRPROC,msg);
  }
return reply;
}
/*

*/
/************************************************/
/*						*/
/*	Processus to grab an advice message	*/
/*	and post it to mentor, if it is not	*/
/*	possible (mentor connexion missing)	*/
/*	send an void reply to requestor.	*/
/*						*/
/************************************************/
static void dopost(BPTR *bptr,CTTYP *cta)

{
#define	OPDP	"modmnt.c:mntworking,"

TIMESPEC timeout;
u_long msec;
char *msgref;
char *msg;
char *locreply;
int mentorok;
int phase;
int proceed;


timeout.tv_sec=(time_t)1;
timeout.tv_nsec=(long)0;
msec=(u_long)0;
msgref=(char *)0;
msg=(char *)0;
locreply=(char *)0;
mentorok=true;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*mentor available	*/
      if ((cta==(CTTYP *)0)||(cta->tls==(TLSFILE *)0)) 
        mentorok=false;
      break;
    case 1	:	/*do we have a message?	*/
      if ((msg=ipc_getpost(bptr,&timeout,&msgref,&msec))==(char *)0) 
	phase=999;	/*no Messages!?		*/
      break;
    case 2	:	/*checking message	*/
      if (msg!=(char *)0) {	/*always true	*/
	char *loccmd;
	VOCTYP *voc;

	loccmd=strndup(msg,100);
        voc=clm_getvoca(vocable,loccmd);
	switch(voc->code) {
	  case a_advice	:	/*request advice*/
	    if (mentorok==false) 
	      locreply=autoreply(msg,voc->key); 
	      phase++;
	    break;
    	  case a_tknote	:	/*remote.. note	*/
	    (void) strcpy(msgref,ONEWAY);
	    break;
	  default	:	/*not unexpected*/
      	    (void) clm_alert(0,"%s unknown command <%s>",OPDP,msg);
      	    locreply=eml_ascstat(ERRPROC,loccmd);
            break;
	  }
	(void) free(loccmd);
	}
      break;
    case 3	:	/*sending msg to mentor	*/
      if (mentorok==true) {
	STRMTYP *stream;

        stream=eml_mkstream(RQSTID,msgref,msg);
        (void) eml_sendstream(cta,stream);
        stream=eml_freestream(stream);
	}
      break;
    case 4	:	/*locreply avaliable	*/
      if (locreply!=(char *)0) {
	timeout.tv_sec=(time_t)2;
	timeout.tv_nsec=(long)0;
	if (ipc_postreply(bptr,&timeout,msgref,locreply)==(u_long)0) 
	  (void) clm_alert(0,"%s timeout to post locreply answer to msgref/msg <%s/%s>",
			      OPDP,msgref,msg);
	(void) free(locreply);
	}
      break;
    default	:	/*SAFE Guard		*/
      msgref=clm_freestr(msgref);
      msg=clm_freestr(msg);
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Processus to open a contact with a 	*/
/*	remote memtor, if mentor is unreachable	*/
/*	log and continue without mentor help.	*/
/*						*/
/************************************************/
static CTTYP *getcontact()

{
#define	OPGE	"modmen.c:getcontact,"

#define	MXWMNT	10	/*waiting 10 seconds	*/

CTTYP *contact;
SQLFILE *sql;
SRVLST **srvlst;
const char *empty;
int phase;
int proceed;

contact=(CTTYP *)0;
sql=(SQLFILE *)0;
srvlst=(SRVLST **)0;
empty=(const char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*openning data-base	*/
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
	(void) jrl_journalling(true,APJRL,"Process '%d' is unable to open default data-base",getpid());
	phase=999;
	}
      break;
    case 1	:	/*getting mentor data	*/
      srvlst=dbc_getsrvlst(sql,APPMNT);
      sql=sql_close(sql);
      if (srvlst==(SRVLST **)0) {
	(void) jrl_journalling(true,APJRL,"Process '%d' is unable to retrieve mentor data",getpid());
	phase=999;
        }
      break;
    case 2	:	/*lets try to do contact*/
      if (srvlst!=(SRVLST **)0) {/*always	*/
	int i;

	for (i=0;srvlst[i]!=(SRVLST *)0;i++) {
	  int socker;

	  socker=0;
	  contact=eml_opencontact(pro_ment,srvlst[i]->srvhost,srvlst[i]->srvport,MXWMNT,&socker);
	  if (contact!=(CTTYP *)0) {
	    if (eml_sendcmd(contact,CMDOK,"%s %s",SAUTH,srvlst[i]->srvaccess)==CMDOK)
	      (void) jrl_journalling(true,APJRL,"Contact established with Mentor at <%s/%d>",
						 srvlst[i]->srvhost,srvlst[i]->srvport);
	     else {
	       (void) jrl_journalling(false,APJRL,"Unable to validate contact with Mentor at <%s/%d>",
						   srvlst[i]->srvhost,srvlst[i]->srvport);
	       contact=eml_closecontact(contact);
	       }
	     break;	/*no need to go further	*/
	     }
	  }
        srvlst=dbc_freesrvlst(srvlst);
	}
      break;
    case 3	:	/*do we have contact?	*/
      if (contact==(CTTYP *)0) 
	(void) jrl_journalling(true,APJRL,"Process '%d' is unable to establish contact with a mentor!",getpid());
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Processus to wait for checker request	*/
/*	and forward them to remote mentor	*/
/*	if the remote mentor is available.	*/
/*						*/
/************************************************/
static void menworking(BPTR *bptr,CTTYP *cta)

{
#define	OPMN	"modmnt.c:mntworking,"

#define	TITLE	"%s Mentor module %08lu/%08lu/%08lu"

u_long cycle;
u_long msgnum;
u_long msgrsp;
STRMTYP *stream;
TLSFILE *tls;
pid_t father;

cycle=(u_long)0;
msgnum=(u_long)0;
msgrsp=(u_long)0;
stream=(STRMTYP *)0;
tls=(TLSFILE *)0;
if (cta!=(CTTYP *)0)
  tls=cta->tls;
father=getppid();
(void) ipc_wakemeup(bptr,true);
while (sigterm==false) {
  IPEVTYP reveil;

  (void) prc_settitle(TITLE,appname,msgrsp,msgnum,cycle);
  reveil=ipc_waitevent(bptr,tls,10000);
  switch (reveil) {
    case ipev_none	:
      break;
    case ipev_tls	:
      if ((stream=eml_collectstream(stream,cta))!=(STRMTYP *)0) {
	if (stream->ready==true) {
          msgrsp++;
	  if (eml_poststream(bptr,stream)==false) 
  	    (void) clm_alert(0,"%s Unable to send active stream, seq=<%s> (bug?)",
			       OPMN,stream->seq);
	  stream=eml_freestream(stream);
	  }
	}
      if (tls_iseof(tls)==true)
	sigterm=true;
      break;
    case ipev_post	:
      msgnum++;
      (void) dopost(bptr,cta);
      break;
    case ipev_unknown	:
    default		:
      (void) clm_core_dump("%s unexpected event",OPMN);
      break;
    }
  if (prc_checkprocess(father)==false)
    break;
  cycle++;
  }
if (stream!=(STRMTYP *)0) {
  (void) clm_alert(0,"%s Unexpected active stream, seq=<%s> (bug?)",
		     OPMN,stream->seq);
  stream=eml_freestream(stream);
  }
(void) ipc_wakemeup(bptr,false);
}
/*

*/
/************************************************/
/*						*/
/*	Processus to start a mentor interface	*/
/*	process, waiting for data to be		*/
/*	tramsmitted to remote mentor.		*/
/*						*/
/************************************************/
void men_collector(BPTR *bptr,u_long mntpool)

{
#define	OPTL	"modmnt.c:,mnt_collector"

CTTYP *contact;
int phase;
int proceed;

contact=(CTTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*subscribing to IPC	*/
      if (ipc_subscribe(bptr,mntpool,true)==false) {
        (void) clm_alert(0,"%s Unable to subscribe to message pool",OPTL);
	phase=999;
	}
      break;
    case   1	:	/*opening contact	*/
      contact=getcontact();
      break;
    case   2	:	/*doing real work	*/
      (void) menworking(bptr,contact);
      break;
    case   3	:	/*closing contact	*/
      contact=eml_closecontact(contact);
      (void) ipc_subscribe(bptr,(u_long)0,false);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
