/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	Implement very sub level procedure to 	*/
/*	handle clement specific needs.		*/
/*						*/
/************************************************/
#include	<sys/time.h>
#include	<sys/stat.h>
#include	<sys/resource.h>
#include	<sys/prctl.h>
#include	<unistd.h>
#include	<time.h>
#include	<stdio.h>
#include	<errno.h>
#include	<syslog.h>
#include	<signal.h>
#include	<stdarg.h>
#include	<string.h>
#include	<ctype.h>

#include	"lowtyp.h"
#include	"utlgac.h"
#include	"utlprc.h"
#include	"utlvec.h"
#include	"subclm.h"

/*debug level					*/
#if DEBUG
#define CLMLOG  LOG_DEBUG
#else
#define CLMLOG  LOG_INFO
#endif

/*sigterm request flag				*/
	int sigterm=false;
/*sigquit request flag				*/
	int sigquit=false;
/*sigint request flag				*/
	int sigint=false;
/*flag application running background/foreground*/
	int clm_foreground=false;
/*base directory (where to store internal data)	*/
	char	*base=""; /*no prefix		*/

/*application name				*/
	char	*appname=CLEMENT;

/*debug level					*/
#define	MXDBGLVL	15	/*maxstack	*/
static  int dbglvl=0;		/*current stakc	*/
static 	int dbgstack[MXDBGLVL];
/*

*/
/************************************************/
/*						*/
/*	Procedure to catch signal and do what is*/
/*	needed.					*/
/*						*/
/************************************************/
static void trapsig(int sig)

{
sigset_t blockmask;
sigset_t oldset;
struct sigaction sa;

(void) sigfillset(&blockmask);
(void) sigprocmask(SIG_SETMASK,&blockmask,&oldset);
if (debug>=20)
  (void) fprintf(stderr,"subclm.c:trapsig <%s/%d> received (by pid='%d')\n",
                         sys_siglist[sig],sig,getpid());
switch (sig) {
  case SIGINT	:
    sigint=true;
    break;
  case SIGQUIT	:
    sigquit=true;
    sigterm=true;
    break;
  case SIGUSR1	:
    debug++;
    (void) syslog(CLMLOG,"Increased debug level to '%d' (pid='%d')",debug,getpid());
    break;
  case SIGUSR2	:
    if (debug>0) {
      debug--;
      (void) syslog(CLMLOG,"Decreased debug level to '%d' (pid='%d')",debug,getpid());
      }
    break;
  case SIGTERM	:
    sigterm=true;
    break;
  default        :
    (void) fprintf(stderr,"%s, Unexpected signal <%s> received\n",
			   appname,sys_siglist[sig]);
    break;
  }
/*restarting signal		*/
(void) sigemptyset(&sa.sa_mask);
sa.sa_handler=trapsig;
sa.sa_flags=0;
(void) sigprocmask(SIG_SETMASK,&oldset,(sigset_t *)0);
(void) sigaction(sig,&sa,(struct sigaction *)0);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to catch signal and do what is*/
/*	needed.					*/
/*						*/
/************************************************/
static void trpmempbls(int sig)

{
#define	OPPB	"subclm.c:trpmempbls,"

switch (sig) {
  case SIGSEGV   :
    (void) clm_core_dump("Clement-Application genuine memory violation");
    break;
  default        :
    (void) clm_core_dump("%s Unexpected signal <%s> received",
			 OPPB,sys_siglist[sig]);
    break;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to change the current ID to	*/
/*	a application working ID.		*/
/*						*/
/************************************************/
static uid_t getappid(char *appnom,DIRENUM dir,int group)

{
#define MAXPATH 300

uid_t id;
char *apppath;
struct stat bufstat;

id=(uid_t)0;
(void) asprintf(&apppath,"%s/%s",clm_appdir(dir),appnom);
if (stat(apppath,&bufstat)<0) {
  (void) clm_alert(0,"%s Unable to get <%s> stat (bug?)",
		     "subclm.c:getappid",apppath);
  }
else {
  if (group==false)
    id=bufstat.st_uid;
  else
    id=bufstat.st_gid;
  }
(void) free(apppath);
return id;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add a file contents to	*/
/*	the env area.				*/
/*						*/
/************************************************/
static void gesenv(char *filename,int toadd)

{
#define	OPAD	"subclm.c:gesenv,"

FILE *fichier;

if ((fichier=fopen(filename,"r"))!=(FILE *)0) {
  u_int numline;
  char line[1000];

  numline=0;
  while (clm_getstr(fichier,line,sizeof(line),'#')!=(char *)0) {
    char var[100];
    char contents[200];

    numline++;
    (void) memset(var,'\000',sizeof(var));
    (void) memset(contents,'\000',sizeof(contents));
    if (strlen(line)==0)	/*empty line	*/
      continue;
    if (sscanf(line,"%100[^=]=%199s",var,contents)!=2) {
      (void) clm_alert(0,"%s filename:<%s> line %04d:<%s> invalid format",
			  OPAD,filename,numline,line);
      continue;
      }
    if (contents[0]=='"') {
      char *ptr;

      (void) memmove(contents,contents+1,strlen(contents+1)+1);
      if ((ptr=strchr(contents,'"'))!=(char *)0)
        *ptr='\000';
      } 
    if (toadd==true) {
      char *ptr;

      (void) asprintf(&ptr,"%s=%s",var,contents);
      if (putenv(ptr)!=0) 
        (void) clm_alert(0,"%s unable to add '%s' to env (error=<%s>)",
			    OPAD,ptr,strerror(errno));
      }
    else {
      if (unsetenv(var)<0) 
        (void) clm_alert(0,"%s unable to remove '%s' from env (error=<%s>)",
			    OPAD,var,strerror(errno));
      }
    } 
  }
else
  (void) clm_alert(0,"%s unable to open file '%s' (error=<%s>)",
		      OPAD,filename,strerror(errno));
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if directory is	*/
/*	already existing.			*/
/*						*/
/************************************************/
int clm_isdir(char *dirpath)

{
int status;
struct stat bufstat;

status=-1;
if (stat(dirpath,&bufstat)==0) {
  if (S_ISDIR(bufstat.st_mode)!=0) {
    status=0;
    }
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to concatenat a string to	*/
/*	another string, allow enough space to	*/
/*	contain the sum of both string.		*/
/*	Return the new str string pointer.	*/
/*						*/
/************************************************/
char *clm_astrcat(char *str,char *ext)

{
if ((ext!=(char *)0)&&(strlen(ext)>0)) {
  if (str==(char *)0)
    str=strdup(ext);
  else {
    str=(char *)realloc(str,(strlen(str)+strlen(ext)+1)*sizeof(char));
    (void) strcat(str,ext);
    }
  }
return str;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to format a string and then	*/
/*	concatenate the result to another string*/
/*	Return the NUMBER of bytes ADDED	*/
/*						*/
/************************************************/
int clm_fasprintf(char **dst,char *fmt,...)

{
va_list args;
char *strloc;
int added;

va_start(args,fmt);
added=vasprintf(&strloc,fmt,args);
*dst=clm_astrcat(*dst,strloc);
(void) free(strloc);
va_end(args);
return added;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove a specific char from*/
/*	a string				*/
/*						*/
/************************************************/
void clm_cutchar(char *str,char cut)

{
char *ptr;

ptr=str;
while ((ptr=strchr(ptr,cut))!=(char *)0)
  (void) memmove(ptr,ptr+1,strlen(ptr+1)+1);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	string, do not proceed if point is NULL.*/
/*						*/
/************************************************/
char *clm_freestr(register char *str)

{
if (str!=(char *)0) {
  (void) free(str);
  str=(char *)0;
  }
return str;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract and return current	*/
/*	version number.				*/
/*						*/
/************************************************/
char *clm_getvers()

{
static char *vernum=(char *)0;
static char numero[30];

if (vernum==(char *)0) {
  (void) sprintf(numero,"%d.%d-%d.%d",VERSION,LEVEL,RELEASE,UPDATE);
  vernum=numero;
  }
return vernum;
}
/*
^L
*/
/************************************************/
/*						*/
/*      Procedure to return the current time	*/
/*	expressed with a millisecond precision.	*/
/*	from the first time it was called	*/
/*						*/
/************************************************/
u_long clm_getmillisec()

{
static time_t start=(time_t)0;

u_long millisec;
struct timeval newtime;

millisec=(u_long)0;
(void) gettimeofday(&newtime,(struct timezone *)0);
if (start==(time_t)0)
  start=(newtime.tv_sec-1);
millisec=newtime.tv_sec-start;
millisec*=1000;
millisec+=(newtime.tv_usec/1000);
return millisec;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to make create a unique name	*/
/*	can be used to store process data	*/
/*	name return is dynamically allocated.	*/
/*						*/
/************************************************/
char *clm_uniquename(unsigned int seq)

{
#define UFTIME	"%Y%m%d%H%M%S"
#define	UNIQUE	"%05d-%s-%08lx"

char *uniquename;
time_t curtime;
char asctemps[100];

curtime=time((time_t)0);
(void) strftime(asctemps,sizeof(asctemps),UFTIME, localtime(&curtime));
(void) asprintf(&uniquename,UNIQUE,getpid(),asctemps,clm_getmillisec()+seq);
return uniquename;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to split a msg structure to	*/
/*	a line structure.			*/
/*						*/
/************************************************/
char **clm_splitstring(char *msg,unsigned char cut)

{
char **lines;
char *msgloc;

lines=(char **)0;
msgloc=(char *)0;
if (msg!=(char *)0)
  msgloc=strdup(msg);
msg=msgloc;
while ((msg!=(char *)0)&&(msg[0]!='\000')) {
  char *ptr;

  if ((ptr=strchr(msg,(int)cut))!=(char *)0) {
    *ptr='\000';
    if ((cut=='\n')&&(strlen(msg)>0)) {
      ptr--;
      if (*ptr=='\r')	/*to get rid of '\r\n' sequence	*/
	*ptr='\000';
      ptr++;
      }
    ptr++;
    }
  lines=(char **)vec_addlstlst((void **)lines,strdup(msg));
  msg=ptr;
  }
msgloc=clm_freestr(msgloc);
return lines;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to merge a list struct to a	*/
/*	msg string.				*/
/*						*/
/************************************************/
char *clm_mergestring(char **lst,unsigned char merge)

{
char *str;

str=(char *)0;
if (lst!=(char **)0) {
  int i;

  for (i=0;lst[i]!=(char *)0;i++)
    (void) clm_fasprintf(&str,"%s%c",lst[i],merge);
  }
return str;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a time seen as local*/
/*	time to a long express as YYYYMMDD	*/
/*						*/
/************************************************/
u_long clm_date(time_t curtime)

{
struct tm *tm;

tm=localtime(&curtime);
return (tm->tm_year*10000)+(tm->tm_mon*100)+tm->tm_mday;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the local time in	*/
/*	HH:MM:SS format.			*/
/*	Time is stored in a STATIC memeory area	*/
/*						*/
/************************************************/
char *clm_ascsystime(time_t curtime)

{
#define TTIME	"%H:%M:%S"

static char asctemps[100];

(void) strftime(asctemps,sizeof(asctemps),TTIME,localtime(&curtime));
return asctemps;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return system time from an	*/
/*	YYYY-MM-DDD HH:MM:SS format.		*/
/*						*/
/************************************************/
time_t clm_datetimesysasc(char *strdate,char *strtime)

{
#define	DBDTOUNIX	"%Y-%m-%d %H:%M:%S"
time_t datetime;

datetime=(time_t)0;
if (strdate!=(char *)0) {
  struct tm tm;
  char strdt[100];

  if ((strtime==(char *)0)||(strlen(strtime)==0))
     strtime="00:00:00";
  (void) memset(&tm,'\000',sizeof(struct tm));
  tm.tm_isdst=-1;
  (void) snprintf(strdt,sizeof(strdt),"%s %s",strdate,strtime);
  if (strptime(strdt,DBDTOUNIX,&tm)!=(char *)0) {
    datetime=mktime(&tm);
    }
  else {
    (void) clm_alert(0,"subclm.c:clm_datetimesysasc Unable to convert <%s> to time_t",strdt);
    }
  }
return datetime;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the local time in	*/
/*	YYYY-MM-DDD format.			*/
/*	date is stored in a STATIC memeory area	*/
/*						*/
/************************************************/
char *clm_ascsysdate(time_t curtime)

{
#define TDATE	"%Y-%m-%d"

static char asctemps[100];

(void) strftime(asctemps,sizeof(asctemps),TDATE,localtime(&curtime));
return asctemps;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the local time in	*/
/*	YYYY-MM-DDD HH:MM:SS format.		*/
/*	date is stored in a STATIC memeory area	*/
/*						*/
/************************************************/
char *clm_ascsysdatetime(time_t curtime)

{
#define TDTIME	"%Y-%m-%d %H:%M:%S"

static char asctemps[100];

(void) strftime(asctemps,sizeof(asctemps),TDTIME,localtime(&curtime));
return asctemps;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to transform the local system	*/
/*	time in ASCII time stamp.		*/
/*	Stored in STATIC memory area.		*/
/*						*/
/************************************************/
char *clm_ascsysstamp(time_t curtime)

{
#define TSTAMP	"%a, %d %b %Y %T %z"

static char ascstamp[100];

(void) strftime(ascstamp,sizeof(ascstamp),TSTAMP,localtime(&curtime));
return ascstamp;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to display an alert in the	*/
/*	debug log.				*/
/*						*/
/************************************************/
void clm_alert(const int dlevel,const char *fmt,...)

{
if (debug>=dlevel)
  {
  va_list args;
  char *strloc;

  va_start(args,fmt);
  (void) vasprintf(&strloc,fmt,args);
  if ((clm_foreground==true)&&((verbose==true))) {
    time_t curtime;
    char logtime[30];

    curtime=time((long *)0);
    (void) fflush(stderr);
    (void) strftime(logtime,sizeof(logtime),"%H:%M:%S",localtime(&curtime));
    (void) fprintf(stderr,"%s %s\n",logtime,strloc);
    (void) fflush(stderr);
    }
  else {
    (void) syslog(CLMLOG,"%s",strloc);
    }
  (void) free(strloc);
  va_end(args);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to display an argv list 	*/
/*						*/
/************************************************/
void clm_argvtrace(const int dlevel,const char *fmt,char *argv[])

{
if ((debug>=dlevel)&&(argv[0]!=(char *)0)) {
  register int i;
  char *tmp;

  tmp=strdup(argv[0]);
  for (i=1;argv[i]!=(char *)0;i++) {
    char *strloc;

    (void) asprintf(&strloc,"%s %s",tmp,argv[i]);
    (void) free(tmp);
    tmp=strloc;
    }
  (void) clm_alert(dlevel,fmt,tmp);
  (void) free(tmp);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to 'compute' a new base 	*/
/*	according the current directory		*/
/*	return a newly allocated memory		*/
/*						*/
/************************************************/
char *clm_mkbase(char *newbase)

{
char *ptr;

ptr=(char *)0;
switch (newbase[0]) {
  case	'.'	: {
    char curdir[512];
    char newpath[1000];

    if (getcwd(curdir,sizeof(curdir))!=(char *)0) {
      (void) snprintf(newpath,sizeof(newpath),"%s/%s",curdir,newbase);
      if (chdir(newpath)<0) {
        (void) clm_alert(0,"Unable to reach directory <%s>",newpath);
	break;	/*immediat exit	*/
        }
      (void) getcwd(newpath,sizeof(newpath));
      if (chdir(curdir)<0) {
        (void) clm_alert(0,"Unable to reach directory <%s>",curdir);
	break;	/*immediat exit	*/
        }
      ptr=strdup(newpath);
      } 
    } 
    break;
  case	'/'	:
    ptr=strdup(newbase);
    break;
  default	:
    break;
  }
if (ptr==(char *)0) {
  (void) clm_alert(0,"Unable to assign new base '%s', immediat exit",newbase);
  (void) exit(-1);
  }
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to 'compute' an application 	*/
/*	directory according dir enum value	*/
/*						*/
/************************************************/
const char *clm_appdir(DIRENUM dir)

{
static char dirname[(int)d_null][200];


if (dir>=(int)d_null)
  return (const char *)0;	/*Trouble	*/
if (strlen(dirname[(int)dir])==0) {
  char *sysbase;
  char *apvers;
  char *subdir;

  sysbase=" ";	/*something impossible		*/
  apvers=clm_getapvers();
  subdir="";
  switch (dir) {
    case (d_tmp)		:
      sysbase="/var/tmp";
      apvers=CLEMENT"/"; 
      break;
    case (d_crash)	:
      sysbase="/var/crash";
      break;
    case (d_tls)		:
      sysbase="/etc/pki";
      break;
    case (d_etc)		:
      sysbase="/etc";
      break;
    case (d_syscnf)	:
      sysbase="/etc/sysconfig";
      apvers=""; 
      break;
    case (d_cnf)		:
      sysbase="/etc";
      subdir="conf.d";
      break;
    case (d_spool)	:
      sysbase="/var/spool";
      break;
    case (d_mails)	:
      sysbase="/var/spool";
      subdir="mails";
      break;
    case (d_tests)	:
      sysbase="/var/spool";
      subdir="tests";
      break;
    case (d_dbgs)	:
      sysbase="/var/spool";
      subdir="dbgs";
      break;
    case (d_queue)	:
      sysbase="/var/spool";
      subdir="mqueue";
      break;
    case (d_lock)		:
      sysbase="/var/lock";
      break;
    case (d_log)		:
      sysbase="/var/spool";
      subdir="logs";
      break;
    case (d_trblq)	:
      sysbase="/var/spool";
      subdir="trblq";
      break;
    case (d_transit)	:
      sysbase="/var/spool";
      subdir="transit";
      break;
    case (d_tobesend)	:
      sysbase="/var/spool";
      subdir="tobesend";
      break;
    case (d_bin)		:
      sysbase="/usr/bin";
      apvers=""; 
      break;
    case (d_sbin)		:
      sysbase="/usr/sbin";
      apvers=""; 
      break;
    case (d_varlib)	:
      sysbase="/var/lib";
      break;
    case (d_usrlib)	:
      sysbase="/usr/lib";
      break;
    default		:
      (void) clm_core_dump("Unexpected DIRENUM value '%d'",(int)dir);
      break;
    }
  (void) snprintf(dirname[(int)dir],sizeof(dirname[(int)dir]),"%s%s/%s%s",base,sysbase,apvers,subdir);
  }
return dirname[(int)dir];
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert a string to a code,*/
/*	remaining of the string is copy at the	*/
/*	beginning.				*/
/*	Return a pointer to table, never a NULL	*/
/*	Pointer unless the table itself is NULL	*/
/*						*/
/************************************************/
VOCTYP *clm_getvoca(const VOCTYP *table,char *str)

{
VOCTYP *voc;

voc=(VOCTYP *)0;
if (table!=(VOCTYP *)0) {
  register char *ptr;
  register int i;

  ptr=str; 
  while (*ptr==' ')
    ptr++;
  if (ptr!=str)
    (void) memmove(str,ptr,strlen(ptr)+1);
  for (i=0;table[i].key!=(char *)0;i++) {
    int max;

    max=strlen(table[i].key);
    if (strncasecmp(table[i].key,str,max)==0) {
      if ((str[max]!='\000')&&(isalnum(str[max])!=0))
        continue;/*it is not the right word	*/
      ptr=str+max;
      (void) memmove(str,ptr,strlen(ptr)+1);
      ptr=str;
      while (isspace(*ptr)!=0)
	ptr++;
      if (ptr!=str)
        (void) memmove(str,ptr,strlen(ptr)+1);
      voc=(VOCTYP *)(table+i);
      break;
      }
    }
  if (voc==(VOCTYP *)0)
   voc=(VOCTYP *)(table+i);/*return the "unknown" info	*/
  }
return voc;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set the SIGV signal trap	*/
/*	purpose is to CORE_DUMP in case we have	*/
/*	a memory failure of some kind.		*/
/*						*/
/************************************************/
void clm_trapsegv(int onoff)

{
static struct sigaction oldsa;

if (onoff==true) {
  struct sigaction newsa;

  newsa.sa_flags=0;
  newsa.sa_handler=trpmempbls;
  (void) sigemptyset(&newsa.sa_mask);
  (void) sigaction(SIGSEGV,&newsa,&oldsa);
  }
else {
  (void) sigaction(SIGSEGV,&oldsa,(struct sigaction *)0);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a core_dump in	*/
/*	the right directory.			*/
/*						*/
/************************************************/
void clm_core_dump(char *frmt,...)

{
va_list args;
char *crashinfo;

va_start(args,frmt);
(void) vasprintf(&crashinfo,frmt,args);
(void) clm_alert(1,crashinfo);
(void) free(crashinfo);
(void) prc_core_dump(clm_appdir(d_crash),frmt,args);
va_end(args);
}
/*

*/
/************************************************/
/*						*/
/*	procedure to set/unset trapped signal	*/
/*						*/
/************************************************/
void clm_settrap(int set)

{
static struct sigaction oldint;
static struct sigaction oldterm;
static struct sigaction oldquit;
static struct sigaction oldusr1;
static struct sigaction oldusr2;

static int alldone=false;

if (set==alldone) {
  switch (set) {
    case true	:
      (void) clm_core_dump("clm_settrap already set");
      break;
    case false	:
      (void) clm_core_dump("clm_settrap not previously set");
      break;
    default	:
      (void) clm_core_dump("clm_settrap unproper set value");
      break;
    } 
  }
(void) clm_trapsegv(set);
if (set==true) {
  struct sigaction newsa;

  newsa.sa_flags=0;
  newsa.sa_handler=trapsig;
  (void) sigaction(SIGUSR2,&newsa,&oldusr2);
  (void) sigaction(SIGUSR1,&newsa,&oldusr1);
  (void) sigaction(SIGINT,&newsa,&oldint);
  (void) sigaction(SIGTERM,&newsa,&oldterm);
  (void) sigaction(SIGQUIT,&newsa,&oldquit);
  }
else {
  (void) sigaction(SIGQUIT,&oldquit,(struct sigaction *)0);
  (void) sigaction(SIGTERM,&oldterm,(struct sigaction *)0);
  (void) sigaction(SIGINT,&oldint,(struct sigaction *)0);
  (void) sigaction(SIGUSR1,&oldusr1,(struct sigaction *)0);
  (void) sigaction(SIGUSR2,&oldusr2,(struct sigaction *)0);
  }
alldone=set;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set a string to lower case	*/
/*	only.					*/
/*						*/
/************************************************/
char *clm_strtolower(char *str)

{
if (str!=(char *)0) {
  register char *ptr;

  for (ptr=str;*ptr!='\000';ptr++)
    *ptr=(char)tolower((int)*ptr);
  }
return str;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure extract line from a file,	*/
/*	forget about any line starting with	*/
/*	carcom character (if carcom character	*/
/*	not null).				*/
/*	Doesn't return the '\r' and '\n' 	*/
/*	character.				*/
/*						*/
/************************************************/
char *clm_getstr(FILE *fichier,char *str,u_int taille,char carcom)

{
char *strloc;

(void) memset(str,'\000',taille);
while ((strloc=fgets(str,taille,fichier))!=(char *)0) {
  char *ptrloc;

  if (carcom!='\000') {
    if (str[0]==carcom)
      continue;
    ptrloc=str+1;
    while ((ptrloc=strchr(ptrloc,carcom))!=(char *)0) {
      if (*(ptrloc-1)=='\\') {
        (void) memmove(ptrloc-1,ptrloc,strlen(ptrloc)+1);	
        ptrloc++;
        }
      else {
        *ptrloc='\000';
        break;
        }
      }
    }
  ptrloc=strloc+strlen(strloc);
  while (ptrloc!=(char *)0) {
    ptrloc--;
    switch (*ptrloc) {
      case '\n'     :
      case '\r'     :
        *ptrloc='\000';
        break;
      default       :
        ptrloc=(char *)0;
	break;
      }
    }
  break;
  }
return strloc;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to create all the directory	*/
/*	needed by a filename.			*/
/*						*/
/************************************************/
int clm_createdirs(char *dirname,int mode)

{
#define	OPCR	"subclm.c:clm_createdirs,"

struct stat bufstat;
int newdir;

newdir=false;
if (dirname[strlen(dirname)-1]=='/')
  dirname[strlen(dirname)-1]='\000';
if (stat(dirname,&bufstat)==0) {
  if (S_ISDIR(bufstat.st_mode)==0) {
    (void) clm_core_dump("%s file <%s> is NOT a directory (error=<%s>) (bug!?)",
			  OPCR,dirname,strerror(errno));
    }
  if (bufstat.st_mode!=mode) {
    if (chmod(dirname,mode)<0) 
      (void) clm_alert(0,"%s Unable to change directory <%s> permission mode to '%d' (error=<%s>)",
			  OPCR,dirname,mode,strerror(errno));
    }
  }
else {
  char *ptr;
  char *dirloc;

  dirloc=strdup(dirname);
  newdir=true;
  if ((ptr=strrchr(dirloc,'/'))!=(char *)0) {
    *ptr='\000';
    ptr++;
    if (strlen(ptr)>0)
      newdir|=clm_createdirs(dirloc,mode);
    }
  if (mkdir(dirname,mode)<0) {
    newdir=false;
    switch (errno) {
      case EEXIST	:	/*already done	*/
	break;			/*no trouble	*/
      default		:
        (void) clm_core_dump("%s unable to create directory <%s> (error=<%s>) (bug?!)",
	  		     OPCR,dirname,strerror(errno));
	break;			/*big trouble	*/
      }
    }
  (void) free(dirloc);
  }
return newdir;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to append a file contents to	*/
/*	a file previously open.			*/
/*	If maxcount is zero, appen the whole	*/
/*	file contents.				*/
/*						*/
/************************************************/
int clm_appenfile(FILE *fout,char *filename,register int maxcount)

{
#define	OPAP	"subclm.c:clm_appenfile,"

int done;
FILE *fin;
int phase;
int proceed;

done=false;
fin=(FILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*opening file		*/
      if ((fin=fopen(filename,"r"))==(FILE *)0) {
	(void) clm_alert(0,"%s unable to open appened file <%s> (error=<%s>)",
			   OPAP,filename,strerror(errno));
        proceed=false;
	}
      break;
    case 1	:	/*appening file		*/
      if (fin!=(FILE *)0) {
	int total;
	int got;
	char buffer[1024];

	total=0;
        do {
	  got=fread(buffer,sizeof(char),sizeof(buffer),fin);
	  if (got>0)
	    (void) fwrite(buffer,sizeof(char),got,fout);
	  if (maxcount>0) {
	    total+=got;
	    if (total>maxcount)
	      break;
	    }
	  } while (got==sizeof(buffer));
        }
      break;
    case 2	:	/*closing file		*/
      (void) fclose(fin);
      done=true;
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to copy one file content to 	*/
/*	another file.				*/
/*	return '1' if successdul, '0' otherwise.*/
/*						*/
/************************************************/
int clm_dupfile(char *original,char *copy)

{
#define	OPDU	"subclm.c:clm_dupfile,"

int done;
FILE *fout;
int phase;
int proceed;

done=0;
fout=(FILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*open input file	*/
      if ((fout=fopen(copy,"w"))==(FILE *)0) {
	(void) clm_alert(0,"%s unable to open duplicated file <%s> (error=<%s>)",
			   OPDU,original,strerror(errno));
        proceed=false;
	}
      break;
    case 1	:	/*appening file		*/
      (void) clm_appenfile(fout,original,0);
      break;
    case 2	:	/*closing duplicate file*/
      (void) fclose(fout);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to change the current ID to	*/
/*	a clement working ID.			*/
/*						*/
/************************************************/
void clm_chgid(char *appnom,DIRENUM dir)

{
#define	OPID	"subclm.c:clm_chgid,"
#define	CHGMSG \
  "%s Unable to %s to '%d' (error=<%s> (cur u/g id='%d/%d',eid='%d/%d') bug?)"

struct rlimit limites;

(void) memset(&limites,'\000',sizeof(struct rlimit));
if (getuid()==0) {	/*is changing ID possible	*/
  uid_t uid;
  uid_t gid;

  gid=getappid(appnom,dir,true);
  uid=getappid(appnom,dir,false);
  if (setegid(gid)<0)
    (void) clm_alert(0,CHGMSG,OPID,"setegid",gid,
				    strerror(errno),
				    getuid(),getgid(),
				    geteuid(),getegid());
  if (seteuid(uid)<0)
    (void) clm_alert(0,CHGMSG,OPID,"seteuid",uid,
				    strerror(errno),
				    getuid(),getgid(),
				    geteuid(),getegid());
  }
if (getrlimit(RLIMIT_CORE,&limites)<0) {
  (void) clm_alert(0,"getrlimit error='%s'",strerror(errno));
  }
limites.rlim_cur=limites.rlim_max;
if (setrlimit(RLIMIT_CORE,&limites)<0) {
  (void) clm_alert(0,"setrlimit error='%s'",strerror(errno));
  }
(void) prctl(PR_SET_DUMPABLE,1,0,0,0);/*to allow core-dump	*/
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to fully change current ID,	*/
/*	(needed before doing a 'system()' call).*/
/*						*/
/************************************************/
void clm_backfullid()

{
#define	OPBA	"subclm.c:clm_backfullid,"
#define	BCKID	"%s Unable to proceed phase='%d' (error=<%s> bug?)"
int status;
int euid;
int egid;
int phase;
int proceed;

status=true;
egid=getegid();
euid=geteuid();
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*setting effective groupe	*/
      status=setegid(getgid());
      break;
    case 1	:	/*setting effective id		*/
      status=seteuid(getuid());
      break;
    case 2	:	/*setting group id		*/
      status=setgid(egid);
      break;
    case 3	:	/*setting user id		*/
      status=setuid(euid);
      break;
    default	:	/*SAFE guard			*/
      if (status!=0) {
        (void) clm_core_dump(BCKID,OPID,status,strerror(errno));
        }
      proceed=false;
      break;
    }
  phase++;
  if (status<0) {
    status=phase;
    phase=999;		/*trouble trouble		*/
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to load a configuration file	*/
/*	within the sysconfig directory		*/
/*						*/
/************************************************/
void clm_loadsysconf(char *appnom,int toadd)

{
char *conffile;

(void) asprintf(&conffile,"%s%s",clm_appdir(d_syscnf),appnom);
(void) gesenv(conffile,toadd);
(void) free(conffile);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to load a configuration file	*/
/*	all values are stored within the	*/
/*	environement.				*/
/*						*/
/************************************************/
void clm_loadconf(char *appnom)

{
char *conffile;

(void) asprintf(&conffile,"%s%s.conf",clm_appdir(d_etc),appnom);
(void) gesenv(conffile,true);
(void) free(conffile);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if file has been	*/
/*	updated since 'tempo' time.		*/
/*	return status:				*/
/*	-1	: file missing			*/
/*	 0	: file is not new		*/
/*	 1	: file is new			*/
/*						*/
/************************************************/
int clm_isnewfile(char *filename,time_t tempo)

{
int status;
struct stat bufstat;

status=-1;
if (stat(filename,&bufstat)==0) {
  status=0;
  if (bufstat.st_mtime>tempo)
    status=1; 
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to application+version name	*/
/*						*/
/************************************************/
char *clm_getapvers()

{
static char *apvers=(char *)0;

if (apvers==(char *)0) {
  static char apinfo[50];

  char *ptr;
  char version[30];

  (void) strcpy(version,clm_getvers());
  if ((ptr=strchr(version,'-'))!=(char *)0)
    *ptr='\000';
  (void) snprintf(apinfo,sizeof(apinfo),"%s-%s/",appname,version);
  apvers=apinfo;
  }
return apvers;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get rid of next blank	*/
/*	character within a string		*/
/*	an E-mail string's end.			*/
/*						*/
/************************************************/
char *clm_skipblank(char *string)

{
if (string!=(char *)0) {
  register char *ptr;

  ptr=string;
  while (isblank(*ptr)!=0)
    ptr++;
  if (ptr!=string)
    (void) memmove(string,ptr,strlen(ptr)+1);
  }
return string;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get rid of any \r or \n	*/
/*	at the end of a string.			*/
/*						*/
/************************************************/
char *clm_skiprtlf(char *string)

{
if (string!=(char *)0) {
  register int taille;
  register char *ptr;

  taille=strlen(string)-1;
  ptr=string+taille;
  while ((*ptr=='\n') || (*ptr=='\r')) {
    *ptr='\000';
    taille--;
    if (taille<0)
      break;
    ptr--;
    }
  }
return string;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure remove any ending \n or \r at	*/
/*	an E-mail string's end.			*/
/*	remove any blank string at the begining	*/
/*						*/
/************************************************/
char *clm_cleanstring(char *string)

{
string=clm_skiprtlf(string);
return clm_skipblank(string);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to format a string and add	*/
/*	the formated string to a string list.	*/
/*						*/
/************************************************/
char **clm_addfrmstr(char **strlst,const char *fmt,...)

{
va_list args;
char *str;

va_start(args,fmt);
(void) vasprintf(&str,fmt,args);
strlst=(char **)vec_addlstlst((void **)strlst,(void *)str);
va_end(args);
return strlst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to put the process in sleep	*/
/*	briefly while keeping the timeout value	*/
/*	in sync.				*/
/*	return 0 if timeout counter are 	*/
/*	exhausted				*/
/*						*/
/************************************************/
int clm_timeslice(TIMESPEC *timeout)

{
struct itimerval timer;
struct itimerval oldtimer;

timer.it_interval.tv_sec=0;
timer.it_interval.tv_usec=0;
timer.it_value.tv_sec=timeout->tv_sec;
timer.it_value.tv_usec=timeout->tv_nsec/1000;
(void) setitimer(ITIMER_REAL,&timer,&oldtimer);
(void) usleep(1000);	/*1 ms sleep		*/
(void) setitimer(ITIMER_REAL,&oldtimer,&oldtimer);
timeout->tv_sec=oldtimer.it_value.tv_sec;
timeout->tv_nsec=oldtimer.it_value.tv_usec*1000;
return timeout->tv_sec+timeout->tv_nsec;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to push debug level and 	*/
/*	replace current debug level with	*/
/*	newdebug value.				*/
/*						*/
/************************************************/
void clm_pushdebug(int newdebug)

{
if (dbglvl<MXDBGLVL) {
  dbgstack[dbglvl]=debug;
  dbglvl++;
  debug=newdebug;
  }
else
  (void) clm_alert(0,"%s, Debug stack overload (Bug?)","subclm.c:clm_pushdebug");
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to pop a debug level		*/
/*						*/
/************************************************/
void clm_popdebug()

{
if (dbglvl>0) {
  dbglvl--;
  debug=dbgstack[dbglvl];
  }
else 
  (void) clm_alert(0,"%s, Debug stack underload (Bug?)","subclm.c:clm_popdebug");
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory which		*/
/*	application lost track.			*/
/*						*/
/************************************************/
void clm_trashmem(int tgtdbg,GACTYP *gat)

{
if (gac_trash(tgtdbg,gat)<0) {
  switch (errno) {
    default :       /*spell trouble!		*/
      (void) clm_alert(0,"%s Unable to trash Memory (errno=<%s> bug?)","subclm.c:clm_trashmem,",strerror(errno));
      break;
    }
  }
}
