/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	UNIDNS:					*/
/*	Take Care of all access to DNS data	*/
/*						*/
/************************************************/
#include	<string.h>
#include	<resolv.h> 
#include	<stdlib.h> 
#include	<ctype.h> 
#include	<errno.h> 
#include	<arpa/inet.h>
#
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unidns.h"

#define	B64K	65536		/*according US Cert VU#738331	*/

typedef	union	{
        HEADER hdr;		/*defined in resolv.h 		*/
        u_char buf[B64K];	/*according US Cert VU#738331	*/
	}RSPTYP;
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare 2 MX and return	*/
/*	a value according preference value	*/
/*						*/
/************************************************/
static int cmpmx(const void *p1,const void *p2)

{
MXTYP *m1;
MXTYP *m2;

m1=((MXTYP **)p1)[0];
m2=((MXTYP **)p2)[0];
return m1->preference-m2->preference;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to query DNS information	*/
/*						*/
/************************************************/
static int myquery(const char *dname,int class,int type,
                   unsigned char *answer, int anslen)

{
#define	OPMQ	"unidns.c:myquery,"
int got;

if ((got=res_query(dname,class,type,answer,anslen))<0) {
  got=-1;			/*trouble trouble*/
  switch (errno) {
    case  0		:	/* do not care	*/
      got=0;			/*NO answer!	*/
      break;
    case ETIMEDOUT	:	/*rqst timed out*/
      (void) clm_alert(0,"%s Timeout on domain <%s>!",OPMQ,dname);
      break;
    default	:
      (void) clm_alert(0,"%s query error for domain <%s> type='%d' "
		         "(error=<%d/%s>)",OPMQ,dname,type,errno,strerror(errno));
    }
  }
else {
  if (got>anslen) {
    (void) clm_alert(0,"%s query too by for domain <%s> type='%d' (got=%d)",
		       OPMQ,dname,type,got);
    got=anslen;
    }
  }
return got;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if IP is a local IP	*/
/*	return true if submitted IP is local.	*/
/*						*/
/************************************************/
static int islocalip(char *ipnum)

{
return ((strcmp(ipnum,LOCIP4)==0)||(strcmp(ipnum,LOCIP6)==0));
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extracte texte information	*/
/*	from DNS record.			*/
/*	SMTP client status.			*/
/*						*/
/************************************************/
static char **extracting(RSPTYP *rsp,int rsplen,char *request,char *field)

{
#define	OPEX	"unidns.c:extracting,"

char **list;
int question;
int answer;
const unsigned char *msg;
const unsigned char *eom;
const unsigned char *cp;
int rc;
char buf[NS_PACKETSZ+1];

list=(char **)0;
question=ntohs(rsp->hdr.qdcount);
answer=ntohs(rsp->hdr.ancount);
msg=(unsigned char *)rsp;
eom=(unsigned char *)rsp+rsplen;
cp=(unsigned char *)rsp+sizeof(HEADER);
while ((question-->0)&&(cp<eom)) {
  if ((rc=dn_expand(msg,eom,cp,(char *)buf,sizeof(buf)-1))<0)
    break;
  cp +=rc+QFIXEDSZ;
  }
while((answer-->0)&&(cp<eom)) {
  if ((rc=dn_expand(msg,eom,cp,(char *)buf,sizeof(buf)-1))<0) {
    break;
    }
  if (rc>0) {
    int type;
    int class;
    long ttl;
    int dlen;
    int collected;
    const unsigned char *txtptr;
    char *dynbuf;

    cp+=rc;
    GETSHORT(type,cp);
    GETSHORT(class,cp);
    GETLONG(ttl,cp);
    GETSHORT(dlen,cp);
    if (class!=1) {
      (void) clm_alert(0,"%s unexpected class='%d' on request <%s> field=<%s>",
			  OPEX,class,request,field);
      cp+=dlen;
      break;		/*no need to go further			*/
      }
    collected=0;
    dynbuf=(char *)0;
    switch (type) {
      case T_MX		:
        if ((rc=dn_expand(msg,eom,cp+2,(char *)buf,sizeof(buf)-1))>0) {
          int pref;
	  char info[30];

	  pref=(((unsigned int)cp[0])<<8)|((unsigned int)cp[1]);
	  (void) snprintf(info,sizeof(info),"%d",pref);
          list=(char **)vec_addlstlst((void **)list,(void *)strdup(info));
          list=(char **)vec_addlstlst((void **)list,(void *)strdup(buf));
	  }
	break;
      case T_CNAME	:	
	#ifdef GETCNAME_2011
	/*we are not interested to collect CNAME for now (JMP Feb 2011)	*/
        if ((rc=dn_expand(msg,eom,cp,(char *)buf,sizeof(buf)))>0) {
          list=(char **)vec_addlstlst((void **)list,(void *)strdup("100"));
          list=(char **)vec_addlstlst((void **)list,(void *)strdup(buf));
	  }
	#endif
	break;
      case T_SPF	:
      case T_TXT	:
	txtptr=cp;
        dynbuf=(char *)calloc(10,sizeof(char));
	while (collected<dlen) {
          u_int taille;

	  (void) memset(buf,'\000',sizeof(buf));
          taille=((u_int)(txtptr[0]))&0X0FF; 
	  txtptr++;
          (void) snprintf(buf,sizeof(buf),"%.*s",taille,txtptr);
	  dynbuf=(char *)realloc(dynbuf,strlen(dynbuf)+strlen(buf)+5);
	  (void) strcat(dynbuf,buf);
	  collected+=taille+1;
	  txtptr+=taille;
          }
        list=(char **)vec_addlstlst((void **)list,(void *)dynbuf);
	break;
      case T_A		:	{
        u_int32_t *addr;

        addr=(u_int32_t *)calloc(1,sizeof(u_int32_t));
        *addr=ntohl(*((u_int32_t *)cp));
        list=(char **)vec_addlstlst((void **)list,(void *)addr);
	}
	break;
      case T_AAAA	:	
	(void) clm_alert(0,"%s Unexpected T_AAAA field",OPEX);
	break;
      case T_PTR	:
        if ((rc=dn_expand(msg,eom,cp,(char *)buf,sizeof(buf)-1))>0) {
          list=(char **)vec_addlstlst((void **)list,(void *)strdup(buf));
	  }
	break;
      default		:
	(void) clm_alert(0,"%s Unexpected '%d' field",OPEX,type);
	break;
      }
    cp+=dlen;
    }
  }
return list;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract a hostname TXT	*/
/*	If the domain doesn't exist or domain	*/
/*	do not have a TXT, a void pointer is	*/
/*	returned.				*/
/*						*/
/*						*/
/************************************************/
static char **gettxt(char *hostname,int ttype,char *field)

{
char **inftxt;
int answer;
RSPTYP rsp;

inftxt=(char **)0;
(void) memset(rsp.buf,'\000',sizeof(rsp.buf));
switch (answer=myquery(hostname,C_IN,ttype,rsp.buf,sizeof(rsp.buf))) {
  case -1	:	/*timeout		*/
    break;
  case  0	:	/*no data available	*/
    break;
  default	:	/*data			*/
    inftxt=extracting(&rsp,answer,hostname,field);
    break;
  }
return inftxt;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if the string is 	*/
/*	really an ipaddress.			*/
/*	Return a dynamic clean ipaddress string	*/
/*	if it is not an IP address return null.	*/
/*						*/
/************************************************/
AFNTYP *dns_isipaddress(char *ipaddr)

{
AFNTYP *afnnum;
char *cleanip;
int phase;
int proceed;


afnnum=(AFNTYP *)0;
cleanip=(char *)0;
phase=0;
proceed=(ipaddr!=(char *)0);
while (proceed==true) {
  switch (phase) {
    case 0	:	/*is it an ipaddr ?	*/
      if (ipaddr[0]!='[') 
	proceed=false;
      break;
    case 1	:	/*duplicating string	*/
      cleanip=strdup(ipaddr+1);
      break;
    case 2	:	/*cleaning 		*/
      if (cleanip!=(char *)0) { /*always	*/
	char *ptr;
	
	if ((ptr=strrchr(cleanip,']'))!=(char *)0)
	  *ptr='\000';
	else
	  phase=999;	/*wrong format		*/
        if (strlen(cleanip)==0)
	  phase=999;	/*no IP address content!*/
	}
      break;
    case 3	:	/*converting		*/
      if ((afnnum=afn_getipnum(cleanip))==(AFNTYP *)0) 
	(void) clm_alert(0,"%s Unable to convert IP <%s> (error=<%s>)",
			   "unidns.c:dns_isipaddress",cleanip,strerror(errno));
      break;
    default	:	/*SAFE Guard		*/
      (void) free(cleanip);
      proceed=false;
      break;
    }
  phase++;
  }
return afnnum;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to extract ipaddr from addr	*/
/*						*/
/************************************************/
char *dns_addrip(struct addrinfo *addinf)

{
static char straddr[INET6_ADDRSTRLEN];

char *ipaddr;
void *addr;

ipaddr=straddr;
addr=(void *)0;
switch (addinf->ai_family) {
  case  AF_INET6	:	/*IPV6		*/
    addr=&(((struct sockaddr_in6 *)addinf->ai_addr)->sin6_addr);
    break;
  case  AF_INET		:	/*IPV4		*/
  default		:
    addr=&(((struct sockaddr_in *)addinf->ai_addr)->sin_addr);
    break;
  }
if (inet_ntop(addinf->ai_family,addr,straddr,sizeof(straddr))==(char *)0) {
  (void) clm_alert(0,"%s Unable to find IP for hostname <%s> (error=<%s>)",
		     "unidns.c:dns_addrip,",addinf->ai_canonname,strerror(errno));
  ipaddr=(char *)0;
  }
return ipaddr;
}

/*
^L
*/
/************************************************/
/*						*/
/*	Routine to extract iplist in string	*/
/*	format from addr list.			*/
/*						*/
/************************************************/
char **dns_addriplist(struct addrinfo *addrlst)

{
char **iplist;
struct addrinfo *pif;

iplist=(char **)0;
for (pif=addrlst;pif!=(struct addrinfo *)0;pif=pif->ai_next) {
  char *straddr;

  if ((straddr=dns_addrip(pif))!=(char *)0)
    iplist=(char **)vec_addlstlst((void **)iplist,(void *)strdup(straddr));
  }
return iplist;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract a hostname SPF	*/
/*	Search first for an SPF record, if not	*/
/*	implemented search for a TXT record	*/
/*	If the domain doesn't exist or domain	*/
/*	do not have a TXT or SPF, a void pointer*/
/*	is returned.				*/
/*						*/
/*						*/
/************************************************/
char **dns_getspf(char *hostname)

{
char **inftxt;

if ((inftxt=gettxt(hostname,T_SPF,"SPF"))==(char **)0)
  inftxt=gettxt(hostname,T_TXT,"TXT");
return inftxt;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract a hostname TXT	*/
/*	If the domain doesn't exist or domain	*/
/*	do not have a TXT, a void pointer is	*/
/*	returned.				*/
/*						*/
/*						*/
/************************************************/
char **dns_gettxt(char *hostname)

{
return gettxt(hostname,T_TXT,"TXT");
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if we can get	*/
/*	information about the domain.		*/
/*	We are interested to detect timeout	*/
/*	condition early				*/
/*						*/
/************************************************/
int dns_gooddomdns(char *domain)

{
int responding;
int phase;
int proceed;

responding=true;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*is domain defined	*/
      if ((domain==(char *)0)&&(strlen(domain)==0)) 
	phase=999;	/*trouble trouble	*/
      break;
    case 1	:	/*is domain an IP?	*/	
      if (domain!=(char *)0) {	/*always	*/
	AFNTYP *afnnum;

        if ((afnnum=dns_isipaddress(domain))!=(AFNTYP *)0) {
	  afnnum=afn_freeipnum(afnnum);
	  phase=999;	/*pure address	*/
	  }
	}
      break;
    case 2	:	/*is DNS with an MX	*/
      if (domain!=(char *)0) {	/*always	*/
	int answer;
	RSPTYP rsp;

	(void) memset(&rsp,'\000',sizeof(rsp)); 
	switch (answer=myquery(domain,C_IN,T_MX,rsp.buf,sizeof(rsp.buf))) {
	  case 0	:	/*no MX answer	*/
	    break;
	  case -1	:	/*time out!	*/
	    responding=false;
	    break;
	  default	:	/*seems to be good	*/
	    phase=999;		/*no need to go further	*/
	    break;
	  }
        }
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return responding;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if an ehlo name is	*/
/*	a FQDN (fully qualified domain name).	*/
/*	condition early				*/
/*						*/
/************************************************/
int dns_isfqdn(char *domain)

{
int fqdn;
struct hostent *hbname;
int phase;
int proceed;

fqdn=false;
hbname=(struct hostent *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*is domain defined	*/
      if ((domain==(char *)0)&&(strlen(domain)==0)) 
	phase=999;	/*trouble trouble	*/
      break;
    case 1	:	/*is domain an IP?	*/	
      if (domain!=(char *)0) {	/*always	*/
	AFNTYP *afnnum;

        if ((afnnum=dns_isipaddress(domain))!=(AFNTYP *)0) {
	  afnnum=afn_freeipnum(afnnum);
	  fqdn=true;	
	  phase=999;	/*pure address	*/
	  }
	}
      break;
    case 2	:	/*is fqdn		*/
      if ((hbname=gethostbyname(domain))!=(struct hostent *)0) 
	fqdn=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return fqdn;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract a domain MX	*/
/*	If the domain doesn't exist or domain	*/
/*	do not have a MX, a void pointer is	*/
/*	returned.				*/
/*	Retuned list is ordered according 	*/
/*	MX preference (low preference first)	*/
/*						*/
/************************************************/
MXTYP **dns_getmx(char *domain)

{
MXTYP **mxlist;

mxlist=(MXTYP **)0;
if ((domain!=(char *)0)&&(strlen(domain)>0)) {
  int answer;
  RSPTYP rsp;
 
  (void) memset(&rsp,'\000',sizeof(rsp)); 
  if ((answer=myquery(domain,C_IN,T_MX,rsp.buf,sizeof(rsp.buf)))>0){
    char **list;

    if (answer>=sizeof(rsp.buf))
      answer=sizeof(rsp.buf)-1;
    if ((list=extracting(&rsp,answer,domain,"MX"))!=(char **)0) {
      register int num;
      register int i;

      num=0;
      for (i=0;list[i]!=(char *)0;i+=2) {
	MXTYP *mx;

	mx=(MXTYP *)calloc(1,sizeof(MXTYP));
	if ((list[i+1]!=(char *)0)&&(strlen(list[i+1])>0)) {
	  mx->preference=atoi(list[i]);
	  mx->mxname=strdup(list[i+1]);
	  mxlist=(MXTYP **)vec_addlstlst((void **)mxlist,(void *)mx);
	  num++;
	  }
        }
      list=(char **)vec_freelstlst((void **)list,VECFREE);
      (void) qsort((void *)mxlist,num,sizeof(MXTYP *),cmpmx);
      }
    }
  }
return mxlist;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	MXLIST.					*/
/*						*/
/************************************************/
MXTYP **dns_cleanmx(MXTYP **mxlist)

{
if (mxlist!=(MXTYP **)0) {
  int i;

  for (i=0;mxlist[i]!=(MXTYP *)0;i++) {
    if (mxlist[i]->mxname!=(char *)0)
      (void) free(mxlist[i]->mxname); 
    (void) free(mxlist[i]);
    }
  (void) free(mxlist);
  mxlist=(MXTYP **)0;
  }
return mxlist;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the complet IP	*/
/*	list for domain FIRST MX		*/
/*						*/
/************************************************/
char **dns_getfirstmxiplst(char *domain)

{
char **iplst;
MXTYP **mxlst;

iplst=(char **)0;
if ((mxlst=dns_getmx(domain))!=(MXTYP **)0) {
  char **stripnum;

  if ((stripnum=dns_getiplist(mxlst[0]->mxname))!=(char **)0) {
    int i;

    for (i=0;stripnum[i]!=(char *)0;i++) {
      iplst=(char **)vec_addlstlst((void **)iplst,(void *)stripnum[i]);
      }
    (void) free(stripnum);
    }
  mxlst=dns_cleanmx(mxlst);
  }
return iplst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return the PTR domain	*/
/*	from an ip number			*/
/*						*/
/************************************************/
char **dns_getptr(AFNTYP *afnnum)

{
char **ptrlist;
int answer;
RSPTYP rsp;
char *raddr;

ptrlist=(char **)0;
if ((raddr=afn_reversipnum(afnnum))!=(char *)0) {
  char *reversip;

  (void) asprintf(&reversip,"%s.%s",raddr,afn_reversdom(afnnum));
  if ((answer=myquery(reversip,C_IN,T_PTR,rsp.buf,sizeof(rsp.buf)))>0){
    ptrlist=extracting(&rsp,answer,reversip,"PTR");
    }
  (void) free(reversip);
  (void) free(raddr);
  }
return ptrlist;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if an Ip (norig) is	*/
/*	part of domain A record list		*/
/*						*/
/************************************************/
int dns_matchIPrec(char *hostname,AFNTYP *afnnum,int mask)

{
#define	OPRC	"unidns.c:dns_matchIPrec,"

int match;
struct addrinfo hints;
struct addrinfo *res;
int status;

match=false;	
(void) memset(&hints, 0, sizeof(hints));
hints.ai_family=PF_UNSPEC;
hints.ai_socktype=SOCK_STREAM;
hints.ai_flags=HINTFLG;
if ((status=getaddrinfo(hostname,"",&hints,&res))==0) {
  match=true;
  if (mask!=0) {/*do not compare IP with 0/0,::0*/
    int proceed;
    struct addrinfo *rp;

    proceed=true;
    match=false;
    for (rp=res;rp!=(struct addrinfo *)0;rp=rp->ai_next) {
      AFNTYP *addrnum;
  
      if ((addrnum=afn_getaddrinfo(rp))==(AFNTYP *)0) {
	(void) clm_alert(0,"%s Unable to resolve addr struct for hostname <%s> (bug?)",
			     OPRC,hostname,strerror(errno));
	continue;	
	}
      switch (afn_cmpipnum(addrnum,afnnum,mask)) {
	case false	:/*no equal continue	*/
	  break;
	case true	:/*found it equal!	*/
	  match=true;
	  proceed=false;
	  break;
	case -1		:/*trouble trouble	*/
	  (void) clm_alert(0,"%s Unable to compare IP for hostname <%s> (error=<%s>)",
			     OPRC,hostname,strerror(errno));
	  proceed=false;
	  break;
	}
      addrnum=afn_freeipnum(addrnum);
      if (proceed==false)
        break;
      }
    }
  (void) freeaddrinfo(res);
  }
else {
  (void) clm_alert(0,"%s Unable to get '%s' host IP (error='%s')",
		     OPRC,hostname,gai_strerror(status));
  }
return match;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert a string IPNUM to	*/
/*	a class IP nume (like IP/mask)		*/
/*						*/
/************************************************/
char *dns_getipclass(char *ipnum,int cidr)

{
char *ipclass;
AFNTYP *afnnum;

ipclass=(char *)0;
if ((afnnum=afn_getipnum(ipnum))!=(AFNTYP *)0) {
  char *strnum;

  if (afnnum->afntype==AF_INET6) 
    cidr*=4;
  (void) afn_maskip(afnnum,cidr);
  strnum=afn_stripnum(afnnum);
  (void) asprintf(&ipclass,"%s/%d",strnum,cidr);
  (void) free(strnum);
  afnnum=afn_freeipnum(afnnum);
  }
return ipclass;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if the listen IP is 	*/
/*	a declared MX for a specific domain	*/
/*	Return false if not the case.		*/
/*						*/
/************************************************/
DNSENUM dns_isgoodmx(char *domain,char *listenip)

{
DNSENUM isgoodmx;
MXTYP **mxlst;

isgoodmx=dns_nomx;
if ((mxlst=dns_getmx(domain))!=(MXTYP **)0) {
  int i;
  AFNTYP *afnnum;

  isgoodmx=dns_extmx;
  afnnum=afn_getipnum(listenip);
  for (i=0;mxlst[i]!=(MXTYP *)0;i++) {
    int mask;

    mask=32;
    if (afnnum->afntype==AF_INET6)
      mask=128;
    if (dns_matchIPrec(mxlst[i]->mxname,afnnum,mask)==true) {
      if (i==0) 
        isgoodmx=dns_bestmx;
      else
        isgoodmx=dns_ourmx;
      break;
      }
    }
  afnnum=afn_freeipnum(afnnum);
  mxlst=dns_cleanmx(mxlst);
  }
return isgoodmx;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check a version number	*/
/*	against an expected one and the last	*/
/*	current one as reported via DNS.	*/
/*	Report false if version is not the same	*/
/*	and return the current one.		*/
/*						*/
/************************************************/
int dns_chkversion(char *domain,char **version)

{
#define	OPCV	"unidns.c:dns_chkversion,"

int itsok;
char **txt;
char *lastvers;

itsok=false;
(void) asprintf(&lastvers,"%s","Unknown");
if (*version==(char *)0) {
  (void) clm_alert(0,"%s version not preset for domain <%s> (bug?)",OPCV,domain);
  *version=strdup("");
  }
if ((txt=dns_gettxt(domain))!=(char **)0) {
  char *ptr;

  if ((ptr=strchr(txt[0],':'))!=(char *)0)
    *ptr='\000';
  (void) free(lastvers);
  (void) asprintf(&lastvers,"%s",txt[0]);
  txt=(char **)vec_freelstlst((void **)txt,VECFREE); 
  }
if (strcmp(*version,lastvers)==0) {
  (void) free(lastvers);
  itsok=true;
  }
else {
  (void) free(*version);
  *version=lastvers;
  }
return itsok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check a SAFE version number*/
/*	against an expected one and the last	*/
/*	current one as reported via DNS.	*/
/*	Report false if version is not the same	*/
/*	and return the current one.		*/
/*						*/
/************************************************/
int dns_chksafeversion(char **version)

{
int compare;
const char *curvers;
char *dnsversion;
char domain[200];

compare=0;
curvers=clm_getvers();
dnsversion=strdup(curvers);
(void) snprintf(domain,sizeof(domain),"version.%s.%s",appname,SAFEDOM);
if (dns_chkversion(domain,&dnsversion)!=true) {
  int i;
  int main1;
  int main2;
  int extens1;
  int extens2;
  u_long numvers[2];
  float diff;

  main1=0;
  main2=0;
  extens1=0;
  extens2=0;
  numvers[0]=0.0;
  numvers[1]=0.0;
  for (i=0;i<2;i++) {
    const char *ptr;

    ptr=(char *)0;
    switch (i) {
      case  0	:
        ptr=(const char *)dnsversion;
        break;
      case  1	:
        ptr=curvers;
        break;
      }
    if (sscanf(ptr,"%d.%d-%d.%d",&main1,&main2,&extens1,&extens2)!=4)
      break;
    numvers[i]=(((((main1*100)+main2)*1000)+extens1)*1000)+extens2;
    }
  diff=numvers[0]-numvers[1];
  if (diff!=0) {
    compare=-1;
    if (numvers[0]<numvers[1])
      compare=1;
    }
  *version=strdup(dnsversion);
  }
(void) free(dnsversion);
return compare;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if an E-mail address	*/
/*	is a done with a fake domain		*/
/*	Return true is domain is OK, false	*/
/*	otherwise.				*/
/*						*/
/************************************************/
int dns_gotmxorip(char *email)

{
int status;
char *dom;
MXTYP **mxlst;
struct hostent *hbname;
int phase;
int proceed;

status=true;
dom=(char *)0;
mxlst=(MXTYP **)0;
hbname=(struct hostent *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase)  {
    case 0	:	/*lets extract domain	*/
      if ((dom=strchr(email,'@'))==(char *)0)
	dom=email;
      else
        dom++;
      break;
    case 1	:	/*is it an IP number?	*/
      if (dom!=(char *)0) {	/*always	*/
	AFNTYP *afnnum;

        if ((afnnum=dns_isipaddress(dom))!=(AFNTYP *)0) {
	  afnnum=afn_freeipnum(afnnum);
	  proceed=false;	/*pure address	*/
	  }
	}
      break;
    case 2	:	/*localdomain?		*/
      if (strcmp(dom,LOCALHOST)==0)
	 proceed=false;	/*dom=localhost.domain?	*/
      if (strcmp(dom,LOCALDOMAIN)==0)
	 proceed=false;	/*OK dom is localdomain	*/
      break;
    case 3	:	/*is domain set right?	*/
      if (dns_gooddomdns(dom)==false) 
	phase=999;	/*no domain answer!	*/
      break;
    case 4	:	/*check for an MX	*/
      if ((mxlst=dns_getmx(dom))!=(MXTYP **)0) {
	int i;

        for (i=0;(mxlst[i]!=(MXTYP *)0)&&(phase<999);i++) {
	  char **addrlist;

          if ((addrlist=dns_getiplist(mxlst[i]->mxname))!=(char **)0) {
	    char **ptr;

	    proceed=false;
	    for (ptr=addrlist;*ptr!=(char *)0;ptr++) {
	      if (islocalip(*ptr)==true) {
		(void) clm_alert(0,"Found IP <%s> for domain <%s> MX <%s> (Rejected!)",*ptr,dom,mxlst[i]->mxname);
		phase=999;/*MX badly configured	*/
		proceed=true;
	        break;
		}
	      }
            addrlist=(char **)vec_freelstlst((void **)addrlist,VECFREE);
	    }
	  }
        mxlst=dns_cleanmx(mxlst);
	}
      break;
    case 5	:	/*check for an IP	*/
      if (dom!=(char *)0) {	/*always	*/
	char **addrlist;

        if ((addrlist=dns_getiplist(dom))!=(char **)0) {
          char **ptr;

	  proceed=false;/*There an IP address		*/
	  for (ptr=addrlist;*ptr!=(char *)0;ptr++) {
	    if (islocalip(*ptr)==true) {
	      (void) clm_alert(0,"Found IP <%s> for domain <%s> (Rejected!)",*ptr,dom);
	      phase=999;/*MX badly configured	*/
	      proceed=true;
	      break;
	      }
            }
          addrlist=(char **)vec_freelstlst((void **)addrlist,VECFREE);
	  }
	}
      break;
    case 6	:	/*is there an Name		*/
      if ((hbname=gethostbyname(dom))!=(struct hostent *)0) 
	proceed=false;/*at least the name is defined	*/
      break;
    default	:	/*SAFE guard, no luck		*/
      status=false;
      proceed=false;
      break;
    } 
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a list of IP related*/
/*	to a host name.				*/
/*						*/
/************************************************/
char **dns_getiplist(char *hostname)

{
char **iplst;
struct addrinfo hints;
struct addrinfo *res;

iplst=(char **)0;
(void) memset(&hints, 0, sizeof(hints));
hints.ai_family = PF_UNSPEC;
hints.ai_socktype = SOCK_STREAM;
if (getaddrinfo(hostname,(char *)0,&hints,&res)==0) {
  iplst=dns_addriplist(res);
  (void) freeaddrinfo(res); 
  }
return iplst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return running server	*/
/*	first IP.				*/
/*						*/
/************************************************/
char *dns_getfirstip(char *hostname)

{
char *firstip;
char **iplist;

firstip=(char *)0;
if ((iplist=dns_getiplist(hostname))!=(char **)0) {
  if (strlen(iplist[0])>0)
    firstip=strdup(iplist[0]);
  iplist=(char **)vec_freelstlst((void **)iplist,VECFREE);
  }
if (firstip==(char *)0)
  firstip=strdup(LOCIP4);
return firstip;
}

/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return running server	*/
/*	hostname.				*/
/*						*/
/************************************************/
char *dns_getmyname()

{
char myname[200];

if (gethostname(myname,sizeof(myname)-1)<0) {
  (void) clm_alert(0,"%s Unable to get my own name (error=<%s> config?)",
		     "unidns.c:dns_getmyname,",strerror(errno));
  (void) strcpy(myname,LOCALHOST);
  }
return strdup(myname);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect MX or IP where	*/
/*	email can be send.			*/
/*						*/
/************************************************/
MXTYP **dns_getmxorip(char *domain)

{
MXTYP **mxlst;
char **iplist;
int phase;
int proceed;

mxlst=(MXTYP **)0;
iplist=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*findind MX		*/
      if ((mxlst=dns_getmx(domain))!=(MXTYP **)0) 
	phase=999;	/*found the MX list	*/
      break;
    case 1	:	/*findind IP		*/
      if ((iplist=dns_getiplist(domain))!=(char **)0) {
	int i;

	for (i=0;iplist[i]!=(char *)0;i++) {
	  MXTYP *mx;

	  mx=(MXTYP *)calloc(1,sizeof(MXTYP));
	  mx->preference=1;
	  mx->mxname=iplist[i];/*moving data ptr*/
	  mxlst=(MXTYP **)vec_addlstlst((void **)mxlst,(void *)mx);
	  }
	(void) free(iplist);
        }
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return mxlst;
}
