/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	Define all routine to manage SMTP low	*/
/*	level exchange.				*/
/*						*/
/************************************************/
#ifndef UNIEML
#define UNIEML

/*default SMTP originator name			*/
#define	POSTM	"postmaster"

/*email quarantine dirname			*/
#define	QUARDIR	"quarantine"

/*define banned table status value		*/
#define	BANUKN	0	/*no entry in table	*/
#define	BANSTD	1	/*remote status not set	*/
#define	BANBAN	2	/*remote is banned	*/

/*define trusted table status flag		*/
#define	TRSUKN	0x0000	/*no entry in table	*/
#define	TRSTRS  0x0001	/*remote is trusted	*/
#define	TRSTUS  0x0002	/*remote trusting us	*/

/*defining FROM status				*/
#define	BADFROM	0x00001	/*big Trouble		*/
#define	AUTH	0x00002	/*from authenticated	*/
#define	RELAY	0x00004	/*remote IP smart relay	*/
#define	FRELAY	0x00008	/*remote IP forced relay*/
#define	NOMX	0x00010	/*From domain without MX*/
#define	SPFFAIL	0x00020	/*From domain is SPF BAD*/
#define	ISSBL	0X00040	/*remote server is SBL	*/
#define	AUTHALR	0X00080	/*diff AUTH ORIGINATOR	*/
#define	BADEHLO	0X00100	/*Bad info given to Helo*/
#define	RBANNED	0X00200	/*Remote IP is banned	*/
#define	TRSTED	0X00400	/*Remote is trusted	*/
#define	VALDTED	0X00800	/*Remote is trusting us	*/
#define	FNOVERS	0X01000	/*no reverse-address	*/
#define	BADUSR	0X02000	/*user is unknown	*/
#define	RLYUSR	0X04000	/*This is relayed user	*/

/*MENTOR avail command				*/
#define	CKCODE  "CKCODE"/*Mentor access code	*/
#define RQSTID	"RQSTID"/*Request uniq ID	*/
#define ANSWID	"ANSWID"/*answer to RQSTID	*/
#define	STRNOOP	"NOOP"	/*NO-OP code		*/

/*SMTP avail command				*/
#define	MAILF	"MAIL FROM"
#define	RCPT	"RCPT TO"
#define SQUIT	"QUIT"
#define STRTTLS	"STARTTLS"
#define	HELO	"HELO %s"
#define	EHLO	"EHLO %s"
#define	DATA	"DATA"
#define	RSET	"RSET"
#define	HLOTRS	"TRUSTED"

/*Mentor API					*/
#define SAUTH	"AUTH"	/*check auth  ->mentor	*/
#define SCHK	"CHK"	/*check remote->mentor	*/
#define SSTA	"STA"	/*email stats ->mentor	*/
#define	SORIG	"ORIGIN"/*Mail origin data	*/
#define	SSUBJ	"SUBJ"	/*Email subject		*/
#define	SSPM	"SPM"	/*Spam level.		*/

#define	RCPTEXT	"R"	/*RCPT file extension	*/
#define	DATAXXT	"D"	/*E-mail data extension	*/
#define	MELQUAR	"-Q"	/*E-mail quarantine ext	*/
#define	MELRLAY	"-R"	/*E-mail relayed ext	*/
#define	MELDRCT	"-D"	/*E-mail direct ext	*/

/*clement/mentor API definition				*/
#define	CURTIME	"CURTIME"	/*operation current time*/
#define	TKNOTE	"TKNOTE"	/*take note from clement*/
#define	GVADV	"ADVICE"	/*ADVICE request	*/
#define	MYADV	"MYADV"		/*ADVICE Response	*/
#define	CLIP	"CLIP"		/*remote CLient IP num	*/
#define	MSIZE	"SIZE"		/*Email size		*/
#define	CKFROM	"CKFROM"	/*check FROM parameters	*/
#define	CKRCPT	"CKRCPT"	/*check RCPT validity	*/
#define	RFRSH	"RFRSH"		/*user status refresh	*/
#define	CKID	"CKID"		/*Check user id		*/
#define	CLAMVS	"CLAMVS"	/*Ask for clamav version*/
#define	CHKVIR	"CHKVIR"	/*check Email for virus	*/
#define GETSPM	"GETSPM"	/*get E-mail spam level	*/
#define	DFLTDOM	"GETDFLTDOM"	/*get default domain	*/
#define	RQSTSND	"RQSTSEND"	/*rqst chkr send email	*/
#define	REPTLOG	"RPTLOG"	/*store emlog in SQL DB	*/
#define	SESLOG	"SESLOG"	/*session logs in SQL DB*/
#define	BANTRST	"BANTRUST"	/*get ban/trust status	*/
#define	CHKTRST	"CHKTRUST"	/*parse data stream	*/
#define	CHKEMID	"CHKEMID"	/*Check previous MSG-ID	*/


#define MFROM	"M-From"	/*Mail SMTP from info	*/
#define HFROM	"H-From"	/*Mail header from info	*/
#define	WITHMX	"MX"		/*send Email to an MX	*/


/*Header add on extension marker			*/
#define	XCLEM	"X-Clement-"

/*SMTP protocol status				*/
#define	GIVADV	110	/*Request an advice	*/
#define	RETADV	120	/*advice return		*/
#define	SENADV	130	/*send file advice	*/
#define	QUAADV	131	/*Quarantine advice	*/
#define	DISADV	140	/*Discarded advice	*/
#define	RLYADV	141	/*Advice E-mail relayed	*/
#define	MNTCHK	150	/*Check done by mentor	*/
#define	TORELAY	151	/*relay mail to rcpt	*/
#define	INRELAY	152	/*relay pending to rcpt	*/
#define	BOUNCE	153	/*bounce mail back orig	*/
#define	QUARAN	154	/*Mail in Quarantine	*/
#define	RETRY	155	/*need to retry sending	*/
#define	SIGNON	220	/*signon information	*/
#define	QUITOK	221	/*status on QUIT	*/
#define	IDOK	235	/*Authentication ID OK	*/
#define	CMDOK	250	/*Everything OK		*/
#define	RELAYOK	251	/*Mail can be relayed	*/
#define	FWRDOK	252	/*Mail forward to inside*/
#define	SENDOK	253	/*Mail not on best MX	*/
#define	ALIOK	254	/*Mail is an alias	*/
#define	MAILPAL	255	/*contact initiated	*/
#define	MUSTFWD	256	/*MUST forward data	*/
#define	OVEROK	257	/*MUST be accepted	*/
#define	SENDB64	334	/*Send B64 data		*/
#define	DATAOK	354	/*ready to accept data	*/
#define	TRUSTOK	355	/*accept trusted data	*/
#define	TMPERR	400	/*Temporary error	*/
#define	ERRPROC	451	/*local error/processing*/ 
#define	CHKUSR	452	/*Need check user exist	*/
#define TOOMANY	453	/*too many mail in queue*/
#define TLSNRDY	454	/*TLS Not ready		*/
#define	DLYUSR	455	/*accepting user delayed*/
#define	ALREAD	456	/*already connected	*/
#define	NEWSITE	457	/*new to me		*/
#define	NOSNDR	458	/*sender not connected	*/
#define	NOCHKR	459	/*No checker available	*/
#define	NOANSWR	460	/*NO answer from remote	*/
#define	SPAMMER	462	/*is detected spammers	*/
#define	SYNERR	500	/*syntax error		*/
#define	BADPAR	501	/*error in parameters	*/
#define	CMDBAD	502	/*cmd not implemented	*/
#define	BADSEQ	503	/*bad command sequence	*/
#define	BADAUTH	504	/*AUTH not available	*/
#define	NOOP	505	/*Not allowed operation	*/
#define	UKNUSR	506	/*Sending user unknow	*/
#define	ONHOLD	507	/*system no ready	*/
#define	MAXDAY	508	/*sending too many today*/
#define	BULKOVR	509	/*Bulk overload		*/
#define	HDRLOOP	510	/*header show email loop*/
#define	BADID	535	/*User ID rejected	*/
#define	EXPIRED	536	/*Contract date expired	*/
#define	REJECT	550	/*E-mail/domain rejected*/
#define	NOLOC	551	/*User not local	*/
#define	MALABRT	552	/*mail aborted		*/
#define	CMDABRT	553	/*aborted command	*/
#define	FAILED	554	/*transaction failed	*/
#define	NORELAY 555	/*we do not relay	*/
#define	BLCKLST 556	/*site Black listed	*/
#define	ISBIG	557	/*Email is too big	*/
#define	MBFULL	558	/*Mail Box full		*/
#define	OPRELAY 559	/*daemon allowing relay	*/
#define	BADVIR	560	/*VIRUS found in E-mail	*/
#define	DOMNRES	561	/*Domain do not resolve	*/
#define	NOSPAMS 562	/*we do not allow spam	*/
#define	MISSMX	563	/*NO MX for recipient	*/
#define	BADSPF	564	/*SPF info do not comply*/
#define	MANYRJT	565	/*Too many reject	*/
#define	NOTGOOD	566	/*remote server is bad	*/
#define	NODNS	567	/*No DNS on rmt domain	*/
#define	BADMX	568	/*NO MX for originator	*/
#define	BANNED	569	/*Originator is banned	*/
#define	WRGRLY	570	/*wrong domain used	*/
#define	IPBAN	571	/*IP itself is banned!	*/
#define	BADRVRS	572	/*IP without Reverse	*/
#define	NCHKED	601	/*ID not checked yet	*/
#define	UNVALID	602	/*ID may be unvalid	*/
#define	TBRECHK	603	/*ID need to be recheck	*/

typedef	enum	{	/*domain possible status	*/
	da_lock,	/*domain is locked out		*/
	da_good,	/*domain is known and active	*/
	da_resend,	/*domain to be resenr		*/
	da_unknown	/*domain is unknown		*/
	}DACTYP;

typedef	enum	{	/*returned by mentor		*/
	m_indata,	/*waiting to get E-mail		*/
	m_acceptable,	/*Acceptable email found ok	*/
	m_quarantine,	/*Quarantine email found bad	*/
	m_relayed,	/*external mail relayed		*/
	m_toforward,	/*To forward to internal server	*/
	m_toresend,	/*internal mail to be resend	*/
	m_direct,	/*mail which was authenticated	*/
	m_rejected,	/*mail reject first hand	*/
	m_discarded,	/*rcpt already seen		*/
	m_empty,	/*no E_mail contents		*/
	m_virusfound,	/*E-mail with a virus		*/
	m_trouble,	/*Email with trouble		*/
	m_progress,	/*email extract from quarantine	*/
	m_delivered,	/*email reach final destination	*/
	m_toresolv,	/*email is an alias to resolv	*/
	m_delayed,	/*sent email delivery is delayed*/
	m_stayput,	/*To forward email quarantined	*/
	m_forwarded,	/*Mail was forwarded to internal*/
	m_mustfwd,	/*Mail must be forwarded	*/
	m_noaddress,	/*Mail without real address.	*/
	m_ipbanned,	/*Remote ip is fully banned	*/
	m_undefined	/*unknown status		*/
	}MENTYP;

typedef	struct	{
	MENTYP status;	/*User internal status		*/
	int smtpstatus;	/*user official status		*/
	int grpid;	/*rcpt group id			*/
	char *rcpt;	/*SMTP RCPT address		*/
	char *inbehalf;	/*email in behaf from		*/
	char *note;	/*note on user (usually empty)	*/
	char *why;	/*why user is rejected		*/
	}RCPTYP;

typedef	enum	{	/*status type for recipient	*/
	rcpt_unknown,	/*00 recipient unknown		*/
	rcpt_issystem,	/*01 recipient a system user!	*/
	rcpt_islocal,	/*02 recipient a local one	*/
	rcpt_islocalias,/*03 recipient is a local alias	*/
	rcpt_isalias,	/*04 recipient is an alias	*/
	rcpt_inactive,	/*05 account not accessible now	*/
	rcpt_nomx,	/*06 no MX for that domain!	*/
	rcpt_noalias,	/*07 alias list is empty!!	*/
	rcpt_mailtb,	/*08 mail to recipient too big	*/
	rcpt_boxtb,	/*09 recipient box too big	*/
	rcpt_forward,	/*10 forward to another SMTP	*/
			/*check ID as far as possible	*/
	rcpt_resend,	/*11 relay to a better MX	*/
	rcpt_banned,	/*12 recipient is banned	*/
	rcpt_alidirect,	/*13 alias accept remote	*/
	rcpt_direct,	/*14 recipient is very welcome	*/
	rcpt_foreign,	/*15 foreign recipient (relay)	*/
	rcpt_notchecked,/*16 status not checked		*/
	rcpt_forcefwd,	/*17 force forward		*/
	rcpt_fforeign,	/*18 force sending to foreign	*/
	rcpt_loopback,	/*19 MX is a loopback to me	*/
	rcpt_uknstatus	/*20 recipient unknown status	*/
	}RCPTFLG;

typedef	struct	{	/*CKRCPT message structure	*/
	char *efrom;	/*From: Extracted from via EHLO	*/
	char *rcpt;	/*recipient Email.		*/
	char *heloname;	/*remote helo name		*/
	char *peerip;	/*remote ip			*/
	char *locip;	/*local connection IP		*/
	char *locport;	/*local connection port		*/
	int auth;	/*remote authenticated		*/
	u_long taille;	/*msg expected size		*/
	u_int ctxflag;	/*originator context flags	*/
	}CKPTYP;

typedef	struct	{	/*dispatch message structure	*/
	char *session;	/*session name.			*/
	char *peerip;	/*Remote peer IP		*/
	char *peername;	/*Remote peer revers address	*/
	char *locip;	/*clement local IP used		*/
	char *locport;	/*clement local port used	*/
	int auth;	/*authenticated connection	*/
	float spmlvl;	/*spam level			*/
	float spmlim;	/*spam limit			*/
	char *rejectto;	/*where to send trouble info	*/
	char *inbehalf;	/*originator behalf (aliases)	*/
	int grpin;	/*efrom groupid			*/
	char *efrom;	/*SMTP extracted From		*/
	char *hfrom;	/*Email header From		*/
	char *subject;	/*Email subject			*/
	char *msgid;	/*Email msgID			*/
	u_long mtaille;	/*email taille			*/
	int quarneeded;	/*quarantine file needed	*/
	RCPTYP **rcpts;	/*email recipient list		*/
	}DSPTYP;

extern int eml_orgn;	/*allowing ORGN directive	*/

extern const char *eml_strmntstatus(MENTYP menstat);
extern char *eml_ascstat(int status,...);
extern char *eml_standardize(char *string);
extern RCPTYP *eml_freercpt(RCPTYP *rcpt);
extern RCPTYP **eml_freercpts(RCPTYP **rcpts);
extern RCPTYP **eml_mergercpts(RCPTYP **rcpts,RCPTYP **rcp);
extern RCPTYP **eml_addnewrcp(RCPTYP **rcpts,RCPTYP **rcp);
extern RCPTYP *eml_duprcpt(const RCPTYP *rcpt);
extern RCPTYP **eml_duprcpts(RCPTYP **rcpt);
extern CKPTYP *eml_freeckrrcpt(CKPTYP *ckrrcpt);
extern CKPTYP *eml_scanckrrqst(char *ckrrqst);
extern char *eml_mkckrrqst(CKPTYP *scn);
extern char *eml_getdomain(char *email);
extern char *eml_splitdomain(char *email);
extern char **eml_setdispatch(DSPTYP *dsp);
extern DSPTYP *eml_getdispatch(char **dspstr);
extern DSPTYP *eml_cleandispatch(DSPTYP *dsp);
extern char *dsp_advicemsg(DSPTYP *dsp);
extern char *eml_getquarfilename(char *session,char *recipient);
extern int eml_attache(char *infile,char *outfile,char *session,
		       char *from,char *to,char *subject,char **explain);
extern MENTYP eml_getcomstat(int status,MENTYP orgstatus);
#endif
