/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	SUBIPC:					*/
/*	Take Care of all access to shared memory*/
/*						*/
/************************************************/
#include	<errno.h>
#include	<memory.h>
#include	<time.h>
#include	<stdio.h>
#include	<syslog.h>
#include	<unistd.h>
#include	<sys/ipc.h>
#include	<sys/sem.h>
#include	<sys/shm.h>
#include	<sys/time.h>
#
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"utlsoc.h"
#include	"subclm.h"
#include	"uniipc.h"

#if defined(DIST_rh9)	|| \
    defined(DIST_rh80)	|| \
    defined(DIST_rh73) 
/*subroutine semtimedop NOT available in distribution	*/
#define	NEED_SEMTIMEDOP
#endif

#ifdef	_SEM_SEMUN_UNDEFINED
union semun {
	int val;		/*<= value for SETVAL			*/
	struct semid_ds *buf;	/*<= buffer for IPC_STAT & IPC_SET	*/
	unsigned short *array;	/*<= array for GETALL & SETALL		*/
	struct seminfo *__buf;	/*<= buffer for IPC_INFO		*/
	};
#undef _SEM_SEMUN_UNDEFINED
#endif

typedef	enum   {	/*lock flag		*/
  	s_lock,		/*locking memory access	*/
	s_release	/*release memory access	*/
	}LCKFLG;

typedef	struct {/*defining shared memory info	*/
	int semid;	/*semaphore ID		*/
	short sm;	/*semaphore array	*/
	int shmid;	/*shared memory ID	*/
	void *shmem;	/*shared memory area	*/
	}SHRTYP;

static	struct	{
	u_int num;	/*# of shared memory	*/
	VPTR *veclst;	/*shmem vec list ref	*/
	}lst;
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to duplicate semtimedop which	*/
/*	is not available in Linux kernel 2.4	*/
/*						*/
/************************************************/
#ifdef	NEED_SEMTIMEDOP
static int semtimedop(int semid,struct sembuf *sops,
		      unsigned nsops,TIMESPEC *timeout)

{
int statut;
int locerrno;
struct itimerval timer;
struct itimerval oldtimer;

timer.it_interval.tv_sec=0;
timer.it_interval.tv_usec=0;
timer.it_value.tv_sec=timeout->tv_sec;
timer.it_value.tv_usec=timeout->tv_nsec/1000;
(void) setitimer(ITIMER_REAL,&timer,&oldtimer);
/*such setting the semaphore won't wait for ever*/
statut=semop(semid,sops,nsops);
locerrno=errno;
(void) setitimer(ITIMER_REAL,&oldtimer,&oldtimer);
timeout->tv_sec=oldtimer.it_value.tv_sec;
timeout->tv_nsec=oldtimer.it_value.tv_usec*1000;
errno=locerrno;
return statut;
}
#endif
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to localise shared information	*/
/*	within list retunr the point if 	*/
/*	existing.				*/
/*						*/
/************************************************/
static SHRTYP *getshared(void *ptr)

{
SHRTYP *shr;

shr=(SHRTYP *)0;
if (lst.veclst!=(VPTR *)0) {
  register int i;
  register VECLST *vec;

  for (i=0,vec=lst.veclst->vcl;i<lst.veclst->numvec;i++,vec=vec->nxt) {
    register SHRTYP *loc;
   
    loc=(SHRTYP *)vec->payload;
    if (loc->shmem==ptr) {
      shr=loc;
      break;
      }
    }
  }
return shr;
}
/*
^L
*/
/************************************************/
/*                                              */
/*	Routine to sync access to a specific	*/
/*	shared memory.				*/
/*                                              */
/************************************************/
int syncsharedaccess(void *ptr,LCKFLG action,struct timespec *timeout)

{
#define	OPSY	"subipc.c:syncsharedaccess,"

int lockdone;
SHRTYP *shr;
struct sembuf directives;

lockdone=false;
if ((shr=getshared(ptr))==(SHRTYP *)0) {
  (void) clm_alert(0,"%s Unable to find shared memory for ptr '%08lx'!\n",
		      OPSY,(u_long)ptr);
  return false;
  }
directives.sem_num=0;
directives.sem_flg=SEM_UNDO;
switch (action) {
  case s_lock     :
    directives.sem_op=-1;
    while (lockdone==false) {
      int toolate;

      toolate=false;
      if (semtimedop(shr->semid,&directives,1,timeout)==0) {
	lockdone=true;
	break;
	}
      switch (errno) {
	case EAGAIN	:
	  toolate=true;
	  break;
	case EINTR	:	/*timer interrupt	*/
	  break;		/*lets wait more	*/
	default		:
	  (void) clm_core_dump("%s Unable to lock semaphore (error=<%s>)",
			        OPSY,strerror(errno));
	  break;
        }
      if (toolate==true)
	break;
      }
    break;
  case s_release  :
    directives.sem_op=1;
    if (semop(shr->semid,&directives,1)<0) {
      (void) clm_alert(0,"%s Unable to Unlock semaphore '%d' (errno=<%s>)",
			  OPSY,shr->semid,strerror(errno));
      }
    lockdone=true;
    break;
  }
return lockdone;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to assign ONE semaphore		*/
/*						*/
/************************************************/
static SHRTYP *getsem()

{
#define	OPGT	"subipc.c:getsem,"
SHRTYP *shared;

int semid;

shared=(SHRTYP *)0;
if ((semid=semget((long)0,1,IPC_CREAT|0600))>=0) {
  struct semid_ds semid_ds;
  union semun arg;

  shared=(SHRTYP *)calloc(1,sizeof(SHRTYP));
  shared->semid=semid;  
  shared->sm=1;
  (void) semctl(shared->semid,1,SETALL,&(shared->sm));
  semid_ds.sem_perm.gid=getgid();
  semid_ds.sem_perm.uid=getuid();
  semid_ds.sem_perm.mode=0600;	/*semaphore acces protection	*/
  arg.buf=&semid_ds; 
  if (semctl(shared->semid,0,IPC_SET,arg)<0) {
    (void) fprintf(stderr,"%s Unable to change ownershipore (error=<%s>",
		   	   OPGT,strerror(errno));
    (void) fflush(stderr);
    }
  }
else {
  (void) fprintf(stderr,"%s Unable to get semaphore (error=<%s>",
		        OPGT,strerror(errno));
  (void) fflush(stderr);
  }
return shared;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to release a semaphore		*/
/*						*/
/************************************************/
static SHRTYP *dropsem(SHRTYP *shared)

{
if (shared!=(SHRTYP *)0) {
  (void) semctl(shared->semid,1,IPC_RMID,0);
  (void) free(shared);
  shared=(SHRTYP *)0;
  }
return shared;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to request shared memory to 	*/
/*	which access is controled by a semaphore*/
/*						*/
/************************************************/
void *ipc_getshrmem(size_t nmemb,size_t size)

{
#define	OPSH	"subipc.c:getshrmem,"

SHRTYP *sharloc;
void *shrmem;

shrmem=(void *)0;
if ((sharloc=getsem())!=(SHRTYP *)0) {
  int shmid;

  if ((shmid=shmget(IPC_PRIVATE,size*nmemb,IPC_CREAT|SHM_R|SHM_W))>=0) {
    sharloc->shmid=shmid;
    if ((sharloc->shmem=shmat(shmid,(void *)0,0))==(void *)-1) {
      (void) fprintf(0,"%s Unable to assigne shared memory (error=<%s>)",
			OPSH,strerror(errno));
      (void) shmctl(shmid,IPC_RMID,0);
      sharloc=dropsem(sharloc);
      }
    else {
      shrmem=sharloc->shmem;
      (void) memset(shrmem,'\000',(size*nmemb)*sizeof(char));
      lst.veclst=(VPTR *)vec_addveclst((VPTR *)lst.veclst,(void *)sharloc);
      }
    }
  else {
    (void) fprintf(stderr,"%s Unable to get shared memory (error=<%s>)",
		   	  OPSH,strerror(errno));
    sharloc=dropsem(sharloc);
    }
  }
return shrmem;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to release shared memory	*/
/*						*/
/************************************************/
void *ipc_dropshrmem(void *ptr)

{
if (ptr!=(void *)0) {
  register SHRTYP *shr;

  if ((shr=getshared(ptr))!=(SHRTYP *)0) {
    (void) vec_rmveclst(&(lst.veclst),(void *)shr);
    (void) shmdt(shr->shmem);
    (void) shmctl(shr->shmid,IPC_RMID,0);
    (void) dropsem(shr);
    }
  ptr=(void *)0;
  }
return ptr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine which is getting back when	*/
/*	access to shared memory is allowed.	*/
/*						*/
/************************************************/
int ipc_getshraccess(void *ptr,struct timespec *timeout)

{
return syncsharedaccess(ptr,s_lock,timeout);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine which is giving back free access*/
/*	to shared memory.			*/
/*						*/
/************************************************/
int ipc_dropshraccess(void *ptr)

{
return syncsharedaccess(ptr,s_release,(struct timespec *)0);
}
