/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	UNILCK:					*/
/*	Take Care of all process locking	*/
/*	function.				*/
/*						*/
/************************************************/
#include	<string.h>
#include	<syslog.h>
#include	<fcntl.h>
#include	<unistd.h>
#include	<malloc.h>

#include	"lowtyp.h"
#include	"utlprc.h"
#include	"subclm.h"
#include	"unilck.h"

/*Application locking directory			*/
#define	LOCKEXT	"-lock"	/*lock file extension	*/

/*

*/
/************************************************/
/*						*/
/*	procedure to set/unset a lock 		*/
/*	return true if successful,		*/
/*	false otherwise.			*/
/*						*/
/************************************************/
int lck_locking(const char *ident,int lock,int tentative)

{
int done;
char *lockname;

done=false;
(void) asprintf(&lockname,"%s%s%s",clm_appdir(d_lock),ident,LOCKEXT);
if (lock==LCK_LOCK) {
  (void) clm_alert(9,"Request locking <%s>",lockname);
  while (tentative>0) {
    int handle;

    tentative--;
    if ((handle=open(lockname,O_RDWR|O_EXCL|O_CREAT,0640))<0) {
      FILE *fichier;

      if ((fichier=fopen(lockname,"r"))!=(FILE *)0) {
        int pid;
        char strloc[80];

	(void) fgets(strloc,sizeof(strloc)-1,fichier);
	(void) fclose(fichier);
        if (sscanf(strloc,"%d",&pid)==1) {
          (void) clm_alert(1,"Locking, check %d process active",pid);
	  if (prc_checkprocess(pid)==false) {
            (void) clm_alert(1,"Locking, remove unactive lock");
	    (void) unlink(lockname);
            }
          else {
	    if (tentative>0)
              {
              (void) clm_alert(2,"waiting 1 sec to be unlocked");
              (void) sleep(1);
	      }
	    }
          }
        }
      continue;
      }
    else
      {
      char numid[30];

      (void) snprintf(numid,sizeof(numid),"%06d\n",getpid());
      (void) write(handle,numid,strlen(numid));
      (void) close(handle);
      done=true;
      break;
      }
    }
  if (done==false)
    (void) clm_alert(0,"Unable to set locking <%s>",lockname);
  }
else
  {
  (void) clm_alert(9,"Request unlocking <%s>",lockname);
  (void) unlink(lockname);
  done=true;
  }
(void) free(lockname);
return done;
}
