/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	UNIQUE:					*/
/*	Take Care of all access to devices	*/
/*	could be system dependent.		*/
/*	Take care about email transit queue 	*/
/*	management.				*/
/*						*/
/************************************************/
#include	<stdlib.h>
#include	<string.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unique.h"

/*

*/
/************************************************/
/*						*/
/*	Procedure to return a clean subject	*/
/*	acceptable for a shell called via popen	*/
/*	transit number				*/
/*						*/
/************************************************/
static char *cleansubject(char *subject)

{
char *sbjt;

if (subject!=(char *)0) { /*should be always	*/
  register char *ptr;

  sbjt=strdup(subject);
  for (ptr=sbjt;*ptr!='\000';ptr++) {
    if ((*ptr<' ')||(*ptr>'~')) {
      *ptr='?';
      continue;
      }
    switch (*ptr) {
      case '"'	:
      case '\''	:
      case '`'	:
	*ptr='_';
	break;
      default	:
        break;
      }
    }
  }
else
  sbjt=strdup("");
return sbjt;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a very unique	*/
/*	transit number				*/
/*						*/
/************************************************/
u_long que_getuniqtransit()

{
#define	TRTCYCLE 1000

static u_long transit;

transit++;
transit%=TRTCYCLE;
return (getpid()*TRTCYCLE)+transit;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to dump the queue contents	*/
/*	needed for debug procedure only.	*/
/*	file.					*/
/*						*/
/************************************************/
void que_dbgqueuedump(QUETYP *q,const char *mrkr,int num)

{
#define	QDSTR	"_q[%d] session=<%s> transit='%08ld' recipient=<%s@%s> smtpmx=<%s> why=<%s>"
(void) clm_alert(0,"%s"QDSTR,mrkr,num,q->session,q->transit,q->recname,q->recdom,q->smtpmx,q->why);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the queue trace	*/
/*	file.					*/
/*						*/
/************************************************/
char *que_tracename(QUETYP *queue)

{
char *tracename;

(void) asprintf(&(tracename),"%s/%s-t%lu",clm_appdir(d_tobesend),queue->session,queue->numtrace);
return tracename;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a dataname build	*/
/*	around session ID and transit number	*/
/*						*/
/************************************************/
char *que_getdatafilename(DIRENUM dir,QUETYP *queue)

{
char *datafilename;

(void) asprintf(&datafilename,"%s/%s-%05lu",clm_appdir(dir),queue->session,queue->transit);
return datafilename;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a subject for 	*/
/*	email report.				*/
/*						*/
/************************************************/
char *que_getreportsubject(QUETYP *queue,int warning)

{
char *subject;
const char *tsubject;
char *locsubject;
char *cmt;
FILE *fichier;

subject=(char *)0;
tsubject="TSUBJECT";
if (warning==true)
  tsubject="WSUBJECT";
locsubject=cleansubject(queue->subject);
(void) asprintf(&cmt,"/bin/sh %s%s/%s %s %d \"%s\"",
		      clm_appdir(d_usrlib),SHELL,TROUBLE,
		      tsubject,queue->lang,locsubject);
if ((fichier=popen(cmt,"r"))!=(FILE *)0) {
  char line[400];
									
  while (fgets(line,sizeof(line)-1,fichier)!=(char *)0) {
    (void) clm_skiprtlf(line);
    if (strlen(line)>0) {
      subject=strdup(line);
      break;	/*1st not empty line is subject*/
      }
    }
  (void) pclose(fichier);
  }
if (subject==(char *)0) {
  subject=strdup("Subject error");
  }
(void) free(cmt);
(void) free(locsubject);
return subject;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a report trouble for*/
/*	email.					*/
/*						*/
/************************************************/
char **que_getreportdata(QUETYP *queue,int warning)

{
char **report;
const char *trouble;
char *locsubject;
int minutes;
char *cmt;
FILE *fichier;

report=(char **)0;
trouble="TROUBLE";
locsubject=cleansubject(queue->subject);
minutes=((queue->lasttime-queue->start)+59)/60;
if (warning==true)
  trouble="WARNING";
(void) asprintf(&cmt,"%s%s/%s %s %d \"%s@%s\" %s \"%s\" \"%s\" \"%s\" %d %ld",
			clm_appdir(d_usrlib),SHELL,TROUBLE,
			trouble,queue->lang,queue->recname,queue->recdom,
			queue->session,locsubject,
			queue->rmtsrvr,queue->why,
			minutes,queue->upto);
if ((fichier=popen(cmt,"r"))!=(FILE *)0) {
  char line[400];
									
  while (fgets(line,sizeof(line)-1,fichier)!=(char *)0) {
    (void) clm_skiprtlf(line);
    report=(char **)vec_addlstlst((void **)report,(void *)strdup(line));
    }
  (void) pclose(fichier);
  }
(void) free(cmt);
(void) free(locsubject);
return report;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to create a email transit	*/
/*	queue entry.				*/
/*						*/
/************************************************/
QUETYP *que_mkqueue(char *outip,char *smtpmx,char *smtpport,char *session,int transit,
		    char *subject,char *rejectto,char *orgeml,int grpin,char *recname,char *recdom,MMOENU mode,int orgstatus)

{
QUETYP *q;

q=(QUETYP *)calloc(1,sizeof(QUETYP));
q->outip=strdup(outip);
q->smtpmx=strdup(smtpmx);
q->smtpport=strdup(smtpport);
q->session=strdup(session);
q->grpin=grpin;
q->transit=transit;
q->mode=mode;
q->orgstatus=orgstatus;
if (subject==(char *)0)
  subject="";
q->subject=strdup(subject);
if (rejectto!=(char *)0) 
  q->rejectto=strdup(rejectto);
else
  q->rejectto=strdup(orgeml);
q->orgeml=strdup(orgeml);
q->recname=strdup(recname);
q->recdom=strdup(recdom);
return q;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	transit queue entry.			*/
/*						*/
/************************************************/
QUETYP *que_freequeue(QUETYP *queue)

{
if (queue!=(QUETYP *)0) {
  queue->logs=(char **)vec_freelstlst((void **)queue->logs,VECFREE);
  queue->why=clm_freestr(queue->why);
  queue->rmtsrvr=clm_freestr(queue->rmtsrvr);
  queue->recdom=clm_freestr(queue->recdom);
  queue->recname=clm_freestr(queue->recname);
  queue->orgeml=clm_freestr(queue->orgeml);
  queue->rejectto=clm_freestr(queue->rejectto);
  queue->subject=clm_freestr(queue->subject);
  queue->session=clm_freestr(queue->session);
  queue->smtpport=clm_freestr(queue->smtpport);
  queue->smtpmx=clm_freestr(queue->smtpmx);
  queue->outip=clm_freestr(queue->outip);
  (void) free(queue);
  queue=(QUETYP *)0;
  }
return queue;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a 	*/
/*	transite email queue list		*/
/*						*/
/************************************************/
QUETYP **que_freequeues(QUETYP **queues)

{
if (queues!=(QUETYP **)0) {
  int i;
  
  for (i=0;queues[i]!=(QUETYP *)0;i++) {
    queues[i]=que_freequeue(queues[i]);
    }
  (void) free(queues);
  queues=(QUETYP **)0;
  }
return queues;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to duplicate a one queue	*/
/*						*/
/************************************************/
QUETYP *que_dupqueue(QUETYP *queue)

{
QUETYP *clone;

clone=(QUETYP *)calloc(1,sizeof(QUETYP));
clone->session=strdup(queue->session);
clone->transit=queue->transit;
clone->numtrace=queue->numtrace;
clone->outip=strdup(queue->outip);
clone->smtpmx=strdup(queue->smtpmx);
clone->smtpport=strdup(queue->smtpport);
if (queue->rmtsrvr==(char *)0)
  clone->rmtsrvr=strdup("");
else
  clone->rmtsrvr=strdup(queue->rmtsrvr);
clone->subject=strdup(queue->subject);
clone->rejectto=strdup(queue->rejectto);
clone->orgeml=strdup(queue->orgeml);
clone->recname=strdup(queue->recname);
clone->recdom=strdup(queue->recdom);
clone->status=queue->status;
if (queue->why==(char *)0)
  clone->why=strdup("");
else
  clone->why=strdup(queue->why);
clone->mode=queue->mode;
clone->retry=queue->retry;
clone->progress=queue->progress;
clone->start=queue->start;
clone->lasttime=queue->lasttime;
clone->nexttime=queue->nexttime;
clone->upto=queue->upto;
clone->orgstatus=queue->orgstatus;
clone->grpin=queue->grpin;
clone->lang=queue->lang;
if (queue->logs!=(char **)0) {
  register char **ptr;
  int i;

  ptr=queue->logs;
  for (i=0;ptr[i]!=(char *)0;i++)
    clone->logs=(char **)vec_addlstlst((void **)clone->logs,(void *)(strdup(ptr[i])));
  }
return clone;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to duplicate a queue structure*/
/*						*/
/************************************************/
QUETYP **que_dupqueues(QUETYP **queues)

{
QUETYP **clones;

clones=(QUETYP **)0;
if (queues!=(QUETYP **)0) {
  int i;
  
  for (i=0;queues[i]!=(QUETYP *)0;i++) {
    QUETYP *clone;

    if ((clone=que_dupqueue(queues[i]))!=(QUETYP *)0)
      clones=(QUETYP **)vec_addlstlst((void **)clones,(void *)clone);
    }
  }
return clones;
}
