/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	Implement very sub level procedure to 	*/
/*	handle socket function.			*/
/*						*/
/************************************************/
#include	<sys/socket.h>
#include	<errno.h>
#include	<stdio.h>
#include	<string.h>
#include	<netdb.h>
#include	<arpa/inet.h>

#include	"lowtyp.h"
#include	"utlsoc.h"
/*

*/
/************************************************/
/*						*/
/*	Procedure to get a name related to an	*/
/*	IP address.				*/
/*	Return a newly allocated memory which	*/
/*	contents is the needed information.	*/
/*						*/
/*	MODE:					*/
/*		pr_loc  -> local reverse address*/
/*		pr_name -> remote reverse name	*/
/*		pr_ipnum-> remote IP number	*/
/*						*/
/************************************************/
char *soc_getaddrname(int handle,PEERTYP peer,char *unknown)

{
char *rvname;
socklen_t taille;
int status;
struct sockaddr_storage connip;
char host[NI_MAXHOST];

taille=sizeof(connip);
status=-1;
rvname=unknown;
(void) strcpy(host,"");
switch(peer) {
  case pr_locname	:
  case pr_locip		:
    status=getsockname(handle,(struct sockaddr *)&connip,&taille);
    break;
  case pr_remname	:
  case pr_remip		:
    status=getpeername(handle,(struct sockaddr *)&connip,&taille);
    break;
  default	:
    break;
  }
if (status<0) {
  switch (errno) {
    case ENOTCONN	:	/*other side just vanished	*/
      break;
    default		:
      (void) fprintf(stderr,"utlsoc.c:soc_getaddrname, Unable to find "
			     "reverse (error='%s')",strerror(errno));
      (void) fflush(stderr);
      break;
    }
  }
else {
  switch (peer) {
    case pr_locname	:
    case pr_remname	:
      if (getnameinfo((struct sockaddr *)&connip,taille,host,sizeof(host),(char *)0,0,NI_NAMEREQD)==0) 
	rvname=host;
      break;
    case pr_remip	:
    case pr_locip	:
      if (getnameinfo((struct sockaddr *)&connip,taille,host,sizeof(host),(char *)0,0,NI_NUMERICHOST)==0) 
	rvname=host;
      /*rvname=inet_ntoa(intaddr.sin_addr);*/
      break;
    }
  }
return strdup(rvname);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Routine to duplicate ppoll which	*/
/*	is not available in Linux kernel 2.4	*/
/*						*/
/************************************************/
#ifdef	NEED_PPOLL
int ppoll(struct pollfd *fds,nfds_t nfds,const struct timespec *timeout,
	  const sigset_t *set)

{
int ready;
int temps;
sigset_t origmask;

ready=-1;
temps=timeout->tv_sec*1000000;
temps+=timeout->tv_nsec/1000;
(void) sigprocmask(SIG_SETMASK,set,&origmask);
ready=poll(fds,nfds,temps);
(void) sigprocmask(SIG_SETMASK,&origmask,(sigset_t *)0);
return ready;
}
#endif
